//******************************************************************************
//*                                  uBee512                                   *
//*       An emulator for the Microbee Z80 ROM, FDD and HDD based models       *
//*                                                                            *
//*                         Floppy disk support module                         *
//*                                                                            *
//*                   Copyright (C) 2007-2009 Stewart Kay                      *
//******************************************************************************
//
// This module implements functions to access floppy disks and/or disk images
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v4.3.0 - 6 August 2009, S.J.Kay
// - Added disk_t structure members density and datarate. (workerbee)
// - Removed 'side1as0' when used for normal read/writes with LibDsk as these
//   now use xread/xwrite exclusively. (workerbee)
// - 'side1as0' is still valid as it determines how a disk will be formatted.
//   This is done from the fdc.c module.
// - Added Dreamdisk's double sided head flag (0x80) to RAW/DIP image types
//   in disk_read_idfield() function.
// - Added conditional libdsk reporting depending on the --verbose option.
//
// v4.2.0 - 9 July 2009, S.J.Kay
// - Added hard disk image type '.hd0' (305 cylinders, 4 heads, 17 sect/track)
//
// v4.1.0 - 4 July 2009, S.J.Kay
// - Made Endian changes to DIP image format.
// - Added hard disk image type '.hd1' (80 cylinders, 4 heads, 63 sect/track)
//
// v4.0.0 - 8 June 2009, S.J.Kay
// - Fixed segmentation error in disk_open() function for '.temp' files when
//   a create file fails.  Typically caused by a non-existent path.  Writing
//   to the file is now skipped and a file error is reported later when not
//   found.
//
// v3.1.0 - 22 April 2009, S.J.Kay
// - Added conditional message reporting based on --verbose option.
// - Removed all occurrences of console_output() function calls.
// - Changed function name find_file_entry() to find_file_alias().
// - Changes made to disk_read() and disk_write() functions for DISK_DSK type
//   so that mixed 512x10 and 1024x5 tracks will work. This required obtaining
//   information from the (already read) track header each time.
// - Changes made disk_read_idfield() function for DISK_DSK and LIBDSK_DSK
//   to update some disk.imagerec values.  In DISK_DSK the idfield->track,
//   idfield->side, idfield->crc1 and idfield->crc2 values are filled in with
//   values from the dskt structure.
// - Added code to disk_write() and disk_format_track() to adjust the timer
//   exit value 'emu.secs_exit', this is done to prevent exiting while disk
//   write activity may be currently active.
// - Changed all printf() calls to use a local xprintf() function.
//
// v2.7.0 - 18 June 2008, S.J.Kay
// - Added creation of disk images when '.temp' is appended to a raw disk
//   file name. Will be over written next time started with same file name.
// - Added find_file_entry() function call to the disk_open() function to
//   find what file to use for file name aliases defined in a configuration
//   file.
// - Changes to disk_open() function, now uses file name in disk_t structure.
// - Added members, 'drive', 'error' and 'image_name' to the disk_t structure.
// - Disk reporting changed.
// - Added double stepping option to support 48tpi disks in 96tpi DD drives.
// - Added double stepping high density option to support 48tpi disks in
//   96tpi HD drives.
//
// v2.6.0 - 8 May 2008, S.J.Kay
// - Use an external upper case string copy function.
//
// v2.4.0 - 18 February 2008, S.J.Kay
// - Added LibDsk providing disk image, floppy access, allowing other
//   drivers to be employed. The existing image and basic Unices floppy
//   access methods remains.
// - Added a side read method to allow disks in LibDsk to correctly read
//   side 1 when the sector header contains side 0. Microbee DSx0 formats
//   can have this problem depending on the format program used.
// - Added disk_format_track() function.
// - Sector range checking is now carried out in this module.
// - Changed itype numbers, 0 is now 'no type selected'.
// - Removed obsolete and unimplemented disk_create() function.
// - Added disk_init function.
//
// v2.3.0 - 21 January 2008, S.J.Kay
// - Added preliminary support for direct floppy access for Unices.
// - Added modio_t structure.
//
// v2.2.0 - 12 December 2007, S.J.Kay
// - added image write protect feature.  A trailing underscore '_' character
//   makes the image write protected.
// - disk->imagerec.datatrack value changed to 1 for non DS80 formats. Not
//   an issue as the system and data tracks have the same sector offsets for
//   either tracks anyway.
//
// v2.1.0 - 24 October 2007, S.J.Kay
// - Added new function disk_read_idfield() to obtain the ID field from the
//   track, side.
// - Undid the changes made in v2.0.0 to the imagerec.datatrack back to how it
//   is in v1.4.0.
// - Microbee DS80 format fixed,  this format has 4 reserved cylinders not 2.
//   disk->imagerec.datatrack = 4
//
// v2.0.0 - 18 October 2007, S.J.Kay
// - DSK format should now be fully complete.
// - Changes to imagerec.datatrack used by all raw and DIP formats now holds
//   the byte offset instead of number of tracks for first data track.
//
// v1.4.0 - 6 October 2007, S.J.Kay
// - Added alternative shortened 3 character filename extentions for raw images.
// - The '.dsk' format now includes 3.5" Modular and should work with other
//   unknown disk formats.  Sector offset information is now read per track.
// - Fixed bug preventing '.DSnn' and '.SSnn' raw images from being opened.
//   Changed image_types[][5] to be image_types[][6].
//
// v1.1.0 - 14 August 2007, S.J.Kay
// - Added Microbee ds40, ss80, ds80, ds82, ds84 raw disk image types.
// - Modified DSK code to detect all Microbee formats.  This needs testing for
//   some formats. (warning in README)
//
// v1.0.0 - 19 June 2007, S.J.Kay
// The following changes/additions have been made:
// - Added new disk image structures,  this now allows for different floppy
//   images to be used instead of only the 40T DS (SBC) images.
// - disk_open function now uses open with write enabled (r+b). This is
//   required to allow disk_write to work.
// - Added an assert to test max_size variable in disk_read function.
// - Implement the disk_write function.
//
// v0.0.0 - 7 June 2007, S.J.Kay
// Start with "nanowasp" source distribution version 0.22. An emulator for the
// microbee 128k. Copyright (C) 2000-2003  David G. Churchill.

//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 ROM, FDD and HDD based models.
 *  Copyright (C) 2007-2009 Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <stdint.h>

#ifdef USE_LIBDSK
#include <libdsk.h>
#endif

#include "ubee512.h"
#include "function.h"
#include "disk.h"

//==============================================================================
// structures and variables
//==============================================================================
extern char userhome_diskpath[];

extern emu_t emu;
extern model_t modelx;
extern modio_t modio;

#ifdef USE_LIBDSK
//==============================================================================
// Callback function for LibDsk to report messages.
//
//   pass: void
// return: DSK_REPORTFUNC
//==============================================================================
static void disk_libdsk_report (const char *s)
{
 if (emu.verbose)
    xprintf("LibDsk: %s\n", s);
}
                
//==============================================================================
// Callback function for LibDsk report end messages.
//
//   pass: void
// return: DSK_REPORTFUNC
//==============================================================================
static void disk_libdsk_report_end (void)
{
}
#endif

//==============================================================================
// Initialise.
//
// Called from fdc.c fdc_init function
//
//   pass: void
// return: int			0 if no errors, else -1
//==============================================================================
int disk_init (void)
{
#ifdef USE_LIBDSK
 dsk_reportfunc_set(disk_libdsk_report, disk_libdsk_report_end);
#endif
 return 0;   
}

//==============================================================================
// Disk open.
//
// Open a disk format for the disk structure requested.  This can be one of
// the following:
//
// * original v0.22 raw disk format (40T DS DD) Side0=1st 200K, Side1=2nd 200K
// * CPC-EMU dsk format (ubee512 built-in support)
// * dip format (disk image plus) This is an in-house spec not finalised.
// * ds40 - Microbee 40T DS DD (SBC) S0-T0, S1-T0, S0-T1, S1-T1 ...
// * ss80 - Microbee 80T SS DD (CIAB) S0-T0, S0-T1, S0-T2 ...
// * ds80 - Microbee 80T DS DD (modular) S0-T0, S1-T0, S0-T1, S1-T1 ...
// * ds82 - Microbee 80T DS DD (dream) S0-T0, S1-T0, S0-T1, S1-T1 ...
// * ds84 - Microbee 80T DS DD (pjb) S0-T0, S1-T0, S0-T1, S1-T1 ...
// * hdx  - Various preset HDD images with x set to a format number.
// * LibDsk - Many types
//
//   pass: disk_t *disk
// return: int				0 if no error, else -1 
//==============================================================================
int disk_open (disk_t *disk)
{
 diski_t temp_imagerec;

 char filename[SSIZE1];

 int type_start = 0;

 int i;
 int itype_temp;
 char *ch;

 dski_t dski;			// dsk information block
 dskt_t dskt;			// dsk track information block

#ifdef USE_LIBDSK
 dsk_err_t dsk_err;
 dsk_format_t fmt; 
 dsk_cchar_t fdesc;
 dsk_cchar_t fname;
#endif

 char uname[SSIZE1];

// these formats are for the built in RAW and DSK driver (not LibDsk)
 char const image_types[][6] =
 {
  ".DIP",			// Disk Image Plus (in-house format)
  ".DSK",			// standard CPC DSK format
  ".IMG",			// Nanowasp v0.22 format
  ".NW",			// Nanowasp v0.22 format
  ".DS40",			// raw images
  ".D40",  
  ".SS80",
  ".S80",
  ".DS80",
  ".D80",  
  ".DS82",
  ".D82",  
  ".DS84",
  ".D84",
  ".HD0",
  ".HD1",
  ""
 };

 // no type selected
 disk->itype = 0;
 
 disk->error = 0;

 // see if the name has an alias file name entry
 if (emu.alias_disks)
    {
     if (! find_file_alias(ALIASES_DISKS, disk->filename, filename))
        strcpy(filename, disk->filename);
    }
 else
    strcpy(filename, disk->filename);   

#ifdef USE_LIBDSK
 if (! strlen(disk->libdsk_format)) // if not using LibDsk
    {
#endif    

     // check and create a temporary file (will grow as needed)
     ch = strstr(filename, ".temp");
     if (ch)
        {
         i = ch - filename;
         if (i == (strlen(filename) - 5))
            {
             disk->fdisk = open_file(filename, userhome_diskpath, disk->filepath, "wb");
             if (disk->fdisk)
                {
                 for (i = 0; i < 51200; i++)  // first 10 ~tracks if 10 @ 512 s/t
                    putc(0xe5, disk->fdisk);
                 fclose(disk->fdisk);
                } 
            }
        }   

      // direct floppy access. ie /dev/fd0u800/.ds84
#ifdef MINGW
#else
     if (strstr(filename, "/dev/") == filename)
        {
         char *ch = strstr(filename, "/.");         
         if (ch)
            *ch = 0;
         type_start = 4;   
        }
#endif 

     // open the image or floppy
     i = (test_file(filename, userhome_diskpath, disk->filepath) == NULL);
     if (emu.verbose)
        xprintf("disk_open: Drive %c: %s\n",  disk->drive+'A', disk->filepath);

     if (i)
        {
         disk->error = DISK_ERR_NOTFOUND;
         return -1;
        } 
     
     disk->fdisk = open_file(filename, userhome_diskpath, disk->filepath, "r+b");
     if (disk->fdisk == NULL)
        {
         disk->error = DISK_ERR_READONLY;
         return -1;
        } 

     // make an upper case copy of the file name  
     toupper_string(uname, disk->filepath);     

     // image is write protected if a trailing '_' is found
     disk->wrprot = uname[strlen(uname)-1] == '_';
     if (disk->wrprot)
        uname[strlen(uname)-1] = 0;		//remove the '_'

     // find the disk type
     for (i = type_start; image_types[i][0] != 0; i++)
        {
         if (strstr(uname, image_types[i]) != NULL)
            break;
        }
     if (image_types[i][0] == 0)
        return -1;    

     itype_temp = i + 1;
#ifdef USE_LIBDSK
    }
 else
    {
     disk->wrprot = 0;
     itype_temp = DISK_LIBDSK;
    }    
#endif

 // configure parameters for the disk type 
 switch (itype_temp)
    {
     case DISK_DIP    : // DIP (in house spec)
                        fseek(disk->fdisk, -sizeof(disk->imagerec), SEEK_END);

                        // read into temporary location as endianness needs addressing
                        fread(&temp_imagerec, sizeof(temp_imagerec), 1, disk->fdisk);
                        memcpy(&disk->imagerec, &temp_imagerec, sizeof(disk->imagerec));
                        disk->imagerec.wrprot = leu16_to_host(temp_imagerec.wrprot);
                        disk->imagerec.tracks = leu16_to_host(temp_imagerec.tracks);
                        disk->imagerec.heads = leu16_to_host(temp_imagerec.heads);
                        disk->imagerec.secsize = leu16_to_host(temp_imagerec.secsize);
                        disk->imagerec.sectrack = leu16_to_host(temp_imagerec.sectrack);
                        disk->imagerec.datatrack = leu16_to_host(temp_imagerec.datatrack);
                        disk->imagerec.systsecofs = leu16_to_host(temp_imagerec.systsecofs);
                        disk->imagerec.datasecofs = leu16_to_host(temp_imagerec.datasecofs);
                        disk->imagerec.secterrs = leu16_to_host(temp_imagerec.secterrs);

                        if (strcmp(disk->imagerec.id, "DISK-IMAGE-PLUS") != 0)
                           return -1;
                        if ((disk->imagerec.secsize != 128) && (disk->imagerec.secsize != 256) &&
                           (disk->imagerec.secsize != 512) && (disk->imagerec.secsize != 1024))
                           return -1;
                        if (! disk->wrprot)   // if not already '_' write protected
                           disk->wrprot = disk->imagerec.wrprot;
			// default image to double density, 250kbps data rate
			disk->density = DISK_DENSITY_DOUBLE;
			disk->datarate = DISK_RATE_250KBPS;
                        strcpy(disk->image_name, "DIP");
                        break;
     case DISK_DSK    : // DSK
                        // get the disk information block
                        fread(&dski, sizeof(dski), 1, disk->fdisk);
                        disk->imagerec.tracks = dski.tracks;
                        disk->imagerec.heads = dski.heads; 

                        // get the first track information block to get some initial values
                        fread(&dskt, sizeof(dskt), 1, disk->fdisk);
                        disk->imagerec.sectrack = dskt.spt;
                        disk->imagerec.secsize = 128 << dskt.bps;
                        disk->imagerec.systsecofs = dskt.sect_numb;
			disk->density = DISK_DENSITY_DOUBLE;
			disk->datarate = DISK_RATE_250KBPS;
                        strcpy(disk->image_name, "DSK");
                        break;
     case DISK_IMG    : // IMG (nanowasp v0.22 format)
     case DISK_NW     : // NW (nanowasp v0.22 format)
                        disk->imagerec.tracks = 40;
                        disk->imagerec.heads = 2;
                        disk->imagerec.sectrack = 10;
                        disk->imagerec.secsize = 512;           
                        disk->imagerec.datatrack = 1;
                        disk->imagerec.datasecofs = 1;
                        disk->imagerec.systsecofs = 1;
			disk->density = DISK_DENSITY_DOUBLE;
			disk->datarate = DISK_RATE_250KBPS;
                        strcpy(disk->image_name, "RAW (IMG/NW)");
                        break;
     case DISK_DS40   : // DS40
     case DISK_D40    : // D40
                        disk->imagerec.tracks = 40;
                        disk->imagerec.heads = 2;
                        disk->imagerec.sectrack = 10;
                        disk->imagerec.secsize = 512;           
                        disk->imagerec.datatrack = 1;
                        disk->imagerec.datasecofs = 1;
                        disk->imagerec.systsecofs = 1;
			disk->density = DISK_DENSITY_DOUBLE;
			disk->datarate = DISK_RATE_250KBPS;
                        strcpy(disk->image_name, "RAW");
                        break;
     case DISK_SS80   : // SS80
     case DISK_S80    : // S80     
                        disk->imagerec.tracks = 80;
                        disk->imagerec.heads = 1;
                        disk->imagerec.sectrack = 10;
                        disk->imagerec.secsize = 512;           
                        disk->imagerec.datatrack = 1;
                        disk->imagerec.datasecofs = 1;
                        disk->imagerec.systsecofs = 1;
			disk->density = DISK_DENSITY_DOUBLE;
			disk->datarate = DISK_RATE_250KBPS;
                        strcpy(disk->image_name, "RAW SS80");
                        break;
     case DISK_DS80   : // DS80 (Microbee modular)
     case DISK_D80    : // D80 (Microbee modular)
                        disk->imagerec.tracks = 80;
                        disk->imagerec.heads = 2;
                        disk->imagerec.sectrack = 10;
                        disk->imagerec.secsize = 512;           
                        disk->imagerec.datatrack = 2;         // 4 reserved tracks
                        disk->imagerec.datasecofs = 21;       // sectors are 21 to 30
                        disk->imagerec.systsecofs = 1;
			disk->density = DISK_DENSITY_DOUBLE;
			disk->datarate = DISK_RATE_250KBPS;
                        strcpy(disk->image_name, "RAW DS80");
                        break;
     case DISK_DS82   : // DS82
     case DISK_D82    : // D82
                        disk->imagerec.tracks = 80;
                        disk->imagerec.heads = 2;
                        disk->imagerec.sectrack = 10;
                        disk->imagerec.secsize = 512;           
                        disk->imagerec.datatrack = 1;
                        disk->imagerec.datasecofs = 1;
                        disk->imagerec.systsecofs = 1;
			disk->density = DISK_DENSITY_DOUBLE;
			disk->datarate = DISK_RATE_250KBPS;
                        strcpy(disk->image_name, "RAW DS82");
                        break;
     case DISK_DS84   : // DS84
     case DISK_D84    : // D84
                        disk->imagerec.tracks = 80;
                        disk->imagerec.heads = 2;
                        disk->imagerec.sectrack = 10;
                        disk->imagerec.secsize = 512;           
                        disk->imagerec.datatrack = 1;
                        disk->imagerec.datasecofs = 1;
                        disk->imagerec.systsecofs = 1;
			disk->density = DISK_DENSITY_DOUBLE;
			disk->datarate = DISK_RATE_250KBPS;
                        strcpy(disk->image_name, "RAW DS84");
                        break;
     case DISK_HD0    : // HDD (10MB) 305 cylinders, 4 heads, 17 sect/track
                        disk->imagerec.tracks = 305;
                        disk->imagerec.heads = 4;
                        disk->imagerec.sectrack = 17;
                        disk->imagerec.secsize = 512;           
                        disk->imagerec.datatrack = 1;
                        disk->imagerec.datasecofs = 1;
                        disk->imagerec.systsecofs = 1;
                        strcpy(disk->image_name, "RAW HD0");
                        break;
     case DISK_HD1    : // HDD (10MB) 80 cylinders, 4 heads, 63 sect/track
                        disk->imagerec.tracks = 80;
                        disk->imagerec.heads = 4;
                        disk->imagerec.sectrack = 63;
                        disk->imagerec.secsize = 512;           
                        disk->imagerec.datatrack = 1;
                        disk->imagerec.datasecofs = 1;
                        disk->imagerec.systsecofs = 1;
			// Density and data rate aren't used for a harddisk image
			disk->density = DISK_DENSITY_DOUBLE;
			disk->datarate = DISK_RATE_250KBPS;
                        strcpy(disk->image_name, "RAW HD1");
                        break;
#ifdef USE_LIBDSK
     case DISK_LIBDSK : // LibDsk
                        disk->itype = 0;

                        if (emu.verbose)
                           xprintf("disk_open: Drive %c: %s\n",  disk->drive+'A', filename);

                        // Find the format name
			{
                        int found = 0;
                        fmt = FMT_180K;
                        while ((dg_stdformat(NULL, fmt, &fname, NULL) == DSK_ERR_OK) && (! found))
                           {
                            if (!strcmp(disk->libdsk_format, fname))
                               found = 1;
                            else
                               ++fmt;
                           }

                        if (! found)
                           {   
                            xprintf("disk_open: LibDsk format name %s not recognised.\n", disk->libdsk_format);
                            return -1;
                           }
			}
                        snprintf(disk->image_name, sizeof(disk->image_name), "LIBDSK-%s", disk->libdsk_format);

                        // Initialise disk geometry                                                       
                        dsk_err = dg_stdformat(&disk->dg, fmt, NULL, &fdesc);
                        if (dsk_err != DSK_ERR_OK)
                           {
                            xprintf("disk_open: LibDsk format name %s not recognised.\n", fdesc);                           
                            return -1;
                           }
                           
                        // Allow the Data rate to be changed for DS40 DD media in HD 80T 1.2MB drives.
                        if (disk->dstep_hd)
                           disk->dg.dg_datarate = RATE_DD;
                        
                        disk->imagerec.tracks = disk->dg.dg_cylinders;
                        disk->imagerec.heads = disk->dg.dg_heads;
                        disk->imagerec.sectrack = disk->dg.dg_sectors;
                        disk->imagerec.secsize = disk->dg.dg_secsize;
                        disk->imagerec.datatrack = 1;
                        disk->imagerec.datasecofs = disk->dg.dg_secbase;
                        disk->imagerec.systsecofs = disk->dg.dg_secbase;

			switch (disk->dg.dg_datarate)
                           {
                            case RATE_HD: // Data rate for 1.4Mb 3.5"  in 3.5"  drive
                               disk->datarate = DISK_RATE_500KBPS;
                               break;
                            case RATE_DD: // Data rate for 360k  5.25" in 1.2Mb drive
                            case RATE_SD: // Data rate for 720k  3.5"  in 3.5"  drive
                               disk->datarate = DISK_RATE_250KBPS;
                               break;
                            case RATE_ED:
                               // Data rate for 2.8Mb 3.5"  in 3.5"  drive
                               // we don't support this.
                            default:
                               assert(0);
                               break;
                            }
			disk->density = disk->dg.dg_fm ? DISK_DENSITY_SINGLE : DISK_DENSITY_DOUBLE;

                        // use side1 as 0 if native formatted disks are being used
                        if ((strcmp(disk->libdsk_format, "ds80") == 0) ||
                            (strcmp(disk->libdsk_format, "ds40") == 0))
                            disk->side1as0 = 1;

                        // Open the disk image or floppy
                        test_file(filename, userhome_diskpath, disk->filepath);
                        if (strlen(disk->libdsk_type))
                           dsk_err = dsk_open(&disk->self, disk->filepath, disk->libdsk_type, NULL);
                        else   
                           dsk_err = dsk_open(&disk->self, disk->filepath, NULL, NULL);
                        
                        // use double stepping if --dstep or --dstep-hd was specified
                        if (disk->dstep)
                           dsk_set_option(disk->self, "DOUBLESTEP", 1);

                        if (disk->self)
                           dsk_set_retry(disk->self, 5);
                        else
                           {
                            xprintf("disk_open: dsk_open error: dsk_err=%d\n", dsk_err);
                            disk->itype = 0;
                            return -1;
                           }
                        break;
#endif
    } 

 disk->itype = itype_temp;

#ifdef use_info_disk_open
 if (emu.verbose)
    xprintf("disk_open: %s   Tracks: %d   Heads: %d   S/T: %d   B/S: %d\n", 
    disk->image_name,
    disk->imagerec.tracks, disk->imagerec.heads,
    disk->imagerec.sectrack, disk->imagerec.secsize);
#endif

 return 0;    
}

//==============================================================================
// Disk close.
//
//   pass: disk_t *disk
// return: void
//==============================================================================
void disk_close (disk_t *disk)
{
#ifdef USE_LIBDSK
 if (disk->itype == DISK_LIBDSK)
    dsk_close(&disk->self);
 else
#endif 
    fclose(disk->fdisk);

 disk->fdisk = NULL;
 disk->itype = 0; 
}

int disk_create(const char *name, disk_t *disk)
{
 return -1;
}

//==============================================================================
// Disk read.
//
//   pass: disk_t *disk
//         char *buf
//         int side                     physical side number
//         int idside                   side number in the ID field
//         int track
//         int sect
//         char rtype			m if a multi sector read operation
// return: int				0 if no errors, else error number
//==============================================================================
int disk_read (disk_t *disk, char *buf, int side, int idside, int track, int sect, char rtype)
{
 dskt_t dskt;	// dsk track information block

 int res = 0;

#ifdef USE_LIBDSK
 dsk_err_t dsk_err = 0;
#endif

 int trackofs;
 int sectofs;
 int sectuse;
 int dskofs;

#ifdef use_debug_disk_read_abort
 xprintf("disk_read: forcing an abort here\n");
 assert(1 != 1);
#endif

#ifdef use_info_disk_read
 if (emu.verbose)
    {
     xprintf("Tracks: %d   Heads: %d   S/T: %d   B/S: %d\n", 
     disk->imagerec.tracks, disk->imagerec.heads,
     disk->imagerec.sectrack, disk->imagerec.secsize);
     xprintf("Track: %3d   Side: %3d   Sector: %3d\n",  track, side, sect);
    } 
#endif

 switch (disk->itype)
    {
     case DISK_DIP    : // DIP (in house spec)
                        trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
                        if (track >= disk->imagerec.datatrack)
                           sectuse = sect - disk->imagerec.datasecofs;
                        else
                           sectuse = sect - disk->imagerec.systsecofs;
                        sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize + (sectuse * disk->imagerec.secsize);
#ifdef use_info_disk_read
                        if (emu.verbose)
                           xprintf("Image: %s   Vers: %s   Type: %s\n", disk->imagerec.id,
                           disk->imagerec.ver, disk->imagerec.type);
#endif
                        if ((sectuse >= 0) && (sectuse < disk->imagerec.sectrack))
                           {
                            fseek(disk->fdisk, trackofs + sectofs, SEEK_SET);
                            fread(buf, disk->imagerec.secsize, 1, disk->fdisk);
                           }
                        else
                           res = -1;
                        break;
     case DISK_DSK    : // DSK
                        trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
                        dskofs = (((track * disk->imagerec.heads) + side + 1) * 0x100) + 0x100;

                        // read the track information header for the track
                        sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize;
                        fseek(disk->fdisk, trackofs + sectofs + dskofs - 0x100, SEEK_SET);

                        fread(&dskt, sizeof(dskt), 1, disk->fdisk);
                        disk->imagerec.datasecofs = dskt.sect_numb;
                        disk->imagerec.systsecofs = dskt.sect_numb;
                        sectuse = sect - disk->imagerec.datasecofs;

                        // the sectors per track and size may change on some DSK images
                        disk->imagerec.sectrack = dskt.spt;
                        disk->imagerec.secsize = 128 << dskt.bps;

                        // seek and read the sector
                        if ((sectuse >= 0) && (sectuse < disk->imagerec.sectrack))
                           {
                            sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize + (sectuse * disk->imagerec.secsize);
                            fseek(disk->fdisk, trackofs + sectofs + dskofs, SEEK_SET);
                            fread(buf, disk->imagerec.secsize, 1, disk->fdisk);
                           }
                        else
                           res = -1;
                        break;
     case DISK_IMG    : // IMG (nanowasp v0.22 format)
     case DISK_NW     : // NW (nanowasp v0.22 format)
                        if ((sect > 0) && (sect <= disk->imagerec.sectrack))
                           {
                            fseek(disk->fdisk, ((side * 40 + track) * 10 + (sect - 1)) * 512, SEEK_SET);
                            fread(buf, 512, 1, disk->fdisk);
                           }
                        else
                           res = -1;
                        break;
     case DISK_DS40   : // DS40
     case DISK_D40    : // D40     
     case DISK_SS80   : // SS80
     case DISK_S80    : // S80     
     case DISK_DS80   : // DS80 (Microbee modular)
     case DISK_D80    : // D80 (Microbee modular)     
     case DISK_DS82   : // DS82
     case DISK_D82    : // D82     
     case DISK_DS84   : // DS84
     case DISK_D84    : // D84
     case DISK_HD0    : // HDD (10MB) 305 cylinders, 4 heads, 17 sect/track
     case DISK_HD1    : // HDD (10MB) 80 cylinders, 4 heads, 63 sect/track
                        trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
                        sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize;

                        if (track >= disk->imagerec.datatrack)
                           sectuse = sect - disk->imagerec.datasecofs;
                        else
                           sectuse = sect - disk->imagerec.systsecofs;

                        sectofs += sectuse * disk->imagerec.secsize;

                        if ((sectuse >= 0) && (sectuse < disk->imagerec.sectrack))
                           {
                            fseek(disk->fdisk, trackofs + sectofs, SEEK_SET);
                            fread(buf, disk->imagerec.secsize, 1, disk->fdisk);
                           }
                        else
                           res = -1;
                        break;
#ifdef USE_LIBDSK
     case DISK_LIBDSK : // LibDsk
	                dsk_err = dsk_xread(disk->self, &disk->dg, buf, track, side, track, idside, sect, disk->secsize, NULL);
                        res =  dsk_err;
                        break;
#endif
    }  

 // check if there were any errors and report and exit
 if (res)
    {
#ifdef USE_LIBDSK
     if (disk->itype == DISK_LIBDSK)
        {
         if ((dsk_err != DSK_ERR_NODATA) && (rtype != 'm')) // Don't report multi sector error
            {
             xprintf("disk_read: dsk_xread");
             xprintf(" error: file=%s dsk_err=%d %s\n", disk->filepath, dsk_err, dsk_strerror(dsk_err));
             xprintf("Track: %3d   Side: %3d   Sector: %3d   IDside: %3d\n",  track, side, sect, idside);
            } 
        }       
     else
#endif
        { 
         if (rtype != 'm') // Don't report multi sector error
            {
             xprintf("disk_read: (inbuilt) file=%s res=%d\n", disk->filepath, res);
             xprintf("Track: %3d   Side: %3d   Sector: %3d\n",  track, side, sect);
            } 
        }
     
     return res;
    }
 
 return 0;
}

//==============================================================================
// Disk write.
//
// A flush call is used on each sector write.
//
//   pass: disk_t *disk
//         char *buf
//         int side
//         int track
//         int sect
//         char wtype			m if a multi sector read operation
// return: int				0 if no errors, else error number
//==============================================================================
int disk_write (disk_t *disk, char *buf, int side, int idside, int track, int sect, char wtype)
{
 dskt_t dskt;			// dsk track information block

 int res = 0;

#ifdef USE_LIBDSK
 dsk_err_t dsk_err = 0;
#endif

 int trackofs;
 int sectofs;
 int sectuse; 
 int dskofs;

 // reset the exit seconds counter to a new minimum value every time we write
 // to disk and emu.secs_exit is not zero.
 if ((emu.secs_exit) && ((emu.secs_run + 3) >= emu.secs_exit))
    emu.secs_exit = emu.secs_run + 3;

#ifdef use_info_disk_read
 if (emu.verbose)
    { 
     xprintf("Tracks: %d   Heads: %d   S/T: %d   B/S: %d\n", 
     disk->imagerec.tracks, disk->imagerec.heads,
     disk->imagerec.sectrack, disk->imagerec.secsize);
     xprintf("Track: %3d   Side: %3d   Sector: %3d\n",  track, side, sect);
    } 
#endif

 switch (disk->itype)
    {
     case DISK_DIP    : // DIP (in house spec)
                        trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
                        if (track >= disk->imagerec.datatrack)
                           sectuse = sect - disk->imagerec.datasecofs;
                        else
                           sectuse = sect - disk->imagerec.systsecofs;       
                        sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize + (sectuse * disk->imagerec.secsize);
#ifdef use_info_disk_read
                        if (emu.verbose)
                           xprintf("Image: %s   Vers: %s   Type: %s\n", disk->imagerec.id,
                           disk->imagerec.ver, disk->imagerec.type);
#endif
                        if ((sectuse >= 0) && (sectuse < disk->imagerec.sectrack))
                           {
                            fseek(disk->fdisk, trackofs + sectofs, SEEK_SET);
                            fwrite(buf, disk->imagerec.secsize, 1, disk->fdisk);
                            fflush(disk->fdisk);
                           }
                        else
                           res = -1;
                        break;
     case DISK_DSK    : // DSK
                        trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
                        dskofs = (((track * disk->imagerec.heads) + side + 1) * 0x100) + 0x100;

                        // read the track information header for the track
                        sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize;
                        fseek(disk->fdisk, trackofs + sectofs + dskofs - 0x100, SEEK_SET);

                        fread(&dskt, sizeof(dskt), 1, disk->fdisk);
                        disk->imagerec.datasecofs = dskt.sect_numb;
                        disk->imagerec.systsecofs = dskt.sect_numb;
                        sectuse = sect - disk->imagerec.datasecofs;

                        // the sectors per track and size may change on some DSK images
                        disk->imagerec.sectrack = dskt.spt;
                        disk->imagerec.secsize = 128 << dskt.bps;

                        // seek and write the sector
                        sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize + (sectuse * disk->imagerec.secsize);
                        if ((sectuse >= 0) && (sectuse < disk->imagerec.sectrack))
                           {
                            fseek(disk->fdisk, trackofs + sectofs + dskofs, SEEK_SET);
                            fwrite(buf, disk->imagerec.secsize, 1, disk->fdisk);
                            fflush(disk->fdisk);
                           } 
                        else
                           res = -1;
                        break;
     case DISK_IMG    : // IMG (nanowasp v0.22 format)
     case DISK_NW     : // NW (nanowasp v0.22 format)
                        if ((sect > 0) && (sect <= disk->imagerec.sectrack))
                           {
                            fseek(disk->fdisk, ((side * 40 + track) * 10 + (sect - 1)) * 512, SEEK_SET);
                            fwrite(buf, 512, 1, disk->fdisk);
                            fflush(disk->fdisk);
                           }
                        else
                           res = -1;
                        break;
     case DISK_DS40   : // DS40
     case DISK_D40    : // D40     
     case DISK_SS80   : // SS80
     case DISK_S80    : // S80     
     case DISK_DS80   : // DS80 (Microbee modular)
     case DISK_D80    : // D80 (Microbee modular)     
     case DISK_DS82   : // DS82
     case DISK_D82    : // D82     
     case DISK_DS84   : // DS84
     case DISK_D84    : // D84
     case DISK_HD0    : // HDD (10MB) 305 cylinders, 4 heads, 17 sect/track
     case DISK_HD1    : // HDD (10MB) 80 cylinders, 4 heads, 63 sect/track
                        trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
                        sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize;

                        if (track >= disk->imagerec.datatrack)
                           sectuse = sect - disk->imagerec.datasecofs;
                        else
                           sectuse = sect - disk->imagerec.systsecofs;

                        sectofs += sectuse * disk->imagerec.secsize;

                        if ((sectuse >= 0) && (sectuse < disk->imagerec.sectrack))
                           {
                            fseek(disk->fdisk, trackofs + sectofs, SEEK_SET);
                            fwrite(buf, disk->imagerec.secsize, 1, disk->fdisk);
                            fflush(disk->fdisk);
                           }
                        else
                           res = -1;
                        break;
#ifdef USE_LIBDSK
     case DISK_LIBDSK : // LibDsk
                        dsk_err = dsk_xwrite(disk->self, &disk->dg, buf, track, side, track, idside, sect, disk->secsize, (int)NULL);
                        res = dsk_err;
                        break;
#endif                        
    }

 // check if there were any errors and report and exit
 if (res)
    {
#ifdef USE_LIBDSK
     if (disk->itype == DISK_LIBDSK)
        {
         if ((dsk_err != DSK_ERR_NODATA) && (wtype != 'm')) // Don't report multi sector error
            {
             xprintf("disk_write: dsk_xwrite");
             xprintf(" error: file=%s dsk_err=%d %s\n", disk->filepath, dsk_err, dsk_strerror(dsk_err));
             xprintf("Track: %3d   Side: %3d   Sector: %3d   IDside: %3d\n",  track, side, sect, idside);
            }
        }
     else
#endif
        {
         if (wtype != 'm') // Don't report multi sector error
            {
             xprintf("disk_write: (inbuilt) file=%s res=%d\n", disk->filepath, res);
             xprintf("Track: %3d   Side: %3d   Sector: %3d\n",  track, side, sect);
            }
        }
        
     return res;
    }

 return 0;
}

//==============================================================================
// Disk read ID field.
//
// Fills the read_addr_t structure id with the ID field.  For internal non
// DSK drivers the structure is set to the passed side and track values and
// values from the disk->imagerec structure.
//
// CPCEMU DSK formats can use different size sectors but must be the same for
// the entire track and the total size of the track must be the same for all
// tracks on the disk.  This means 512x10 and 1024x5 tracks work ok.
//
//   pass: disk_t *disk
//         read_addr_t *idfield
//         int side
//         int track
// return: int				0 if no errors, else error number
//==============================================================================
int disk_read_idfield (disk_t *disk, read_addr_t *idfield, int side, int track)
{
 dskt_t dskt;			// dsk track information block
 
 int res = 0; 

#ifdef USE_LIBDSK
 dsk_err_t dsk_err;
 DSK_FORMAT result;
#endif

 int trackofs;
 int sectofs;
 int dskofs;

 switch (disk->itype)
    {
     case DISK_DSK    : // if DSK format (not LibDsk)
                        trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
                        dskofs = (((track * disk->imagerec.heads) + side + 1) * 0x100) + 0x100;

                        // offset calculation when side 1 of disk accessed
                        sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize;

                        // read the track information header for the track
                        fseek(disk->fdisk, trackofs + sectofs + dskofs - 0x100, SEEK_SET);

                        fread(&dskt, sizeof(dskt), 1, disk->fdisk);

                        disk->imagerec.sectrack = dskt.spt;
                        disk->imagerec.secsize = 128 << dskt.bps;

                        idfield->track = dskt.track_numb;
                        idfield->side = dskt.hnumb_sectid;
                        idfield->secaddr = dskt.sect_numb;
                        idfield->seclen = disk->imagerec.secsize >> 8;
                        idfield->crc1 = dskt.state1_errcode;
                        idfield->crc2 = dskt.state2_errcode;
                        break;
#ifdef USE_LIBDSK
     case DISK_LIBDSK : // LibDsk
                        dsk_err = dsk_psecid(disk->self, &disk->dg, track, side, &result);
                        idfield->track = result.fmt_cylinder;
                        idfield->side = result.fmt_head;
                        idfield->seclen = result.fmt_secsize >> 8; 
                        idfield->secaddr = result.fmt_sector;

                        // should update disk->imagerec.sectrack too, but how (but don't need to ??)
                        disk->imagerec.secsize = result.fmt_secsize;

                        if (dsk_err)
                           xprintf("disk_read_idfield: dsk_psecid error: file=%s dsk_err=%d %s\n",
                           disk->filepath, dsk_err, dsk_strerror(dsk_err));
                        idfield->crc1 = 0xff;
                        idfield->crc2 = 0xff;
                        res = dsk_err;
                        break;
#endif
              default : // internal raw and DIP formats
                        idfield->track = track;

                        idfield->side = side;

                        // if a Dreamdisk double sided disk then set the 'double sided' flag
                        // FIXME - this should only apply to real Dreamdisk images.
                        if ((modelx.fdc == MODFDC_DD) && (disk->imagerec.heads == 2))
                           idfield->side |= 0x80;

                        idfield->seclen = disk->imagerec.secsize >> 8; 
                        if (track >= disk->imagerec.datatrack)
                           idfield->secaddr = disk->imagerec.datasecofs;
                        else
                           idfield->secaddr = disk->imagerec.systsecofs;
                        idfield->crc1 = 0xff;
                        idfield->crc2 = 0xff;

                        break;
    }

 return res;
}

//==============================================================================
// Disk write protected query.
//
//   pass: disk_t *disk
// return: int				write protect status
//==============================================================================
int disk_iswrprot (disk_t *disk)
{
#ifdef USE_LIBDSK
 dsk_err_t dsk_err;
 unsigned char result;
#endif

 switch (disk->itype)
    {
#ifdef USE_LIBDSK
     case DISK_LIBDSK : // LibDsk
                        dsk_err = dsk_drive_status(disk->self, &disk->dg, 0, &result);
                        disk->wrprot = (result & DSK_ST3_RO);

                        if (dsk_err)
                           xprintf("disk_iswrprot: dsk_drive_status error: file=%s dsk_err=%d %s\n",
                           disk->filepath, dsk_err, dsk_strerror(dsk_err));
                        break;

                        break;
#endif                        
              default : // internal formats
                        break;               
    }

 return disk->wrprot;
}

//==============================================================================
// Disk set geometry member.
//
//   pass: disk_t *disk
// return: void
//==============================================================================
int disk_setgeom_member (disk_t *disk)
{
 return 0;
}

//==============================================================================
// Disk format track
//
//   pass: disk_t *disk
//         int side
//         char *buf
//         int ddense			MFM=1, FM=0
//         int track
//         int side
//         int sectors
// return: int				0 if no error
//==============================================================================
int disk_format_track (disk_t *disk, char *buf, int ddense, int track, int side, int sectors)
{
 int res;
 
#ifdef USE_LIBDSK
 dsk_err_t dsk_err;
 static DSK_GEOMETRY geom;
 static DSK_FORMAT format[128]; // up to 128 sector headers
 int i;
 int s;
#endif

 // reset the exit seconds counter to a new minimum value every time we write
 // to disk and emu.secs_exit is not zero.
 if ((emu.secs_exit) && ((emu.secs_run + 3) >= emu.secs_exit))
    emu.secs_exit = emu.secs_run + 3;

 switch (disk->itype)
    {
#ifdef USE_LIBDSK
     case DISK_LIBDSK : // LibDsk
                        memcpy(&geom, &disk->dg, sizeof(geom));
                        geom.dg_fm = (ddense == 0);
                        i = 0;
                        for (s = 0; s < sectors; s++)
                           {
                            format[s].fmt_cylinder = buf[i++];
                            format[s].fmt_head = buf[i++];
                            format[s].fmt_sector = buf[i++];
                            format[s].fmt_secsize = buf[i++] << 8;
                           }
                        dsk_err = dsk_pformat(disk->self, &geom, track, side, format, 0xe5);
                        
                        if (dsk_err)
                           xprintf("disk_format_track: dsk_pformat error: file=%s dsk_err=%d %s\n",
                           disk->filepath, dsk_err, dsk_strerror(dsk_err));
                        res = dsk_err;
                        break;
#endif                        
              default : // internal formats
                        res = 0;
                        break;
    }

 return res;                       
}
