//******************************************************************************
//*                                  uBee512                                   *
//*       An emulator for the Microbee Z80 ROM, FDD and HDD based models       *
//*                                                                            *
//*                               Sound module                                 *
//*                                                                            *
//*                   Copyright (C) 2007-2009 Stewart Kay                      *
//******************************************************************************
//
// This module is used to emulate the internal speaker using SDL.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v4.1.0 - 11 June 2009, S.J.Kay
// - Added sound_get_buffer_count() function to return current buffer count.
//
// v4.0.0 - 13 May 2009, S.J.Kay
// - Changes made to incorporate a Z80 API.  Specific Z80 code used here
//   has been replaced with code usable by any Z80 emulator.
// - Changed SOUND_HOLDOFF defaults from 50 to 75 mS.
//
// v3.1.0 - 22 April 2009, S.J.Kay
// - Removed all occurrences of console_output() function calls.
// - Changed all printf() calls to use a local xprintf() function.
//
// v3.0.0 - 5 October 2008, S.J.Kay
// - Simplified the timer by introducing a time_get_ms() function.
// - Changed SOUND_LOWESTFREQ to 20Hz to improve re-sync on quiet periods.
// - Removed SDL_LockAudio() and SDL_UnlockAudio() calls and sound_unsilence()
//   function.  Reworked the code for sound improvments concerning the holdoff.
//
// v2.8.0 - 1 September 2008, S.J.Kay
// - Added sound_volume() function to set the audio volume from an integer
//   percentage value. 
// - Changed documentation references from XTAL to CPU clock.
//
// v2.7.0 - 15 June 2008, S.J.Kay
// - Changes as many values are now definable as an option.
// - Added an optional algorithm to keep sound timing tracking better.
// - Added cycles_adj variable to help keep better timing and prevent audio
//   break up caused by audio finishing too early.
// - Added sound_update() function, moved code over from the ubee512.c module.
// - Number of sound samples for refreshing SDL buffer is now configurable.
// - Added a sound frequency +/- adjustment value.
// - Added a mute member to sound_t structure to mute the sound.
// - Added structures emu_t, sound_t and moved variables into it.
// - Moved the defines to sound.h.
//
// v1.4.0 - 26 September 2007, S.J.Kay
// - Removed sound on/off conditional testing,  all done in pio.c module now.
// - Added changes to error reporting.
//
// v1.1.0 - 13 August 2007, S.J.Kay
// - Changes to the volume level to allowing level to be set by user.
// - Added sound_clock function that determines the sound rate playback rate
//   depending on the XTAL frequency and the method chosen.
//
// v1.0.0 - 14 July 2007, S.J.Kay
// - Created a new file to implement the sound emulation.

//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 ROM, FDD and HDD based models. 
 *  Copyright (C) 2007-2009 Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#ifdef MINGW
#include <windows.h>
#else
#include <sys/stat.h>
#include <signal.h>		// signal name macros, and the signal() prototype
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <SDL.h>

#include "sound.h"
#include "z80api.h"
#include "function.h"
#include "gui.h"
#include "ubee512.h"

//==============================================================================
// structures and variables
//==============================================================================
sound_t sound =
{
 .enabled = 1,
 .mute = 0,
 .volume = SOUND_VOLUME_PERCENT,
 .samples = SOUND_SAMPLES,
 .holdoff = SOUND_HOLDOFF,
 .frequency = SOUND_FREQUENCY,
 .freq_adjust = SOUND_FREQADJUST,
 .freq_lowest = SOUND_LOWESTFREQ
};

static int audio_volume;
static int audio_count;
static int audio_outpos;
static int audio_inpos;
static int audio_divval;
static int audio_off_cmp;

static uint8_t speaker;

static uint64_t cycles_now;
static uint64_t cycles_before;
static uint64_t cycles_elapsed;
static int cycles_adj;

static uint8_t sound_before;

static uint64_t ticks_s;	// for sound hold-off period

// the sound buffer time capacity in seconds is equal to the size of the
// buffer divided by the SOUND_FREQUENCY value. A 500K buffer, and an
// SOUND_FREQUENCY value of 44100 is 11.60 seconds (t=512000/44100)
static SOUND_BUFTYP sndbuff[1024 * SOUND_BUFSIZE_KB];

static SDL_AudioSpec wanted;

static void fill_audio(void *udata, Uint8 *stream, int len);

extern emu_t emu;
extern gui_t gui;
extern gui_status_t gui_status;

//==============================================================================
// Sound volume set.  Set the volume level from an integer percentage value.
//
//   pass: int vol			volume percentage (0-100%) integer
// return: void
//==============================================================================
void sound_volume (int vol)
{
 audio_volume = (int)(128.0 * (vol/100.0));
 if (audio_volume > 127)
    audio_volume = 127;
}

//==============================================================================
// Sound Initialise
//
//   pass: void
// return: int			0 if success, -1 if error
//==============================================================================
int sound_init (void)
{
// set the audio format desired
 wanted.freq = sound.frequency + (int)((float)sound.frequency * (sound.freq_adjust / 100.0));
 wanted.format = SOUND_FORMAT;
 wanted.channels = SOUND_CHANNELS;
 wanted.samples = sound.samples;
 wanted.callback = fill_audio;
 wanted.userdata = NULL;

// set the sound volume
 sound_volume(sound.volume);

// open the audio device, forcing the desired format
 if (SDL_OpenAudio(&wanted, NULL) < 0)
    {
     xprintf("sound_init: Couldn't open audio: %s\n", SDL_GetError());
     return -1;
    }

 return 0;
}

//==============================================================================
// The audio function callback function declared in for wanted.callback
//
//   pass: void *udata
//         Uint8 *stream	A pointer to the audio buffer to be filled
//         int len		The length (in bytes) of the audio buffer
// return: void
//==============================================================================
static void fill_audio (void *udata, Uint8 *stream, int len)
{
 int count;
 int audio_len1, audio_len2;
 int audio_outnext;

// xprintf("audio_outpos=%d   audio_count=%d   len=%d\n", audio_outpos, audio_count, len);        

 if (audio_count)
    {
     if (len > audio_count)
        count = audio_count;
     else
        count = len;
        
     if ((count + audio_outpos) <= sizeof(sndbuff))
        {
         audio_len1 = count;
         audio_len2 = 0;
         audio_outnext = audio_outpos + audio_len1;
         if (audio_outnext >= sizeof(sndbuff))
            audio_outnext = 0;
        }
     else   
        {
         audio_len1 = sizeof(sndbuff) - audio_outpos; 
         audio_len2 = count - audio_len1;
         audio_outnext = audio_len2;
        }

// xprintf("audio_outpos=%d   audio_len1=%d  audio_len2=%d\n", audio_outpos, audio_len1, audio_len2);        

// mix in the audio data, we must do this for audio_len1 and then audio_len2
// if the buffer has wrapped around.
     if (audio_len1)
        {
         SDL_MixAudio(stream, sndbuff + audio_outpos, audio_len1, SDL_MIX_MAXVOLUME);
         if (audio_len2)
             SDL_MixAudio(stream + audio_len1, sndbuff, audio_len2, SDL_MIX_MAXVOLUME);
         audio_count -= count;
         audio_outpos = audio_outnext;
        }

// xprintf("len=%d   audio_count=%d\n", len, audio_count);
    }
 else
    SDL_PauseAudio(1);    // pause audio
}

//==============================================================================
// Sound de-initialise
//
//   pass: void
// return: int				0
//==============================================================================
int sound_deinit (void)
{
 SDL_CloseAudio();

 return 0;
}

//==============================================================================
// Sound reset
//
//   pass: void
// return: int				sound init result
//==============================================================================
int sound_reset (void)
{
 SDL_CloseAudio();

 audio_count = 0;
 audio_outpos = 0;
 audio_inpos = 0;

 speaker = 0;
 cycles_now = 0;
 cycles_before = 0;
 cycles_elapsed = 0;
 cycles_adj = 0;
 
 sound_before = 0;

 sound.dobuffering = 0;

 return sound_init();
}

//==============================================================================
// Sound write
//
//   pass: uint8_t data
// return: void
//==============================================================================
void sound_w (uint8_t data)
{
 int count;
 int audio_len1;
 int audio_len2;

 uint8_t sound_now;

// mask of the speaker bit
 sound_now = data & 0x40;

// only do something if the speaker state changes.
 if ((sound_now != sound_before) && (! sound.mute))
    {
     sound_before = sound_now;

// calculate the elapsed time since the last change of the speaker
// output,  this will be used to determine the amount of data to be
// placed into the buffer.
     cycles_now = z80api_get_tstates();
     cycles_elapsed = cycles_now - cycles_before;
     cycles_before = cycles_now;

// use a constant value here, determines the number of repeat values to be
// placed into the buffer at the desired SOUND_FREQUENCY chosen. It is based
// on 1 channel of 8 bit sound samples.
     count = cycles_elapsed / audio_divval;

// accumulate the left over cycles to be used later to keep timing correct,
// this code works to keep things on track better but constant tones don't
// sound very good.
     cycles_adj += (cycles_elapsed % audio_divval);
     if ((cycles_adj / audio_divval) && (sound.use_alg1))
        {
         count++;
         cycles_adj -= audio_divval;
        }

// pause the audio (no call backs) while we work on the buffer.
     SDL_PauseAudio(1);    // pause audio

// this value determines when a change in speaker state with a long lead up
// delay should be ignored. The value should be kept high enough for low
// frequency sound but low enough to allow the audio to re-sync to the Z80
// CPU speed.  The time is: i.e t=value/44100 (44100 being the
// SOUND_FREQUENCY rate).
//
// for the sound to be truely emulated would require long time periods of no
// speaker change to also be put out, but this would mean re-syncing will
// not work as intended and a lot of data would also need to be filled in. 
// Quiet periods longer than 'audio_off_cmp' will therefore rely on the Z80
// emulation speed.
     if (count > audio_off_cmp)
        {
         if (audio_count)  // if something is already in the sound buffer
            count = audio_off_cmp;
         else              // else skip the lead time
            count = 1;
        }    

// enable buffering if no current audio and not already in buffer mode.
     if ((! audio_count) && (! sound.dobuffering))
        sound.dobuffering = 1;

// truncate the number of fill values if not enough room to fit in the buffer
     if ((audio_count + count) > sizeof(sndbuff))
        count = sizeof(sndbuff) - audio_count;

// calculate the buffer fill quantity,  as the buffer is circular in nature
// values are calulated to fill from the current input position and from the
// start of the buffer when wrapping occurs.
     if ((count + audio_inpos) > sizeof(sndbuff))
        {
         audio_len1 = sizeof(sndbuff) - audio_inpos;
         audio_len2 = count - audio_len1;
        }
     else
        {
         audio_len1 = count;
         audio_len2 = 0;
        } 

// get the actual fill value to be placed into the buffer.  The value used is
// the speaker value on the previous occasion but it probably makes
// no difference anyway.
     if (sound_before)
        speaker = 128 + audio_volume;
     else
        speaker = 128 - audio_volume;

// fill the buffer with the speaker value, audio_len1 is the fill length
// for the current audio_inpos and audio_len2 is the fill length from the
// start of the buffer when wrapping occurs.
     if (audio_len1)
        {
         memset(sndbuff + audio_inpos, speaker, audio_len1);
         audio_inpos += audio_len1;
         if (audio_len2)
            {
             memset(sndbuff, speaker, audio_len2);
             audio_inpos = audio_len2;
            } 
        }

// ensure that the audio input position is in kept inbounds
     if ((audio_inpos) >= sizeof(sndbuff))
        audio_inpos = 0;

// adjust the count of samples now in the buffer ready to be played
     audio_count += count;

// every time a re-syncing period occurs a small delay period is enabled
// before sound starts to play.  This is required when running with a real
// Microbee CPU speed as the sound out is trying to play at the real rate
// too but race conditions means that audio will sometimes not be ready on
// the next call back and so the sound will be become jerky and distorted.
     if (! sound.dobuffering)
        SDL_PauseAudio(0);    // un-pause audio
    }
}

//==============================================================================
// Sound update.
//
// Sound buffering.  This is required as CPU emulation speed is affected by
// other processes occuring at the same time, we need to try to ensure the
// sound is slightly in front so that it does not matter if the Z80 lags
// behind slightly in creating the sound.  A 100mS delay has been found to
// be about right.  If the buffer delay is too long the lag will be
// noticable by the user.  If too short then break up, and crackling will be
// evident.
//
//   pass: void
// return: void
//==============================================================================
void sound_update (void)
{
 if (sound.dobuffering)
    {
     if (sound.dobuffering == 1)   // if just starting the buffering
        {
         sound.dobuffering++;
         ticks_s = time_get_ms();
        } 
     else
        {
         if ((time_get_ms() - ticks_s) >= sound.holdoff)
            {
             sound.dobuffering = 0;
             SDL_PauseAudio(0);    // un-pause audio
            } 
        }
    }
}

//==============================================================================
// Sound clock values - set values determined by the CPU clock frequency.
// Microbee application software was designed for 2 and 3.75 MHz CPU clock
// rates so if these are required then method should be 0. (proportional)
//
// If the method is 1 then playback will always be at the rate expected for a
// 3.375 MHz CPU clock and software regardless of the CPU clock frequency.
//
// The audio_off_cmp value is calculated for the lowest audio frequency (f)
// allowed: i.e 4410 = 10.0Hz @ 3.375 MHz & 44050KHz sampling rate.
//
// audio_off_cmp = (1/f) * sound.frequency
//
//   pass: int cpuclock		CPU clock frequency in Hz.
//         int method		0=proportional sound rate determined by CPU
//                              clock.
//                              1=sound rate always as if 3.375 MHz CPU clock
//                              used.
// return: void
//==============================================================================
void sound_clock (int cpuclock, int method)
{
 if (method != SOUND_PROP)
    cpuclock = 3375000;

 audio_divval = cpuclock / sound.frequency;
 audio_off_cmp = (int)((1.0 / sound.freq_lowest) * sound.frequency);
}

//==============================================================================
// Get the count of bytes currently in the sound buffer;
//
//   pass: void
// return: int
//==============================================================================
int sound_get_buffer_count (void)
{
 return audio_count;
}

//==============================================================================
// sound commands
//
//   pass: int cmd			sound command
// return: void
//==============================================================================
void sound_command (int cmd)
{
 switch (cmd)
    {
     case      EMU_CMD_MUTE : sound.mute = ! sound.mute;
                              break;
     case   EMU_CMD_VOLUMEI : if ((gui_status.vol) || (gui.persist_flags & GUI_PERSIST_VOL))
                                 {
                                  sound.volume += EMU_VOLUME_CHANGE;
                                  if (sound.volume > 100)
                                     sound.volume = 100;
                                  sound_volume(sound.volume);
                                 }
                              if (! gui_status.vol)
                                 gui_status_set_persist(GUI_PERSIST_VOL, 0);
                              break;
     case   EMU_CMD_VOLUMED : if ((gui_status.vol) || (gui.persist_flags & GUI_PERSIST_VOL))
                                 {
                                  sound.volume -= EMU_VOLUME_CHANGE;
                                  if (sound.volume < 0)
                                     sound.volume = 0;
                                  sound_volume(sound.volume);
                                 }    
                              if (! gui_status.vol)
                                 gui_status_set_persist(GUI_PERSIST_VOL, 0);
                              break;
    }
}
