/* ubee512 Header */

#ifndef HEADER_UBEE512_H
#define HEADER_UBEE512_H

#include <SDL.h>
#include "console.h"
#include "log.h"

#ifdef MINGW
#define SLASHCHAR '\\'
#define SLASHCHAR_STR "\\"
#define SLASHCHAR_OTHER '/'
#define DIR_CONF "\\"
#define DIR_DOCS "\\doc\\"
#define DIR_DISKS "\\disks\\"
#define DIR_ROMS "\\roms\\"
#define DIR_TAPES "\\tapes\\"
#define DIR_PRINTER "\\printer\\"
#define DIR_IMAGES "\\images\\"
#define DIR_RTC "\\rtc\\"
#define DIR_TOOLS "\\tools\\"
#define DIR_FILES "\\files\\"
#define DIR_SHARE "\\share\\"
#else
#define SLASHCHAR '/'
#define SLASHCHAR_STR "/"
#define SLASHCHAR_OTHER '\\'
#define DIR_CONF "/"
#define DIR_DOCS "/doc/"
#define DIR_DISKS "/disks/"
#define DIR_ROMS "/roms/"
#define DIR_TAPES "/tapes/"
#define DIR_PRINTER "/printer/"
#define DIR_IMAGES "/images/"
#define DIR_RTC "/rtc/"
#define DIR_TOOLS "/tools/"
#define DIR_FILES "/files/"
#define DIR_SHARE "/share/"
#endif

#define SSIZE1 512

//==============================================================================
// Model property definitions
//==============================================================================
#define MODROM  1
#define MODCOL1 1
#define MODCOL2 2
#define MODSPD  1

#define MODPB7_PUP 0
#define MODPB7_VS 1
#define MODPB7_RTC 2
#define MODPB7_NET 3

#define MODRTC 1

#define MODFDC_AT 1		// Applied Technology floppy controller
#define MODFDC_DD 2		// DreamDisk floppy controller

//==============================================================================
// Compile time options,  these are normally only changed for testing purposes.
//==============================================================================
#ifdef MINGW
#define FRAMERATE 50		// target frame rate (frames per second) wanted
#define EMU_MAXLAG_MS 250	// maximum time (in mS) Z80 CPU is allowed lag
#define EMU_Z80_DIVIDER 25	// Z80 blocks executed in 1 Z80 emulation frame
#else
#define FRAMERATE 50		// target frame rate (frames per second) wanted
#define EMU_MAXLAG_MS 250	// maximum time (in mS) Z80 CPU is allowed lag
#define EMU_Z80_DIVIDER 25	// Z80 blocks executed in 1 Z80 emulation frame
#endif

// default host conversion of path slash characters
#define EMU_SLASHCONV 1

// paths to shared directories.
#define PATH_SHARED_IMAGES "/usr/local/share/ubee512/images/"
#define PATH_SHARED_DOCS "/usr/local/share/ubee512/doc/"
#define PATH_SHARED_TOOLS "/usr/local/share/ubee512/tools/"
#define PATH_SHARED_DISKS "/usr/local/share/ubee512/disks/"
#define PATH_SHARED_CONFIG "/usr/local/share/ubee512/config/"

// alias configuration files
#define ALIASES_ROMS "roms.alias"
#define ALIASES_DISKS "disks.alias"

// default fall-back boot image
#define BOOT_IMAGE "boot.dsk"

// default model to emulate
#define MOD_DEFAULT MOD_P512K

// define this for the default CPU clock frequency (in MHz)
#define	CPU_CLOCK_FREQ 3.375

// define where the logged data will go
#define LOGFILE	"ubee512_log.txt"

// define the following debug options to report on ports on stdout
#define use_resetinfo
//#define use_debug_disk_read_abort

//#define use_debug_getkeystate_shift
//#define use_debug_getkeystate_lower

// define the following options to be reported to stdout
#define use_info_disk_open
//#define use_info_disk_read

// define this to use faster memory reads but slower MMU configuring. The
// reverse is true if not defined (default is to use it)
#define MEMMAP_HANDLER_1

//==============================================================================
// emulator commands activated from the keyboard or joystick
//==============================================================================
#define EMU_CMD_FULLSCR    0
#define EMU_CMD_TAPEREW    1
#define EMU_CMD_DUMP       2
#define EMU_CMD_DUMP_N1    3
#define EMU_CMD_DUMP_N2    4
#define EMU_CMD_DUMP_B1    5
#define EMU_CMD_DUMP_B2    6
#define EMU_CMD_DUMP_REP   7
#define EMU_CMD_DUMPREGS   8
#define EMU_CMD_DBGOFF     9
#define EMU_CMD_DBGON      10
#define EMU_CMD_DBGTRACE   11
#define EMU_CMD_DBGSTEP01  12
#define EMU_CMD_DBGSTEP10  13
#define EMU_CMD_DBGSTEP20  14
#define EMU_CMD_JOYSTICK   15
#define EMU_CMD_MUTE       16
#define EMU_CMD_PAUSE      17
#define EMU_CMD_VOLUMEI    18
#define EMU_CMD_VOLUMED    19
#define EMU_CMD_SCREENI    20
#define EMU_CMD_SCREEND    21
#define EMU_CMD_VIDSIZE1   22
#define EMU_CMD_GL_FILTER  23
#define EMU_CMD_MWHEEL     24
#define EMU_CMD_MOUSE      25
#define EMU_CMD_CONSOLE    26

#define EMU_VOLUME_CHANGE  1.0
#define EMU_REPEAT1        500
#define EMU_REPEAT2        50

#define EMU_EMU_CONTEXT    0
#define EMU_OSD_CONTEXT    1

//==============================================================================
// hardware definitions (up to 32 hardware bits)
//==============================================================================
#define HW_WD2793          0x00000001

int set_account_paths (void);
void write_id_file (void);
void event_handler (void);
void set_clock_speed (float clock, int blocks, int frate);

enum
   {
    MOD_256TC,   
    MOD_P512K,
    MOD_512K,
    MOD_P128K,
    MOD_128K,
    MOD_P256K,
    MOD_256K,
    MOD_P64K,
    MOD_64K,
    MOD_56K,
    MOD_TTERM,
    MOD_PPC85,    
    MOD_PC85B,
    MOD_PC85,
    MOD_PC,
    MOD_IC,
    MOD_2MHZ,
    MOD_2MHZDD,                 // dreamdisk 2 MHz
    MOD_DD,                     // dreamdisk 3.375 MHz
    MOD_SCF,
    MOD_PCF,
    MOD_TOTAL
   };

typedef struct emu_t
   {
    int done;
    int runmode;
    int model;
    int turbo;
    uint64_t z80_cycles;
    int z80_blocks;		// working number of Z80 blocks
    int z80_ratio;		// current z80 execution ratio (default = 1)
    int z80_divider;
    int maxcpulag;
    int cpuclock;
    int framerate;
    int keyesc;
    int keym;
    int display_context;
    int osd_focus;
    int install_files_req;
    int paused;
    int quit;
    int reset;
    int slashconv;
    int alias_roms;
    int alias_disks;
    int exit_check;
    int win32_lock_key_fix;
    int x11_lock_key_fix;
    int cmd_repeat1;
    int cmd_repeat2;
    int home_account_set;
    int secs_init;
    int secs_run;
    int secs_exit;
    int century;
    int hardware;
    int verbose;
    int cfmode;
    int port50h;
    int port51h;
    int port58h;
    int proc_delay_type;
    float cpuclock_def;
    char sysname[SSIZE1];
    char prefix_path[SSIZE1];
    SDL_Event event;
   }emu_t; 

typedef struct model_t
   {
    int alphap;
    int tckeys;
    int rom;
    int ide;
    int hdd;
    int bootaddr;
    int fdc;
    int ram;
    int pcg;
    int vdu;
    int colour;
    int hwflash;
    int halfint;
    int lpen;
    int speed;
    int piob7;
    int rtc;
    float cpuclock;
   }model_t; 

typedef struct model_custom_t
   {
    char charrom[512];
    char rom1[512];
    char rom2[512];
    char rom3[512];
    char rom256k[512];
    char pak[8][512];
    char netrom[512];
    char basica[512];
    char basicb[512];
    char basicc[512];
    char basicd[512];
    char colprom[512];
    char systname[512];
    int pakram[8];
    int basram;
    int netram;
    int paksel;
   }model_custom_t; 

typedef struct messages_t
   {
    int opengl_no;
   }messages_t;

//==============================================================================
// Debugging related
//==============================================================================
typedef struct modio_t
   {
    FILE *log;
    int level;
   
    int clock;    
    int crtc;
    int fdc;
    int fdc_wtd;
    int fdc_wth;    
    int func;
    int ide;
    int hdd;
    int joystick;
    int keystd;
    int keytc;    
    int mem;
    int options;
    int roms;
    int pioa;
    int piob;
    int piocont;
    int raminit;
    int rtc;
    int ubee512;
    int vdu;
    int vdumem;
    int video;
    int z80;
   }modio_t;

typedef struct regdump_t
   {
    int clock;
    int crtc;
    int fdc;
    int keytc;    
    int mem;
    int roms;
    int pio;
    int rtc;
    int vdu;
    int z80;
   }regdump_t;

typedef struct init_func_t
   {
    int (*memory_call_init)(void);
    int (*memory_call_deinit)(void);
    int (*memory_call_reset)(void);
    char func_name[20];
   }init_func_t;

#endif /* HEADER_UBEE512_H */
