//******************************************************************************
//*                                  uBee512                                   *
//*       An emulator for the Microbee Z80 ROM, FDD and HDD based models       *
//*                                                                            *
//*                             z80ex API module                               *
//*                                                                            *
//*                   Copyright (C) 2007-2009 Stewart Kay                      *
//******************************************************************************
//
// Provides an API for z80ex Z80 emulator.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v4.3.0 - 5 August 2009, S.J.Kay
// - Added functions z80api_register_action(), z80api_deregister_action() and
//   z80api_call_actions() needed for Dreamdisk model. (workerbee)
// - z80api_nonmaskable_intr() implemented for Dreamdisk model. (workerbee)
// - Changed z80api_maskable_intr() to accumulate tstates in emu.z80_cycles.
//
// v4.2.0 - 17 July 2009, S.J.Kay
// - Changes to z80api_execute_complete() to include pio polling calls after
//   every instruction.
//
// v4.1.0 - 27 June 2009, S.J.Kay
// - Fixed z80api_execute_complete() function as was not completing on all
//   dd/fd/cb/ed prefixed opcodes.
//
// v4.0.0 - 15 May 2009, S.J.Kay
// - Created new z80ex API file.

//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 ROM, FDD and HDD based models. 
 *  Copyright (C) 2007-2009 Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <z80ex/z80ex.h>

#include "z80api.h"
#include "z80.h"
#include "ubee512.h"
#include "pio.h"

#define NUM_Z80_ACTIONS 10

static Z80EX_CONTEXT *z80;

static int exec_tstates;
static int poll_want_tstates;
static int poll_want_tstates_def;
static int poll_wait_tstates;
static int poll_repeats;

static int intr_vector;

static int z80_action_count;

static struct
{
 z80_event_t when;
 z80api_action_fn_t function;
}z80actions[NUM_Z80_ACTIONS];

extern struct z80_memory_read_byte z80_mem_r[];
extern struct z80_memory_write_byte z80_mem_w[];

extern uint16_t (*z80_ports_r[])(uint16_t, struct z80_port_read *);
extern void (*z80_ports_w[])(uint16_t, uint8_t, struct z80_port_write *);

extern emu_t emu;
extern model_t modelx;
extern modio_t modio;

Z80EX_BYTE read_mem_cb (Z80EX_CONTEXT *cpu, Z80EX_WORD addr, int m1_state, void *user_data);
void write_mem_cb (Z80EX_CONTEXT *cpu, Z80EX_WORD addr, Z80EX_BYTE value, void *user_data);
Z80EX_BYTE read_port_cb (Z80EX_CONTEXT *cpu, Z80EX_WORD port, void *user_data);
void write_port_cb (Z80EX_CONTEXT *cpu, Z80EX_WORD port, Z80EX_BYTE value, void *user_data);
Z80EX_BYTE read_interrupt_vector_cb (Z80EX_CONTEXT *cpu, void *user_data);

//==============================================================================
// Z80 Initilization
//
//   pass: void
// return: int				0
//==============================================================================
int z80api_init (void)
{
 // create and initialise CPU
 z80 = z80ex_create(read_mem_cb, NULL, write_mem_cb, NULL, read_port_cb, NULL,
 write_port_cb, NULL, read_interrupt_vector_cb, NULL);
 z80_action_count = 0;

 return 0;
}

//==============================================================================
// Z80 De-Initilization
//
//   pass: void
// return: int				0
//==============================================================================
int z80api_deinit (void)
{
 z80ex_destroy(z80);
 return 0;
}

//==============================================================================
// Z80 Reset
//
// poll_want_tstates_def needs to be set to a reasonably low value of ~300
// to allow SW serial to work for Microbee Telcom v2.0 and Telcom ROMs pre
// v3.2.1.
//
//   pass: void
// return: int				0
//==============================================================================
int z80api_reset (void)
{
 z80ex_reset(z80);

 // set the Z80 PC execution address
 z80ex_set_reg(z80, regPC, modelx.bootaddr);

 exec_tstates = 0;

 poll_want_tstates_def = 300;

 poll_want_tstates = poll_want_tstates_def;
 poll_repeats = 0;
 
 return 0;
}

//==============================================================================
// Get Z80 Program Counter (PC)
//
//   pass: void
// return: int				PC
//==============================================================================
int z80api_getpc (void)
{
 return z80ex_get_reg(z80, regPC);
}

//==============================================================================
// Set default PIO polling rate.
//
// Set Z80 tstates between between each PIO polling.  This value is used by
// default and when the poll_repeats counter reaches 0.
//
//   pass: int tstates                  tstates between each PIO poll
// return: void
//==============================================================================
void z80api_set_poll_tstates_def (int tstates)
{
 poll_want_tstates_def = tstates;
} 

//==============================================================================
// Set PIO polling rate and repeat count.
//
// Set Z80 tstates between between each PIO polling and the number of times
// to repeat before the default tstate count is used.
//
//   pass: int tstates                  tstates between each PIO poll
//         int repeats                  how many times to repeat until reset
// return: void
//==============================================================================
void z80api_set_poll_tstates (int tstates, int repeats)
{
 poll_want_tstates = tstates;
 poll_wait_tstates = 0;
 poll_repeats = repeats;
} 

//==============================================================================
// Register an action to occur on a Z80 state change.
//
//   pass: z80_event_t when, pointer to action function
// return: void
//==============================================================================
void z80api_register_action (z80_event_t when, z80api_action_fn_t function)
{
 assert(z80_action_count >= 0 && z80_action_count < NUM_Z80_ACTIONS);
 z80actions[z80_action_count].when = when;
 z80actions[z80_action_count].function = function;
 z80_action_count++;
}

//==============================================================================
// De-register an action to occur on a Z80 state change.
//
//   pass: z80_event_t when, pointer to action function
// return: void
//==============================================================================
void z80api_deregister_action (z80_event_t when, z80api_action_fn_t function)
{
 int i;
 
 for (i = 0; i < z80_action_count; ++i)
    {
     if (z80actions[i].function == function && z80actions[i].when == when)
        {
         memmove(&z80actions[i], &z80actions[i + 1],
         sizeof(z80actions[0]) * (z80_action_count - i - 1));
         z80_action_count--;
         break;
        }
    }
}

//==============================================================================
// Call actions to occur on a Z80 state change.
//
//   pass: z80_event_t when
// return: void
//==============================================================================
void z80api_call_actions (z80_event_t when)
{
 int i;

 for (i = 0; i < z80_action_count; ++i)
 if (z80actions[i].when == when)
    (*z80actions[i].function)();
}

//==============================================================================
// Execute Z80 tstates.
//
//   pass: int tstates
// return: void
//==============================================================================
void z80api_execute (int tstates)
{
 int ts;
 exec_tstates = 0;

 while (exec_tstates < tstates)
    {
     ts = z80ex_step(z80);    
     exec_tstates += ts;

     if (z80ex_doing_halt(z80))
	 z80api_call_actions(Z80_HALT);

     // FIXME - can the actions() mechanism be generalised to
     //  encompass this sort of periodic callback?

     poll_wait_tstates -= ts;

     if (poll_wait_tstates < 1)
        {
         pio_polling();
         if (poll_repeats)
            poll_repeats--;
         else   
            poll_want_tstates = poll_want_tstates_def;
         poll_wait_tstates = poll_want_tstates;
        } 
    }

 emu.z80_cycles += exec_tstates;
 exec_tstates = 0;
}

//==============================================================================
// Execute a single instruction until completed.
//
// Executes dd/fd/cb/ed prefixed opcodes until completed. This function is
// intended for debug stepping. The method employed here is specific to the
// Z80 emulator in use.
//
//   pass: void
// return: void
//==============================================================================
void z80api_execute_complete (void)
{
 z80api_execute(1);
 pio_polling();
 while (z80ex_last_op_type(z80) != 0)
    {
     z80api_execute(1);
     pio_polling();
    } 
} 

//==============================================================================
// Return current Z80 tstate count.
//
//   pass: void
// return: uint64_t
//==============================================================================
uint64_t z80api_get_tstates (void)
{
 return emu.z80_cycles + exec_tstates;
}

//==============================================================================
// Z80 Non Maskable interrupt.
//
//   pass: void
// return: void
//==============================================================================
void z80api_nonmaskable_intr (void)
{
 emu.z80_cycles += z80ex_nmi(z80);
}

//==============================================================================
// Z80 Maskable interrupt.
//
//   pass: int vector
// return: void
//==============================================================================
void z80api_maskable_intr (int vector)
{
 if (z80ex_int_possible(z80) == 1)
    {
     intr_vector = vector;
     emu.z80_cycles += z80ex_int(z80);
    } 
}

//==============================================================================
// Return 1 if maskable interrupts are possible in current Z80 state.
//
//   pass: void
// return: int
//==============================================================================
int z80api_intr_possible (void)
{
 return z80ex_int_possible(z80);
} 

//==============================================================================
// Return all Z80 registers (not called during execution of an instruction)
//
//   pass: z80regs_t *z80regs
// return: void
//==============================================================================
void z80api_get_regs (z80regs_t *z80regs)
{
 z80regs->af = z80ex_get_reg(z80, regAF);
 z80regs->bc = z80ex_get_reg(z80, regBC);
 z80regs->de = z80ex_get_reg(z80, regDE);
 z80regs->hl = z80ex_get_reg(z80, regHL);

 z80regs->af_p = z80ex_get_reg(z80, regAF_);
 z80regs->bc_p = z80ex_get_reg(z80, regBC_);
 z80regs->de_p = z80ex_get_reg(z80, regDE_);
 z80regs->hl_p = z80ex_get_reg(z80, regHL_);

 z80regs->ix = z80ex_get_reg(z80, regIX);
 z80regs->iy = z80ex_get_reg(z80, regIY);
 z80regs->pc = z80ex_get_reg(z80, regPC);
 z80regs->sp = z80ex_get_reg(z80, regSP);
  
 z80regs->i = z80ex_get_reg(z80, regI);
 z80regs->r = z80ex_get_reg(z80, regR);
} 

//==============================================================================
// Return Z80 reg A while processing an IO call.
//
//   pass: void
// return: int                          Z80 reg A
//==============================================================================
int z80api_get_proc_reg_a (void)
{
 return (z80ex_get_reg(z80, regAF) & 0xff00) >> 8;
}

//==============================================================================
// Return Z80 reg B while processing an IO call.
//
//   pass: void
// return: int                          Z80 reg B
//==============================================================================
int z80api_get_proc_reg_b (void)
{
 return (z80ex_get_reg(z80, regBC) & 0xff00) >> 8;
}

//==============================================================================
// Return opcode for current Z80 IO instruction.
//
//   pass: void
// return: int                          Z80 opcode
//==============================================================================
int z80api_get_proc_opcode (void)
{
 return read_mem_cb(z80, z80ex_get_reg(z80, regPC) - 2, 0, NULL);
}

//==============================================================================
// Return name of Z80 emulator and version.
//
// The format of the string should be 'name version'
//
//   pass: char *vers
//         int size                     size of string
// return: void
//==============================================================================
void z80api_get_version (char *vers, int size)
{
 // the version function is implemented in z80ex-1.1.17.  If compiling against an
 // older version then define Z80EX_NO_VERSION_CODE in the Makefile.

#ifdef Z80EX_NO_VERSION_CODE
 snprintf(vers, size, "z80ex v???");
#else 
 Z80EX_VERSION *version = z80ex_get_version();
 snprintf(vers, size, "z80ex %s", version->as_string);
#endif

 vers[size-1] = 0;
}

//==============================================================================
// Z80 register dump
//
// Dump the contents of the Z80 registers.
//
//   pass: void
// return: void
//==============================================================================
void z80api_regdump (void)
{
 z80regs_t z80regs;
 
 z80api_get_regs(&z80regs);

 xprintf("\n");
 xprintf("Z80 Registers    Contents\n");
 xprintf("--------------------------------\n");
 xprintf("%-16s 0x%04x (%dd)\n", "AF", z80regs.af, z80regs.af);
 xprintf("%-16s 0x%04x (%dd)\n", "BC", z80regs.bc, z80regs.bc);
 xprintf("%-16s 0x%04x (%dd)\n", "DE", z80regs.de, z80regs.de);
 xprintf("%-16s 0x%04x (%dd)\n", "HL", z80regs.hl, z80regs.hl);
 xprintf("%-16s 0x%04x (%dd)\n", "AF_p", z80regs.af_p, z80regs.af_p);
 xprintf("%-16s 0x%04x (%dd)\n", "BC_p", z80regs.bc_p, z80regs.bc_p);
 xprintf("%-16s 0x%04x (%dd)\n", "DE_p", z80regs.de_p, z80regs.de_p);
 xprintf("%-16s 0x%04x (%dd)\n", "HL_p", z80regs.hl_p, z80regs.hl_p);
 xprintf("%-16s 0x%04x (%dd)\n", "IX", z80regs.ix, z80regs.ix);
 xprintf("%-16s 0x%04x (%dd)\n", "IY", z80regs.iy, z80regs.iy);
 xprintf("%-16s 0x%04x (%dd)\n", "SP", z80regs.sp, z80regs.sp);
 xprintf("%-16s 0x%04x (%dd)\n", "PC", z80regs.pc, z80regs.pc);
 xprintf("%-16s 0x%02x (%dd)\n", "I", z80regs.i, z80regs.i);
 xprintf("%-16s 0x%02x (%dd)\n", "R", z80regs.r & 0xff, z80regs.r & 0xff);
 xprintf("\n");
 xprintf("%-16s 0x%02x (%dd)\n", "A", z80regs.af >> 8, z80regs.af >> 8);
 xprintf("%-16s 0x%02x (%dd)\n", "F", z80regs.af & 0xff, z80regs.af & 0xff);
 xprintf("%-16s 0x%02x (%dd)\n", "B", z80regs.bc >> 8, z80regs.bc >> 8);
 xprintf("%-16s 0x%02x (%dd)\n", "C", z80regs.bc & 0xff, z80regs.bc & 0xff);
 xprintf("%-16s 0x%02x (%dd)\n", "D", z80regs.de >> 8, z80regs.de >> 8);
 xprintf("%-16s 0x%02x (%dd)\n", "E", z80regs.de & 0xff, z80regs.de & 0xff);
 xprintf("%-16s 0x%02x (%dd)\n", "H", z80regs.hl >> 8, z80regs.hl >> 8);
 xprintf("%-16s 0x%02x (%dd)\n", "L", z80regs.hl & 0xff, z80regs.hl & 0xff);
}

//==============================================================================
// Read a byte from a Z80 memory location.
//
//   pass: int addr
// return: Z80EX_BYTE
//==============================================================================
Z80EX_BYTE z80api_read_mem (int addr)
{
 return read_mem_cb(NULL, addr, 0, NULL);
}

//==============================================================================
// Write a byte to a Z80 memory location.
//
//   pass: int addr
//         uint8_t value
// return: void
//==============================================================================
void z80api_write_mem (int addr, uint8_t value)
{
 write_mem_cb(NULL, addr, value, NULL);
}

//==============================================================================
// Z80ex Read memory call back.
//
//   pass: Z80EX_CONTEXT *cpu
//         Z80EX_WORD addr
//         int m1_state
//         void *user_data
// return: Z80EX_BYTE
//==============================================================================
Z80EX_BYTE read_mem_cb (Z80EX_CONTEXT *cpu, Z80EX_WORD addr, int m1_state, void *user_data)
{
#ifdef MEMMAP_HANDLER_1
 return (Z80EX_BYTE)z80_mem_r[(addr & 0xF000) >> 12].memory_call(addr, NULL);
#else
 int i;
 
 for (i=0;;i++)
    {
     if (((int)addr >= z80_mem_r[i].low_addr) && ((int)addr <= z80_mem_r[i].high_addr))
        return (Z80EX_BYTE)z80_mem_r[i].memory_call(addr, NULL);
    }
#endif
}

//==============================================================================
// Z80ex Write memory call back.
//
//   pass: Z80EX_CONTEXT *cpu
//         Z80EX_WORD addr
//         Z80EX_BYTE value
//         void *user_data
// return: void
//==============================================================================
void write_mem_cb (Z80EX_CONTEXT *cpu, Z80EX_WORD addr, Z80EX_BYTE value, void *user_data)
{
#ifdef MEMMAP_HANDLER_1
 z80_mem_w[(addr & 0xF000) >> 12].memory_call(addr, value, NULL);
#else
 int i;

 for (i=0;;i++)
    {
     if (((int)addr >= z80_mem_w[i].low_addr) && ((int)addr <= z80_mem_w[i].high_addr))
        {
         z80_mem_w[i].memory_call(addr, value, NULL);
         return;
        }
    }
#endif
}

//==============================================================================
// Z80ex Read port call back.
//
//   pass: Z80EX_CONTEXT *cpu
//         Z80EX_WORD port
//         void *user_data
// return: Z80EX_BYTE
//==============================================================================
Z80EX_BYTE read_port_cb (Z80EX_CONTEXT *cpu, Z80EX_WORD port, void *user_data)
{
 return z80_ports_r[port & 0x00ff](port, NULL);
}

//==============================================================================
// Z80ex Write port call back.
//
//   pass: Z80EX_CONTEXT *cpu
//         Z80EX_WORD port
//         Z80EX_BYTE value
//         void *user_data
// return: void
//==============================================================================
void write_port_cb (Z80EX_CONTEXT *cpu, Z80EX_WORD port, Z80EX_BYTE value, void *user_data)
{
 z80_ports_w[port & 0x00ff](port, value, NULL);
}

//==============================================================================
// Z80ex Read interrupt vector call back.
//
//   pass: Z80EX_CONTEXT *cpu
// return: Z80EX_BYTE
//==============================================================================
Z80EX_BYTE read_interrupt_vector_cb (Z80EX_CONTEXT *cpu, void *user_data)
{
 return (Z80EX_BYTE)intr_vector;
}
