//******************************************************************************
//*                                  uBee512                                   *
//*       An emulator for the Microbee Z80 ROM, FDD and HDD based models       *
//*                                                                            *
//*                              options module                                *
//*                                                                            *
//*                   Copyright (C) 2007-2009 Stewart Kay                      *
//******************************************************************************
//
// Provides command line and initilisation file options processing.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v4.4.0 - 14 August 2009, S.J.Kay
// - Added 'pc85b' model to --model option in help section.
// - Fixed --pak and --pakx index bug introduced after enumeration of
//   options.
//
// v4.3.0 - 31 July 2009, S.J.Kay
// - Added '2mhzdd' and 'dd' Dreamdisk model parameters to help section.
// - Fixed --hddx-close options, incorrect index was being used.
//
// v4.2.0 - 19 July 2009, S.J.Kay
// - Added options --hdd0 to --hdd6 for WD1002-6 Winchester/floppy card.
// - Added options --hdd3-close, --hdd4-close, --hdd5-close, --hdd6-close.
// - Added '+hdd' argument to --modio option.
// - Added --mouse option for emulation of a Microbee mouse.
// - Added '+mouse' argument argument to --status option.
// - Added --bootkey option to pass a key (scan code) to boot ROMs.
// - Added --bpr option for repeatable break point action in debug mode.
// - Added '+pio' argument to --regs option.
// - Changed --verbose option to take optional level argument.
// - Enumerated all options for easier maintenance.
//
// v4.1.0 - 4 July 2009, S.J.Kay
// - Added --rom256k override option for 3rd party models.
// - Added --ide-a0, --ide-a1, --ide-b0 and --ide-b1 IDE drive options.
// - Added models 'scf' and 'pcf' (Compact Flash CB Standard/Premium)
// - Added --cfmode option to force start-up mode when emulation CF model.
// - Added --cpu-delay option to determine the method used for delays when
//   controlling emulation speed.
// - Added '+ide' argument to --modio option.
// - Changed +paknet argument to +roms.
// - Fixed options_ubee512_envvar_get() as was incorrectly matching
//   variables that had the same search name as the leading part of the
//   name.
//
// v4.0.0 - 18 May 2009, S.J.Kay
// - Changes made to incorporate a Z80 API.  Specific Z80 code used here
//   has been replaced with code usable by any Z80 emulator.
// - Changes made to --version option to also include Z80 emulator.
//
// v3.1.0 - 22 April 2009, S.J.Kay
// - Added --verbose option to switch on additional emulator reporting.
// - Removed all occurrences of console_output() function calls.
// - Added 'clear' argument to --regs option.
// - Added --hardware option to disable emulation of various hardware.
// - Changed --tapevol option to now use a percentage value.
// - Added --tapei-det option to set an optional input level detection value.
// - Added --osd option to configure OSD operation.
// - Added --output option to set text output devices for xprintf().
// - Added --century option to change century dates where appropriate.
// - Added --keystd-mod option for setting standard keyboard behaviour flags.
// - The --psec option no longer does anything and is ignored.
// - Added --hwflashr option to set the video hardware flashing rate.
// - Options --mon-fgl-b, --mon-fgl-g and --mon-fgl-r are now obsolete and
//   will be ignored.  A warning message will be reported if any of these
//   are used.
// - Added --mon-bgi-b, --mon-bgi-g, --mon-bgi-r, --mon-fgi-b, --mon-fgi-g,
//   and --mon-fgi-r options for setting the user dual monochrome intensity
//   RGB values.
// - Added --dint option (dual intensity) to replace --hint option.  The
//   latter option will remain but --dint should be used.
// - Changed --hwflash option, takes 'off', 'on', 'v3' and 'v4' as arguments.
// - Fixed debug option parameters for step and trace.
// - Added --runsecs option to set number of seconds to run before exiting.
// - Added colour, amber, green, black, white, and user to the single
//   character --monitor option parameters of c,a,g,b,w and u.
// - Added Windows OSVERSIONINFOEX dwMajorVersion and dwMinorVersion members
//   as variables UBEE_SYS_MAJOR_VAL and UBEE_SYS_MINOR_VAL to mingw build.
// - Changed -f, --fullscreen and -t, --turbo options to now have support
//   for optional on/off arguments.
// - Added --coms-close option to close open serial communications ports.
// - Added --a-close, --b-close, --c-close, and --d-close options.
// - Added --print-close and --printa-close options.
// - Added --tapei-close and --tapeo-close options.
// - Added --options-warn to issue a warning or produce an error when an
//   option is encountered that is not supported in run mode.
// - Fixed sound volume set option(s), was unable to set any level.
// - Removed the options_exit() function as closing open file handled on
//   entry to options_process() now.
// - Changed all printf() calls to use a local xprintf() function.
// - Changed all fflush() calls to use a local xflush() function.
// - Added ability to process options during the running of the emulator.
//   This has resulted in a large number of changes to some options. Options
//   now only set affected values if the parameter is legal.
// - Added optional administration sections 'global-start-runmode' and
//   'global-end-runmode' for configuration files when in run mode.
// - Added --prefix option to specify the installation path prefix.
// - Added --account option to specify an alternative home account location.
// - getopt_long() now uses local xgetopt_long() function.
// - Moved the 'UBEE512' and 'ubee512' home variable setting to ubee512.c
//   Added OpenGL conditional code compilation using USE_OPENGL.
// - Changed --config option to no longer cause the program to terminate if
//   the specified file does not exist.  A message will be output and no
//   config file will be used.
//
// v3.0.0 - 13 October 2008, S.J.Kay
// - Re-organised the hugely long options_getopt() function by moving each
//   option into their own respective functions. option_*name*()
// - Added code to correctly handle double quoted command line arguments
//   when using win32 before being processed by getopts_long().
// - Added options_init() function.
// - Added --varset and --varuset option to set/unset built in uBee512
//   variables.
// - Added conditional options --if-egt, --if-elt, --if-eq, --if-gt,
//   --if-lt, --if-negt, --if-nelt, --if-neq, --if-ngt, --if-nlt, --if-nset,
//   --if-set, --if-false, --if-true, --if-system, --if-else, --if-end and
//   --if-cmpmode.
// - Added --echo and --echoq options.
// - Added +video and +options to the --modio option arguments.
// - Added 'win' parameter to the --mouse-wheel and --status options.
// - Added 'gl' OpenGL parameter to the --video-type option.
// - Added gl-aspect-bee, gl-aspect-mon, gl-max, gl-vsync, gl-winpct,
//   gl-winpix, --gl-filter-fs, --gl-filter-max and --gl-filter-win
//   options.
// - Changed gui.fullscreen to video.fullscreen, crtc.video_depth to
//   video.depth, and crtc.video_type to video.type.
// - Changed --snd-freqlow help as is now 20Hz default.
// - Changed help information concerning the --cmd-repeat2 option.
// - Added UBEE_HOST and UBEE_VERSION internal pre-defined variables.
// - Added UBEE_SYS_MAJOR and UBEE_SYS_MINOR internal pre-defined variables.
// - Added options_ubee512_envvar_set() and options_ubee512_envvar_get()
//   functions for setting and retrieving uBee512 variables.
// - Changes to extract_environment_vars() function for uBee512 variables.
// - Added --exit option to allow exiting from start up in a script.
// - Renamed the options_free() function to options_exit() and removed the
//   code used to free the allocated memory as nothing really to be gained.
//
// v2.8.0 - 2 September 2008, S.J.Kay
// - Added --mouse-wheel option to associate actions for wheel scroll events.
// - Added --cmd-repeat1 and --cmd-repeat2 options to set command repeat delays.
// - Added +vol argument parameter to --status options.
// - Changed --snd-volume, --vol to now take a percentage integer value.
// - Added --gui-persist option to set persist times in milliseconds for
//   values that appear on the status line.
// - Changed --js-kset option to now also be the default joystick key set.
// - Added --lcons option to set the list start point for --lcon and --lconw.
// - Added --js-clist option to list the commands available for mapping to
//   a joystick.
// - Modified --file-list and --file-list-q options to allow more than one
//   file string to be passed to Z80 applications.
// - Added --lockfix-win32 and --lockfix-x11 options. These allow the LOCK
//   key work around code to be enabled/disbabled for each system type.
// - crtc.early_col_type is now crtc.std_col_type. Edited help documentation
//   to refer to 'standard colour' instead of 'early colour'.
// - Added --video-depth option to select 8, 8gs, 16 and 32 bit video depths.
// - Added --video-type option to select SW and HW video modes.
// - Fixed --col-type option to set the colour type only if emulating a
//   non-alpha+ model.
// - Changed the default colour monitor type to RGBrgb when --col is used.
// - Changed --frate help as Unices is now also 50 FPS.
//
// v2.7.0 - 2 July 2008, S.J.Kay
// - Changed rgb-nn-x ordering from bgr to rgb as table has changed.
// - Added --col-type option to select early colour circuit monitor types.
// - Added --lconw option to list section names in config file in wide format.
// - Added --clock-def option to be used by uBee512 API functions.
// - Added --args-error options to disable some argument option errors.
// - Added --exit-check option to allow quick exits with no checking.
// - Added application options --file-app, --file-exec, --file-exit,
//   --file-list, --file-list-q, --file-load, and --file-run.
// - Added +mute argument parameter to --status options.
// - Added --alias-disks and --alias-roms options.
// - options_readline() function moved over to functions.c module and is now
//   file_readline() with file pointer and string size also passed.
// - Added --slashes option to enable/disable conversion of path slashes.
// - Added support for environment variables in configuration files.
// - Added joystick options --js-axis, --js-axisb, and --js-axisl to allow
//   analogue axis movements to return joystick buttons.
// - Changed the way files names are handled. Determining the paths is now
//   handled by the open_file() function in the functions.c module.
// - Added --dstep option to support 48tpi disks in 96tpi DD drives.
// - Added --dstep-hd option to support 48tpi disks in 96tpi HD drives.
// - Added +ubee512 to the --modio option arguments.
// - Added option --z80div to set the number of Z80 blocks executed per Z80
//   frame.
// - Added option --maxcpulag that sets the maximium Z80 CPU lag time allowed.
// - Added options --snd-alg1, --snd-freq=f, --snd-freqadj, --snd-freqlow,
//   --snd-holdoff, --snd-samples, --snd-hq, --snd-mute, and --snd-volume.
// - Added option --config to allow alternative configuration files to be used.
// - Added structures emu_t, crtc_t, func_t, sound_t and moved the associated
//   variables into it.
// - Made changes for the --nodisk option.
// - Added --video option to disable/enable video output on startup.
// - Added function calls to set colours instead of doing direct.
//
// v2.6.0 - 13 May 2008, S.J.Kay
// - Many tape, printer and serial variables have been placed into structures
//   and declared in the appropriate modules.
// - Added --status option to cutomize the status line in the title bar.
// - Added --title option to set a title name.
// - Added --spad option to set padding between status information.
// - Added joystick select and enable option --js.
// - Added Microbee joystick options: --js-clear, --js-mbee, --js-up,
//   --js-right, --js-down, --js-left, --js-fire, --js-play1, --js-play2,
//   and --js-spare.
// - Added joystick to keys mapping options: --js-klist --js-kbd, --js-kk,
//   --js-kb, and --js-kset.
// - Added +joystick parameter to --modio options.
// - Added --dclick option to set double click rate for full screen toggle.
// - Added --basram option to use SRAM instead of ROM at 0xA000-0xBFFF.
// - Added --pakram option to allow a PAK location to use SRAM instead of ROM
//   at 0xC000-0xDFFF.
// - Added --netram option to use SRAM instead of ROM at 0xE000-0xEFFF.
// - Moved the string_search() function over to the function.c module.
//
// v2.5.0 - 30 March 2008, S.J.Kay
// - Restructured the help output and increased lineswanted to 24 from 23.
// - Added --lmodel option to list the available models.
// - Added Teleterm model emulation.
// - Added user's monochrome configuration type 'u' to the --monitor option.
// - Added --mon-fg-x, --mon-bg-x, and --mon-fgl-x options for a user
//   configurable monochrome monitor.
// - Added --basic, --basica, --basicb,--basicc, --basicd, options to override
//   default BASIC in ROM based models.
// - Added --pak, --pak0 to --pak7, and --netrom options to override PAK and
//   net ROMs in ROM based models.
// - Added --rom1, --rom2, --rom3 options to override ROMs in DRAM models.
// - Added --charrom option to override the default character ROM.
// - Added --sys option.  Used to allow some files to include the system name.
// - Added --lcon option to list section names in the configuration file.
// - Added --colprom option to override early colour internal PROM values.
// - Implement the modelc structure.
// - Added 48 --rgb-nn-x options to custom configure the alpha+ colours.
// - Removed un-necessary information output used by some options.
// - Modified --fullscreen option to now have a toggling action.
// - Moved the options usage help information into this module.
// - Created a new file containing the options processing code for command
//   line and initilisation options.

//============================================================================== 
/*
 *  uBee512 - An emulator for the Microbee Z80 ROM, FDD and HDD based models. 
 *  Copyright (C) 2007-2009 Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//==============================================================================   

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <SDL.h>

#ifdef USE_LIBDSK
#include <libdsk.h>
#endif

#ifdef MINGW
#include <windows.h>
#include <conio.h>
#else
#include <sys/utsname.h>
#endif

#include "getopt.h"
#include "ubee512.h"
#include "options.h"
#include "z80.h"
#include "z80api.h"
#include "crtc.h"
#include "vdu.h"
#include "crtc.h"
#include "memmap.h"
#include "keyb.h"
#include "fdc.h"
#include "hdd.h"
#include "ide.h"
#include "gui.h"
#include "osd.h"
#include "video.h"
#include "roms.h"
#include "pio.h"
#include "sound.h"
#include "tape.h"
#include "serial.h"
#include "mouse.h"
#include "printer.h"
#include "joystick.h"
#include "rtc.h"
#include "clock.h"
#include "function.h"
#include "z80debug.h"
#include "console.h"
#include "keystd.h"

#include "macros.h"

//==============================================================================
// structures and variables
//==============================================================================
char *c_argv[OPTIONS_SIZE];
int c_argc;

help_t help;

#define TRY_MESG "%s: Try `%s --help' or `%s --usage' for more information.\n"

char *monitor_args[] =
{
 "c",
 "a",
 "g",
 "b",
 "w",
 "u",
 "colour",
 "amber",
 "green",
 "black",
 "white",
 "user",
 ""
};

char *offon_args[] =
{
 "off",
 "on",
 ""
};

const SDL_version *sdlv;

static int int_arg;
static float float_arg;

static int if_state_prev;

static char temp_str[512];
 
static char e_optarg[OPTIONS_PARM_SIZE];
static char e_optarg_q[OPTIONS_PARM_SIZE];
static char e_optarg_u[OPTIONS_PARM_SIZE];

static fdc_drive_t fdc_d;
static hdd_drive_t hdd_d;
static ide_drive_t ide_d;

static int runmode_warn = 1;

#ifdef USE_LIBDSK
static dsk_format_t format;
static dsk_cchar_t fname, fdesc;
static char use_driver_type[40];
static char use_format_type[40];
static char *xstr;
static int side1as0;  
static int dstep;
static int dstep_hd; 
#endif 

static int exitstatus;
static int args_err_flags = 0xffffffffL;
static int list_config_start;

static char config_file[SSIZE1];

static char *ndefsv[OPTIONS_SIZE];
static int ndefsc;

static char *xargv[OPTIONS_SIZE];
static int xargc;

static char *emuenv[OPTIONS_ENV_SIZE];
static int emuenvc;

static FILE *fp;

static int if_pos;
static int if_state[OPTIONS_MAXCOND];
static int if_cmp_mode;

#ifdef MINGW
static char win_major_ver[20];
static char win_minor_ver[20];
#else
static struct utsname buf;
#endif

extern char userhome_confpath[];
extern char userhome[];
extern uint8_t col_table_p[];

extern char *model_args[];
extern fdc_drive_t fdc_drive[];
extern ide_drive_t ide_drive[];

extern emu_t emu;
extern model_t model_data[];
extern model_t modelx;
extern model_custom_t modelc;
extern crtc_t crtc;
extern fdc_t fdc;
extern gui_t gui;
extern osd_t osd;
extern video_t video;
extern gui_status_t gui_status;
extern mouse_t mouse;
extern debug_t debug;
extern modio_t modio;
extern regdump_t regdump;
extern func_t func;
extern joystick_t joystick;
extern serial_t serial;
extern sound_t sound;
extern tape_t tape;
extern keystd_t keystd;
extern console_t console;

static char *options_malloc (int size);

//==============================================================================
// Options initilisation.
//
// Convert Windows command line arguments to be similar to Unix arguments
// by preserving double quoted arguments.
//
// Set the internal variable values for Unix and Windows systems and some
// other built in variables.
//
//   pass: void
// return: void
//==============================================================================
void options_init (void)
{
 char temp_str[SSIZE1];

#ifdef MINGW
 options_make_pointers(GetCommandLine());

 // get host system version information
 OSVERSIONINFO osvi;

 ZeroMemory(&osvi, sizeof(OSVERSIONINFO));
 osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
 GetVersionEx(&osvi);

 switch (osvi.dwMajorVersion)
    {
     case 3 : strcpy(win_major_ver, "NT3");
              break;
     case 4 : if ((osvi.dwMinorVersion == 0) || (osvi.dwMinorVersion == 10) ||
              (osvi.dwMinorVersion == 90))
                 {
                  strcpy(win_major_ver, "WIN9X_ME");
                  switch (osvi.dwMinorVersion)
                     {
                      case  0 : strcpy(win_minor_ver, "W95");
                                break;
                      case 10 : strcpy(win_minor_ver, "W98");
                                break;
                      case 90 : strcpy(win_minor_ver, "ME");
                                break;
                     }
                  break;
                 } 
              if ((osvi.dwMinorVersion == 1) || (osvi.dwMinorVersion == 3))
                 {
                  strcpy(win_major_ver, "NT4");
                  switch (osvi.dwMinorVersion)
                     {
                      case 1 : strcpy(win_minor_ver, "NT4_WS");
                               break;
                      case 3 : strcpy(win_minor_ver, "NT4_SERVER");
                               break;
                     }
                 }
              break;
     case 5 : strcpy(win_major_ver, "NT5");
              switch (osvi.dwMinorVersion)
                 {
                  case 0 : strcpy(win_minor_ver, "W2000");
                           break;
                  case 1 : strcpy(win_minor_ver, "XP");
                           break;
                  case 2 : strcpy(win_minor_ver, "SERVER_2003");
                           break;
                 }
              break;
     case 6 : strcpy(win_major_ver, "NT6");
              switch (osvi.dwMinorVersion)
                 {
                  case 0 : strcpy(win_minor_ver, "VISTA");
                           break;
                 }
              break;

    }
 options_ubee512_envvar_set("UBEE_HOST=WIN");
 snprintf(temp_str, sizeof(temp_str), "UBEE_SYS_MAJOR=%s", win_major_ver);
 options_ubee512_envvar_set(temp_str);
 snprintf(temp_str, sizeof(temp_str), "UBEE_SYS_MAJOR_VAL=%d", (int)osvi.dwMajorVersion);
 options_ubee512_envvar_set(temp_str);
 snprintf(temp_str, sizeof(temp_str), "UBEE_SYS_MINOR=%s", win_minor_ver);
 options_ubee512_envvar_set(temp_str);
 snprintf(temp_str, sizeof(temp_str), "UBEE_SYS_MINOR_VAL=%d", (int)osvi.dwMinorVersion);
 options_ubee512_envvar_set(temp_str);
#else
 uname(&buf);
 strcpy(emu.sysname, buf.sysname);
 toupper_string(emu.sysname, emu.sysname);
 
 options_ubee512_envvar_set("UBEE_HOST=UNIX");
 options_ubee512_envvar_set("UBEE_SYS_MAJOR=UNIX");
 snprintf(temp_str, sizeof(temp_str), "UBEE_SYS_MINOR=%s", emu.sysname);
 options_ubee512_envvar_set(temp_str);
#endif 

 // set some pre-configured local environment variables
 options_ubee512_envvar_set("UBEE_VERSION="APPVER);
}

//==============================================================================
// Convert console arguments to pointers for each argument found in 's'.
// Memory is allocated with malloc for each argument found.  Doubled quoted
// arguments are preserved and treated as one argument.  The pointer array
// created will be used by the xgetopts_long() function.  A 'ubee512'
// argument is inserted for the first argument if no arguments were passed.
//
// This function is used for Windows command line as double quoted arguments
// have no special meaning.
//
// This function is also used when using an inbuilt console during the
// running of the emulator.
//
//   pass: char *s
// return: void
//==============================================================================
void options_make_pointers (char *s)
{
 char argx[SSIZE1];
 int i;

 // free existing pointers in use
 while (c_argc)
    free(c_argv[--c_argc]);
 
 while ((*s) && (c_argc < OPTIONS_SIZE))
    {
     argx[0] = 0;
     while ((*s) && (*s <= ' '))  // move past white space
        s++;

     i = 0;   

     while ((*s) && (*s != ' ') && (*s != '"'))     
        {
         if (*s)
            {
             argx[i++] = *s;
             s++;
            } 
        }

     if (*s == '"')
        {
         s++;
         while ((*s) && (*s != '"'))
            {
             argx[i++] = *s;
             s++;
            }
         s++;   
        }

     argx[i] = 0;

     if (argx[0])
        {
         c_argv[c_argc] = options_malloc(i+1);

         if (c_argv[c_argc])
            {
             strcpy(c_argv[c_argc], argx);
             c_argc++;
            }
        }
    }

 if (! c_argc)  // insert an argv[0] value of 'ubee512' if no args
    {
     strcpy(argx, "ubee512");
     c_argv[c_argc] = options_malloc(strlen(argx)+1);
     if (c_argv[c_argc])
        {
         strcpy(c_argv[c_argc], argx);
         c_argc++;
        } 
    }

#if 0
 for (i = 0; i < c_argc; i++)
    xprintf("|%s| string len=%d  c_argc=%d\n", c_argv[i], strlen(c_argv[i]), c_argc);
 xprintf("END OF ARGS\n");    
#endif
}

//==============================================================================
// Options modio information
//
//   pass: void
// return: void
//==============================================================================
void options_modio_info (void)
{
#ifdef MINGW
 if (modio.options)
    {
     xprintf("options_init: win_major_ver=%s   win_minor_ver=%s\n",
     win_major_ver, win_minor_ver);
     if (modio.level)
        fprintf(modio.log, "options_init: win_major_ver=%s   win_minor_ver=%s\n",
        win_major_ver, win_minor_ver);
    }    
#else
 if (modio.options)
    {
     xprintf("options_init: uname.sysname field=%s\n", buf.sysname);
     if (modio.level)
        fprintf(modio.log, "options_init: uname.sysname field=%s\n", buf.sysname);
    }    
#endif 
}

//==============================================================================
// Options usage
//
// Provide help using a state machine or blocking mode.  The state machine
// will be used by the OSD console dialogue.  Blocking mode will be used when
// help is requested from the command line or when using the blocking ALT+C
// console command.
//
// The help information is arranged in sections, each section and members of
// each section are arranged in alphabetical order.
//
// The last printable character as it appears in this source should be less
// than column position 81.
//
//   pass: help_t help
// return: void
//==============================================================================
void options_usage_state (help_t *help)
{
 char ch;

 char const usage[] =
TITLESTRING"\n"
"\n"
"Usage: ubee512 [options]\n"
"\n"
// +++++++++++++++++++++++++++ Control related +++++++++++++++++++++++++++++++++
" Control related:\n\n"
"  --account=path          Specify an alternative account location. To create\n"
"                          other accounts in the home path on Unices or Windows\n"
"                          use path=@UBEE_USERHOME@\\name.  Accounts may also be\n"
"                          created on removable media.  Use this option whenever\n"
"                          an alternative account should be used,  the default\n"
"                          account is @UBEE_USERHOME@/.ubee512 on Unices and\n"
"                          the location of the executed ubee512.exe on Windows\n"
"                          machines.  This option if used must be the first\n"
"                          option declared on the command line.\n"
"\n"
"  --alias-disk=x          Enables/disables checking for disk aliases in the\n"
"                          disks.alias file. x=on to enable, x=off to disable.\n"
"                          Default is enabled.\n"
"\n"
"  --alias-roms=x          Enables/disables checking for ROM aliases in the\n"
"                          roms.alias file. x=on to enable, x=off to disable.\n"
"                          Default is enabled.\n"
"\n"
"  --args-error=options    Clears an option error detection flag.  Arguments\n"
"                          commence with a '-' character to disable a flag.\n"
"                          Additional arguments may be declared in the same\n"
"                          --args-error option.\n"
"\n"
"                          The arguments supported are:\n"
"                          -unknown - disable non-recognised argument error.\n"
"\n"
"  --bootkey=key           Forces a light-pen key scan code on start-up. This\n"
"                          is needed by some ROMs to enter certain operating\n"
"                          modes. The ASCII key value is converted to a scan\n"
"                          code. a-z, and A-Z are converted to codes 1-26 and\n"
"                          ASCII 0-9 to codes 32-41.\n"
"\n"
"  --cfmode=x              Force emulation mode for CF model. x=pc85 for PC85\n"
"                          mode, x=boot for normal boot emulation. Default is\n"
"                          boot mode.\n"
"\n"
"  --config=file           Allows an alternative configuration file to be used\n"
"                          or if file='none' then no configuration file will be\n"
"                          used.  This option if used must be the first or\n"
"                          second option declared on the command line.  The\n"
"                          default file used for configuration is 'ubee512rc'\n"
"                          and must be located in the ubee512 directory.\n"
"\n"
"  --cmd-repeat1=n         Set the first delay period in milliseconds to be\n"
"                          used for repeated emulator commands. Normally these\n"
"                          are activated with EMUKEY and joystick buttons\n"
"                          mapped to commands. Default value is 500mS.\n"
"  --cmd-repeat2=n         Same as --cmd-repeat1 except this value determines\n"
"                          the delay period to be used after the first period.\n"
"                          Default value is 50mS.\n"
"\n"
"  --cpu-delay=n           Set the delay method used for controlling Z80 CPU\n"
"                          emulation speed. The default value is 0.\n"
"                          The arguments supported are:\n"
"\n"
"                          0 : delays give up processor time.\n"
"                          1 : delays do not give up processor time.\n"
"                          2 : if data is in the sound buffer then use method 1\n"
"                              otherwise method 0 applies.\n"
"\n"
"  --dclick=n              Set the double click speed for mouse button events.\n"
"                          n may be 100-3000 milliseconds, default is 300mS.\n"
"\n"
"  --exit=x                Forces the emulator to exit. This option is intended\n"
"                          to be used inside start up scripts when a condition\n"
"                          is not met.  x is the exit status value.\n"
"\n"
"  --exit-check=x          Enables/disables exit checking. if enabled the user\n"
"                          must confirm before exiting the emulator. x=on to\n"
"                          enable, x=off to disable. Default is enabled.\n"
"\n"
"  --gui-persist=n         Set the persist time in milliseconds for values that\n"
"                          appear on the status line, default is 3000mS.\n"
"\n"
"  --keystd-mod=options    Set a standard keyboard behaviour modifier flag.\n"
"                          Arguments commence with a '+' character. Additional\n"
"                          arguments may be declared in the same --keystd-mod\n"
"                          option.  These flags provide workarounds when\n"
"                          emulating the 6545 light pen keys.  Default value is\n"
"                          'ctrl_shift' enabled.\n"
"\n"
"                          The arguments supported are:\n"
"                          +clear - clear all current flag values.\n"
"                          +ctrl_shift - allows extended function keys to work.\n"
"\n"
"  --lockfix-win32=x       Enables/disables LOCK key code fix for Win32. This\n"
"                          option has been provided in case it needs to be\n"
"                          disabled. x=on to enable, x=off to disable. Default\n"
"                          is enabled.\n"
"\n"
"  --lockfix-x11=x         Enables/disables LOCK key code fix for x11. This\n"
"                          option has been provided in case it needs to be\n"
"                          enabled (possibly on some x11 set-ups). x=on to\n"
"                          enable, x=off to disable. Default is disabled.\n"
"\n"
"  --mmode                 Forces the return of the 'M' key once when the\n"
"                          emulator is started.  This is used for jumping\n"
"                          directly into the ROM's Monitor mode.\n"
"\n"
"  --mouse-wheel=x         Set the action associated with mouse wheel scrolling\n"
"                          The default is 'vol', the association can also be\n"
"                          changed with the EMUKEY+W hot key.\n"
"\n"
"                          The arguments supported are:\n"
"                          none : no association (does nothing).\n"
"                           vol : adjust application volume level.\n"
"                           win : resize windows display when in OpenGL mode.\n"
"\n"
"  --nodisk                Set no disks flag,  use to start some boot ROMs in\n"
"                          menu mode.  This flag will be cleared when any key\n"
"                          is pressed.\n"
"\n"
"  --options-warn=x        Turn warnings on/off for unsupported options\n"
"                          encountered during run mode.  If warning is enabled\n"
"                          the offending option is not processed but the\n"
"                          processing of options continues.  If warning is\n"
"                          disabled the offending and all remaining options\n"
"                          will not be processed. x=on to enable, x=off to\n"
"                          disable. Default is enabled.\n"
"\n"
"  --osd=options           OSD configuration.  Arguments commence with a '+'\n"
"                          character.  Additional arguments may be declared in\n"
"                          the same --osd option.\n"
"\n"
"                          The arguments supported are:\n"
"                          +clear - clear all current OSD arguments.\n"
"                          +animate - animate OSD window minimising (default).\n"
"\n"
"  --output=options        Set output devices for all text output.  Arguments\n"
"                          commence with a '+' character. Additional arguments\n"
"                          may be declared in the same --output option.  Default\n"
"                          output is set to 'osd' and 'stdout' on Unices and\n"
"                          'osd' on Windows systems.\n"
"\n"
"                          The arguments supported are:\n"
"                          +none - no output device (use to clear).\n"
"                          +osd - output to the emulator's OSD console.\n"
"                          +stdout - output to STDOUT (or stdout.txt on win32).\n"
"\n"
"  --prefix=path           Specify an alternative installation location to be\n"
"                          used when creating an account on a Unix system.\n"
"                          Installed files are normally located in /usr/local/\n"
"                          but may be prefixed with 'path'.\n"
"\n"
"  --runsecs=n             Run the emulator for n seconds then exit.  A minimum\n"
"                          value of 5 seconds is allowed.  Any disk write\n"
"                          activity will increase the run value until several\n"
"                          seconds of disk write inactivity has passed,  this is\n"
"                          used to reduce the chances of exiting at a critical\n"
"                          moment.  Use this option with care.  Setting n=0 will\n"
"                          disable this feature and is the default.\n"
"\n"
"  --slashes=x             Conversion of path slashes to host format. x=on to\n"
"                          enable, x=off to to disable.  Default is enabled.\n"
"\n"
"  --spad=n                Sets the number of spaces to be placed between each\n"
"                          status entry on the title bar.  The actual spacing\n"
"                          achieved will be dependent on the title font used.\n"
"                          Default value is 2 spaces.\n"
"\n"
"  --status=options        Status configuration for title bar.  Arguments\n"
"                          commence with a '+' character. Additional arguments\n"
"                          may be declared in the same --status option.\n"
"\n"
"                          The arguments supported are:\n"
"                          +clear - clear all current status arguments.\n"
"                          +d - show short drive access.\n"
"                          +drive - show long drive access.\n"
"                          +emu - show emulator name.\n"
"                          +emuver - show emulator name and version.\n"
"                          +joy - show joystick status.\n"
"                          +left - force left hand justification.\n"
"                          +model - show base model emulated.\n"
"                          +mouse - show Microbee mouse emulated.\n"
"                          +mute - show the sound mute state.\n"
"                          +print - show parallel printer enable.\n"
"                          +ram - show amount of RAM emulated.\n"
"                          +serial - show serial port set up if enabled.\n"
"                          +speed - show CPU clock speed.\n"
"                          +sys - show system name.\n"
"                          +tape - show tape input/output state.\n"
"                          +title - show customised title.\n"
"                          +ver - show emulator version.\n"
"                          +vol - always show volume level.\n"
"                          +win - always show window size.\n"
"\n"
"  --title=name            Define the customised title name to be used when\n"
"                          '+title' is used in the --status option.\n"
"\n"
"  --varset=var[=val]      Set a uBee512 built in environment variable.\n"
"                          var contains the variable name and val is an optional\n"
"                          value to assign to it. i.e. --varset=myvar=myvalue.\n"
"  --varuset=var           Un-set (remove) a uBee512 built in environment\n"
"                          variable.  var is the variable name.\n"
"\n"
"  --verbose=level         Switch on additional emulator reporting.  The default\n"
"                          setting does not report any messages during start-up\n"
"                          unless error(s) occur.  The level value is optional\n"
"                          and defaults to 1 if omitted.\n"
"\n"
// +++++++++++++++++++++++ Conditional processing ++++++++++++++++++++++++++++++
" Conditional processing:\n\n"
"                          If any of the following conditionals returns a true\n"
"                          result then option processing is enabled, a false\n"
"                          result turns processing off until a true condition\n"
"                          is met.\n"
"\n"
"  --if-egt=str1,str2      If str1 is equal to or greater than str2.\n"
"  --if-elt=str1,str2      If str1 is equal to or less than str2.\n"
"  --if-eq=str1,str2       If str1 is equal to str2.\n"
"  --if-gt=str1,str2       If str1 is greater than str2.\n"
"  --if-lt=str1,str2       If str1 is less than str2.\n"
"  --if-negt=str1,str2     If str1 is not equal or greater than str2.\n"
"  --if-nelt=str1,str2     If str1 is not equal or less than str2.\n"
"  --if-neq=str1,str2      If str1 is not equal to str2.\n"
"  --if-ngt=str1,str2      If str1 is not greater than str2.\n"
"  --if-nlt=str1,str2      If str1 is not less than str2.\n"
"  --if-nset=var           If variable var has not been set.\n"
"  --if-set=var            If variable var has been set.\n"
"\n"
"  --if-false              If this is used then set false.\n"
"  --if-true               If this is used then set true.\n"
"\n"
"  --if-system=x           If the host system is equal to system x. On POSIX\n"
"                          systems (Unix) this value is tested against the value\n"
"                          returned by the uname function sysname field. Known\n"
"                          arguments supported are:\n"
"\n"
"                                  bsd : Unix BSD system.\n"
"                              freebsd : Unix FreeBSD system.\n"
"                                linux : Unix Linux system..\n"
"                                 unix : Unix system.\n"
"                                  win : Windows system.\n"
"                             win9x_me : Windows 95, 98 or Me system\n"
"                                  w95 : Windows 95 system.\n"
"                                  w98 : Windows 98 system.\n"
"                                   me : Windows Millennium system.\n"
"                                  nt4 : Windows NT4 systems.\n"
"                               nt4_ws : Windows NT4 Work station system.\n"
"                           nt4_server : Windows NT4 Server system.\n"
"                                  nt5 : Windows NT5 systems.\n"
"                                w2000 : Windows 2000 system.\n"
"                                   xp : Windows XP system.\n"
"                          server_2003 : Windows NT5 Server system.\n"
"                                  nt6 : Windows NT6 systems.\n"
"                                vista : Windows Vista system.\n"
"\n"
"  --if-else               If last conditional resulted in false.\n"
"  --if-end                End of a conditional block.\n"
"  --if-cmpmode=x          Set the method used for comparing values, x=0 uses\n"
"                          'C' style strverscmp() and x=1 uses strcmp(). Default\n"
"                          method is 0.\n"
"\n"
// ++++++++++++++++++++++++++++ Debugging tools ++++++++++++++++++++++++++++++++
" Debugging tools:\n\n"
"  --bp=addr               Set a Z80 address break point.  This option can be\n"
"                          specified as many times as is required.  A break\n"
"                          point can only be detected when in debug mode. The\n"
"                          break point is cleared after detection.\n"
"  --bpr=addr              Same action as --bp option except the break point is\n"
"                          not cleared after detection.\n"
"\n"
"  --bpc=count             Set a Z80 break point determined by the number of\n"
"                          instructions executed. Can only be specified once.\n"
"                          A break point can only be detected when in debug\n"
"                          mode.\n"
"\n"
"  -z, --debug=options     Debugging mode.  Arguments commence with a '+'\n"
"                          character,  additional arguments may be declared in\n"
"                          the same --debug option.\n"
"\n"
"                          The arguments supported are:\n"
"                          +alt - output the alternate and the I, R registers.\n"
"                          +all - output all options.\n"
"                          +count - use instruction counter in disassembly.\n"
"                          +index - output the index registers.\n"
"                          +on - turn debugging mode on without disassembly.\n"
"                          +step - turn on debugging mode and start stepping.\n"
"                          +trace - turn on debugging  mode and start tracing.\n"
"\n"
"  --dump=addr             Set the initial dump address value when using the\n"
"                          dump commands.  addr may be a valid Z80 address from\n"
"                          0 to 0xFFFF. If not specified then address 0 is used.\n"
"\n"
"  --echo=x                Echo a string to stdout.  The string may also contain\n"
"                          an environment variable.\n"
"  --echoq=x               Same as --echo option but echoes a quoted version of\n"
"                          the environment variable if any spaces are found.\n"
"\n"
"  --modio=options         Module I/O debugging output.  Arguments commence with\n"
"                          a '+' character,  additional arguments may be\n"
"                          declared in the same --modio option.\n"
"\n"
"                          The arguments supported are:\n"
"                          +log - logs to ubee512_log.txt (makes big log files!)\n"
"                          +raminit - uses bank numbers as DRAM init values.\n"
"\n"
"                          These arguments turn on port debugging for modules:\n"
"                          +clock, +crtc, +fdc, +fdc_wtd, +fdc_wth, +func, +hdd\n"
"                          +ide, +joystick, +keystd, +keytc, +mem, +options,\n"
"                          +roms, +pioa, +piob, +piocont, +rtc, +ubee512,\n"
"                          +vdu, +vdumem, +video, +z80.\n"
"\n"
"  --regs=options          Register dump. Determines what registers will be\n"
"                          dumped when the EMUKEY+R key is pressed.  Arguments\n"
"                          commence with a '+' character,  additional arguments\n"
"                          may be declared in the same --regs option.\n"
"                          The arguments supported are:\n"
"                          +clear, +crtc, +pio, +rtc, +z80.\n"
"\n"
// +++++++++++++++++++++++++++++ Disk drives +++++++++++++++++++++++++++++++++++
" Disk drives:\n\n"
"  --hdd(n)=file           The --hdd(n) options allow emulation of WD1002-5\n"
"                          Winchester and floppy disk controller drives. n=0-2\n"
"                          are hard disk drives and n=3-6 are floppy drives.\n"
"                          file=file path for drive (n).\n"
"\n"
"  --hdd3-close            close WD1002-5 floppy disk (1st)\n"
"  --hdd4-close            close WD1002-5 floppy disk (2nd)\n"
"  --hdd5-close            close WD1002-5 floppy disk (3rd)\n"
"  --hdd6-close            close WD1002-5 floppy disk (4th)\n"
"\n"
"  --ide-a0=file           file path for emulator IDE primary master drive.\n"
"  --ide-a1=file           file path for emulator IDE primary slave drive.\n"
"  --ide-b0=file           file path for emulator IDE secondary master drive.\n"
"  --ide-b1=file           file path for emulator IDE secondary slave drive.\n"
"\n"
"  -a, --image_a=file      file path for emulator floppy drive A\n"
"  -b, --image_b=file      file path for emulator floppy drive B\n"
"  -c, --image_c=file      file path for emulator floppy drive C\n"
"  -d, --image_d=file      file path for emulator floppy drive D\n"
"\n"
"  --a-close               close core board floppy disk A\n"
"  --b-close               close core board floppy disk B\n"
"  --c-close               close core board floppy disk C\n"
"  --d-close               close core board floppy disk D\n"
"\n"
#ifdef USE_LIBDSK
"                          LibDsk usage:\n"
"                          If LibDsk is to be used to access a floppy drive then\n"
"                          file path may be 'A:' or 'B:' for Windows or a device\n"
"                          file for Unices. i.e. /dev/fd0 and /dev/fd1 on Linux.\n"
"\n"
#endif
"                          General usage:\n"
"                          If a drive already has a disk open then the disk is\n"
"                          closed before opening a new one. Do NOT change disks\n"
"                          while the Z80 system is actively working on the\n"
"                          drive or has files open, changing disks requires the\n"
"                          same rules required by a Microbee to be observed.\n"
"                          i.e. type '^C' in CP/M 2.2 after changing disk(s).\n"
"\n"
"                          See 'File path searching' further on for detailed\n"
"                          information.  The default area for disks is:\n"
"\n"
"                          @UBEE512@\\disks\\\n"
"\n"
#ifdef USE_LIBDSK
"  --dstep                 Informs LibDsk the next Disk drives option uses\n"
"                          double stepping to support 48tpi DD disks in a 96tpi\n"
"                          DD drive. This option must precede each Disk drive\n"
"                          option when LibDsk is required.\n"
"\n"
"  --dstep-hd              Same use as the --dstep option except this is for\n"
"                          48tpi DD disks in 96tpi HD drives.\n"
"\n"
"  --format=type           Determines the disk format type when using LibDsk.\n"
"                          Using this option will cause the next Disk drives\n"
"                          option to use the LibDsk driver.  This option must\n"
"                          precede each Disk drive option when LibDsk is\n"
"                          required. Additional disk formats can be placed into\n"
"                          the local libdskrc file.\n"
"\n"
"                          ds40 and ds80 formats will automatically make use of\n"
"                          the --side1as0 option.  Use the ds401 and ds801\n"
"                          formats if this behaviour is not required. i.e. PC\n"
"                          formatted disks)\n"
"\n"
"  --lformat               Lists all the LibDsk built in and additional disk\n"
"                          formats that are available.\n"  
"\n"
"  --ltype                 Lists all the LibDsk driver types that are available.\n"
"\n"
"  --side1as0              Informs LibDsk the next Disk drives option uses a\n"
"                          disk that has physical side one sectors containing 0\n"
"                          in the sector headers. The FDC write track emulation\n"
"                          will force the side information in the sector header\n"
"                          to use the physical side value with this option.\n"
"                          This option is no longer required to read and write\n"
"                          disks that have this issue.\n"
"\n"
"  --type=driver           Determines what LibDsk driver will be used for the\n"
"                          next Disk drives option. This option must precede\n"
"                          each Disk drive option when LibDsk is required, if\n"
"                          not then LibDsk will attempt to automatically detect\n"
"                          the driver type to use.\n"
"\n"
#endif
// +++++++++++++++++++++++++++ Display related +++++++++++++++++++++++++++++++++
" Display related:\n\n"
"  --aspect=n              Set the display aspect when using an SDL video mode\n"
"                          for rendering, n=1 is 1:1,  default aspect 2:1 (n=2),\n"
"                          n may be set to 1 or 2.  1:1 scaling may be enforced\n"
"                          for some CRTC6545 display sizes'.\n" 
"\n"
"  -f, --fullscreen[=x]    Toggle state of full screen mode, the display\n"
"                          defaults to a window (use EMUKEY+ENTER to toggle).\n"
"                          If 'x' is specified then fullscreen mode can be set\n"
"                          with x=on or window mode set with x=off.\n"
"\n"
"  -m, --monitor=type      Monitor type, if this option is not specified a\n"
"                          colour monitor is the default when emulating colour\n"
"                          and green if a monochrome model. <type> may be one\n"
"                          of the following:\n"
"\n"
"                          a,  amber : amber monitor.\n"
"                          g,  green : green monitor.\n"
"                          w,  white : white monitor, white foreground on black\n"
"                                      background.\n"
"                          b,  black : black monitor, black foreground on white\n"
"                                      background.\n"
"                          u,   user : user's monochrome configuration.\n"
"                          c, colour : colour monitor.\n"
"\n"
"                          Note: This option by itself does not force the\n"
"                          emulation into a standard model Microbee, it's use\n"
"                          simply determines what monitor type is connected\n"
"                          to the emulated Microbee.\n"
"\n"
"  --mon-bg-x=level        Set the 3 user customised monochrome background\n"
"                          colours.  x is the gun colour ('r', 'g', 'b').  The\n"
"                          level value is 0-255.\n"
"  --mon-bgi-x=level       Set the 3 user customised monochrome dual intensity\n"
"                          background colours.  x is the gun colour\n"
"                          ('r', 'g', 'b').  The level value is 0-255.  This\n"
"                          option is only for the Premium (alpha+) models for\n"
"                          dual intensity monochrome (see --dint).\n"
"  --mon-fg-x=level        Set the 3 user customised monochrome foreground\n"
"                          colours.  x is the gun colour ('r', 'g', 'b').  The\n"
"                          level value is 0-255.\n"
"  --mon-fgi-x=level       Set the 3 user customised monochrome dual intensity\n"
"                          foreground colours.  x is the gun colour\n"
"                          ('r', 'g', 'b').  The level value is 0-255.  This\n"
"                          option is only for the Premium (alpha+) models for\n"
"                          dual intensity monochrome (see --dint).\n"
"\n"
"  --rgb-nn-x=level        48 options to customise the Premium (alpha+) colours.\n"
"                          nn is the colour value (00-15), x is the gun colour\n"
"                          ('r', 'g', 'b').  The level value is 0-255.\n"
"\n"
"  --video=x               Video initial start state. x=on to enable, x=off to\n"
"                          to disable.  Default is enabled.\n"
"\n"
"  --video-depth=x         Video depth. Default is 16 bits per pixel. Other\n"
"                          depths may improve or degrade performance, i.e. sound\n"
"                          quality. These values only apply to SDL rendering. x\n"
"                          may be one of the following:\n"
"\n"
"                            8 : 8 bit colour.\n"
"                          8gs : 8 bit grey scale.\n"
"                           16 : 16 bit colour.\n"
"                           32 : 32 bit colour.\n"
"\n"
"  --video-type=type       Video type. The default type used is SDL hw rendering\n"
"                          Other types may improve or degrade performance, i.e.\n"
"                          sound quality. <type> may be one of the following:\n"
"\n"
"                          gl : OpenGL (textured) hardware rendering.\n"
"                          hw : SDL hardware rendering.\n"
"                          sw : SDL software rendering.\n"
"\n"
#ifdef USE_OPENGL
// +++++++++++++++++++++++++++++ OpenGL rendering ++++++++++++++++++++++++++++++
" OpenGL rendering:\n\n"
"  --gl-aspect-bee=n       The aspect value you want for the Microbee display.\n"
"                          This default value is 4:3 aspect (1.333) but may be\n"
"                          changed with this option.  Don't use this to fix\n"
"                          monitor aspects, use --gl-aspect-mon for that.\n"
"                          Use a floating point value for n.  i.e. 4:3 aspect\n"
"                          is 4/3=1.333.\n"
"\n"
"  --gl-aspect-mon=n       This option overrides the monitor aspect ratio worked\n"
"                          out by the emulator. This should not be needed for\n"
"                          LCD monitors running in native resolution. The value\n"
"                          may be required if running a 4:3 CRT monitor with\n"
"                          non 4:3 resolution. Use a floating point value for n.\n"
"                          i.e. 4:3 aspect is 4/3=1.333.\n"
"\n"
"  filter options:         The OpenGL filter settings provide sharp or soft\n"
"                          display rendering.  One is provided for each display\n"
"                          mode.  The value for the current mode can be toggled\n"
"                          with the EMUKEY+F hot keys.  The values allowed are:\n"
"\n"
"                          sharp : sharp display.\n"
"                           soft : soft display.\n"
"\n"
"  --gl-filter-fs=x        filter setting for full screen mode. (sharp)\n"
"  --gl-filter-max=x       filter setting for maximised window mode. (sharp)\n"
"  --gl-filter-win=x       filter setting for resizable window mode. (soft)\n"
"\n"
"  --gl-max=x              Start up maximised if x=on, if x=off then start up in\n"
"                          a window or full screen mode depending on the use of\n"
"                          --fullscreen option. Default is off. This option is\n"
"                          currently not supported on Windows machines.\n"
"\n"
"  --gl-vsync=x            Vsync: swap buffers every n'th retrace. x=off to\n"
"                          disable, x=on to enable.  default is enabled.\n"
"\n"
"  --gl-winpct=n           The default window size on start up determined by a\n"
"                          percentage value from 5-100\% of the desktop X\n"
"                          resolution.  If this option is not used the window\n"
"                          X size is 50\% of the desktop width.\n"
"\n"
"  --gl-winpix=n           The default window size on start up determined by\n"
"                          the number of pixels between 50 and the desktop X\n"
"                          resolution.  If this option is not used the window\n"
"                          X size is 50\% of the desktop width.\n"
"\n"
#endif
// ++++++++++++++++++++++++++++++ File related +++++++++++++++++++++++++++++++++
" File related:\n\n"
"                          The file options are for use by the uBee512 support\n"
"                          tools.  How these are used is entirely dependent on\n"
"                          the Z80 application accessing these values.  See the\n"
"                          TOOLS.TXT file for detailed usage information.\n"
"\n"
"  --file-app=name         String name of up to 255 characters. Default is an\n"
"                          empty string.\n"
"  --file-exec=n           Z80 address 0-65535 (0000-FFFF hex). Default is 0.\n"
"  --file-list=files       Host file path(s) of up to 255 characters. @UBEE512@\n"
"                          variable if used will have double quotation\n"
"                          characters added if any space characters are found.\n"
"                          Default is an empty string. This option may be used\n"
"                          repeatedly to build up an array of strings.\n"
"  --file-list-q=files     Same as --file-list option except the entire string\n"
"                          will have double quotation characters added. No\n"
"                          double quotation characters will be placed around\n"
"                          the @UBEE512@ variable. This option may be used\n"
"                          repeatedly to build up an array of strings.\n"
"  --file-load             Z80 address 0-65535 (0000-FFFF hex). Default is 0.\n"
"  --file-run=name         String name of up to 255 characters. Default is an\n"
"                          empty string.\n"
"  --file-exit=x           Enables/disables the state of the flag value, x=on\n"
"                          to enable, x=off to disable. Default is enabled.\n"
"\n"
// +++++++++++++++++++++++++++ Information output ++++++++++++++++++++++++++++++
" Information output:\n\n"
"  -h, --help, --usage     Display help information on command line usage.\n"
"\n"
"  --conio                 Switches on verbose console output for Windows port.\n"
"                          By default only fatal errors and some option's output\n"
"                          is sent to the console.\n"
"\n"
"  --lcon                  List the [section] names found in the configuration\n"
"                          file.\n"
"  --lconw                 Same as --lcon option except uses a wide format.\n"
"  --lcons=n               Sets the list start point for --lcon and --lconw\n"
"                          options. Default value is 1.\n"
"\n"
"  --version               Obtain the version number of uBee512 and other\n"
"                          components (z80, SDL) being used.\n"
"\n"
// +++++++++++++++++++++++++++ Joystick emulation ++++++++++++++++++++++++++++++
" Joystick emulation:\n\n"
"  --js=n                  Joystick number to use, n=0 for first joystick. n=-1\n"
"                          to disable an existing setting.\n"
"\n"
"  --js-axis=x             Joystick axis mapping to buttons.  x=on to enable,\n"
"                          x=off to disable. Default is enabled.\n"
"  --js-axisb=n            Joystick axis buttons base number.  The axis button\n"
"                          offsets are 0=up, 1=right, 2=down and 3=left.  The\n"
"                          base number n is added to the offsets to generate a\n"
"                          button number. n may be any value from 0 to 124.\n"
"                          Default value is 0.\n"
"  --js-axisl=n            Determines the thresh hold level for button\n"
"                          detection.  n may be any value from 1 to 32767.\n"
"                          Default value is 3200.\n"
"  --js-clear              Clear all Microbee joystick button settings.\n"
"  --js-mbee=x             Microbee joystick emulation control. x=on to enable,\n"
"                          x=off to disable.  Default is enabled.\n"
"  --js-up=n               Button number for 'up', default is 0\n"
"  --js-right=n            Button number for 'right', default is 1\n"
"  --js-down=n             Button number for 'down', default is 2\n"
"  --js-left=n             Button number for 'left', default is 3\n"
"  --js-fire=n             Button number for 'fire', default is 4\n"
"  --js-play1=n            Button number for 'player 1', default is 5\n"
"  --js-play2=n            Button number for 'player 2', default is 6\n"
"  --js-spare=n            Button number for 'spare', default is 7\n"
"\n"
"  --js-clist              List the command names available for joystick mapping.\n"
"  --js-klist              List the key names available for joystick mapping.\n"
"  --js-kbd=x              Joystick to Microbee keys mapping control.  x=on to\n"
"                          enable, x=off to disable. Default is disabled except\n"
"                          when a --js-kset or --js-ksel option is used.\n"
"  --js-kk=k               Microbee key to be mapped to a joystick button. See\n"
"                          --js-klist option for more information.\n"
"  --js-kb=n               Button n to be associated with last --js-kk=k option.\n"
"  --js-ksel=n             Select the joystick key set to use. n may be a value\n"
"                          of (0-255) Alternatively specify n as a character\n"
"                          from A-Z, the letter will be converted to numbers\n"
"                          0-25.  Default selection is determined by --js-kset.\n"
"  --js-kset=n             Place the joystick keys into set n and make active.\n"
"                          There are 256 sets of joystick keys (0-255) that may\n"
"                          be used. Alternatively specify n as a character from\n"
"                          A-Z, the letter will be converted to numbers 0-25.\n"
"\n"
// ++++++++++++++++++++++++++++ Model emulation ++++++++++++++++++++++++++++++++
" Model emulation:\n\n"
"                          See 'File path searching' further on for detailed\n"
"                          information.  The default area for roms is:\n"
"\n"
"                          @UBEE512@\\roms\\\n"
"\n"
"  --basic=file            Used for defining 4, 8, and 16K BASIC ROM parts.\n"
"  --basica=file           Same as --basic option.\n"
"  --basicb=file           Used for 4, 8K and ppc85/tterm ROM part B.\n"
"  --basicc=file           Used for 4K ROM part C.\n"
"  --basicd=file           Used for 4K ROM part D.\n"
"\n"
"                          These --basicx options allows the ROM file image\n"
"                          specified to be used instead of the built in model\n"
"                          defaults for the BASIC ROM(s).\n"
"\n"
"  --basram                The memory locations 0xA000-0xBFFF for ROM based\n"
"                          models defaults to ROM.  This option will force this\n"
"                          8K area to SRAM emulation.  The contents will contain\n"
"                          a typical SRAM pattern on start up,  it will not be\n"
"                          associated with a ROM image file.\n"
"\n"
"  --charrom=file          Allows the character ROM file image specified to be\n"
"                          used instead of the built in predefined 'charrom.bin'\n"
"                          ROM.\n"
"\n"
"  --col                   Enables colour emulation for standard models. This\n"
"                          option has no affect when emulating a Premium, 256TC\n"
"                          or Teleterm model.  Defaults to an RGBrgb digital\n"
"                          monitor type when used.\n"
"\n"
"  --col-type=n            Same as --col option except the monitor type may be\n"
"                          selected. n=0 selects RGB analogue, and n=1 selects\n"
"                          RGBrgb monitor emulation. Default is disabled.  This\n"
"                          option has no affect when emulating a Premium, 256TC\n"
"                          or Teleterm model.\n"
"\n"
"  --colprom=file          Use the file values to override the internal 82s123\n"
"                          IC 7 standard colour values. This option has no\n"
"                          affect when emulating a Premium, 256TC or Teleterm\n"
"                          model.\n"
"\n"
"  --dint=x, --hint=x      Dual intensity monochrome emulation for Premium\n"
"                          (alpha+) models.  x=on to enable, x=off to disable.\n"
"                          This is set to 'on' by default for the 256TC and\n"
"                          upgraded Premium models.  This option has no\n"
"                          affect when emulating a standard model.  The --hint\n"
"                          option name should not be used any more.\n"
"\n"
"  --hardware=x            Disables emulation of various hardware sections\n"
"                          allowing variations in models.  Arguments commence\n"
"                          with a '-' character. Additional arguments may be\n"
"                          declared in the same --hardware option.\n"
"\n"
"                          The arguments supported are:\n"
"                          -wd2793 - disable WD2793 emulation.\n"
"\n"
"  --hwflash=x             Hardware inverse and flashing video emulation for\n"
"                          Premium (alpha+) models using one of two possible\n"
"                          methods.  x=on to enable (v4), x=off to disable, x=v3\n"
"                          enables using Premium version 3 main board flashing.\n"
"                          x=v4 may also be used instead of x=on for Premium\n"
"                          version 4 main board, 256TC and Teleterm models.\n"
"                          This is set to 'on' (v4) by default for the 256TC\n"
"                          and upgraded Premium models.  This option has no\n"
"                          affect when emulating a standard model.\n"
"\n"
"  --hwflashr=x            Set the video hardware flashing rate for Premium\n"
"                          (alpha+) models, the true rate is determined by\n"
"                          CRTC 6545 values (VSYNC).  The default value is 320\n"
"                          milliseconds. x may be one of the following timer\n"
"                          values: 20, 40, 80, 160, 320, 640, 1280 or 2560\n"
"                          milliseconds.  Alternatively a link setting may be\n"
"                          used, v3 and v4 boards differ for W63 and W64:\n"
"\n"
"                          Link  v3    v4   (milliseconds)\n"
"                          w61   160   160\n"
"                          w62   320   320\n"
"                          w63   1280  640\n"
"                          w64   640   1280\n"
"\n"
"  --lmodel                List the available model types.\n"
"\n"
"  --lpen                  Enables the use of the 6545 Light pen keys emulation.\n"
"                          This is enabled by default for all models except for\n"
"                          the 256tc and Teleterm models.\n"
"\n"
"  --model=type            Model type, if this option is not specified the p512k\n"
"                          model is emulated.  The 'p' denotes a Premium\n"
"                          variation in a model.  This option should be used\n"
"                          before any other options.  The following models are\n"
"                          supported: 256tc, p512k, 512k, p256k, 256k, p128k,\n"
"                          128k, p64k, 64k, 56k, tterm, ppc85, pc85b, pc85, pc,\n"
"                          ic, 2mhz, 2mhzdd, dd, scf and pcf.\n"
"\n"
"  --mono                  Disables colour circuit emulation for standard models.\n"
"                          This option when emulating a Premium, 256TC or\n"
"                          Teleterm model has the same affect as --monitor=g.\n"
"\n"
"  --netram                The memory locations 0xE000-0xEFFF for ROM based\n"
"                          models defaults to ROM.  This option will force this\n"
"                          4K area to SRAM emulation.  The contents will contain\n"
"                          a typical SRAM pattern on start up,  it will not be\n"
"                          associated with a ROM image file.\n"
"\n"
"  --netrom=file           Allows the ROM file image specified to be used\n"
"                          instead of the built in model defaults for the NET\n"
"                          ROM.\n"
"\n"
"  --pak(n)=file           The --pak(n) options allows the ROM file image\n"
"                          specified to be used instead of the built in model\n"
"                          default for the PAKn ROM 0 to 7 locations.\n"
"\n"
"  --pakram=n              The PAK n location will use SRAM instead of ROM\n"
"                          emulation.  The contents will contain a typical SRAM\n"
"                          pattern on start up,  it will not be associated with\n"
"                          a ROM image file. n may be any PAK number from 0-7.\n"
"\n"
"  --pcg=n                 Premium (alpha+) model option that sets size of PCG\n"
"                          RAM to be emulated.  n is the size of the PCG memory\n"
"                          in Kilobytes and can be any even value between 2 and\n"
"                          32. 256TC, Premium and Teleterm models are 16K,\n"
"                          upgraded Premium models are 32K.  This option has no\n"
"                          affect when emulating a standard model.\n"
"\n"
"  --piob7=signal          Determines what signal is used for PIO port B bit 7.\n"
"                          The default value depends on the model emulated.\n"
"                          The source values are: rtc, vsync, net, and pup.\n"
"\n"
"  --rom1=file             Allows the ROM file image specified to be used\n"
"                          instead of the built in model default for the boot\n"
"                          ROM. This ROM is only used by all FDC models.\n"
"  --rom2=file             Allows the ROM file image specified to be used\n"
"                          instead of the built in model default for ROM2. This\n"
"                          ROM is used by all DRAM FDC models, except for the\n"
"                          256TC.\n"
"  --rom3=file             Allows the ROM file image specified to be used\n"
"                          instead of the built in model default for ROM3. This\n"
"                          ROM is used by all DRAM FDC models, except for the\n"
"                          256TC.\n"
"\n"
"  --rom256k=file          Allows the 256K ROM file image specified to be used\n"
"                          instead of the built in model default for the 256K\n"
"                          ROM.  This ROM is used by some 3rd party designs.\n"
"                          Set file to 'none' to disable the 256K ROM image.\n"
"\n"
"  --sys=name              Defines a system name,  this will be appended to some\n"
"                          files so that different operating systems using the\n"
"                          same model can still have unique names for certain\n"
"                          files.  By default this name contains nothing.  An\n"
"                          example is the loading and saving of RTC values.  By\n"
"                          default for a p128k model this would be 'p128k.rtc',\n"
"                          by defining name to be 'mysys' would then use\n"
"                          'p128k-mysys.rtc' for the file, the emulator inserts\n"
"                          the hyphen character.\n"
"\n"
"  --vdu=n                 Premium (alpha+) model option that sets size of VDU\n"
"                          RAM to be emulated.  n may be 2 or 8.  The VDU RAM\n"
"                          size determines the number of screen, attribute and\n"
"                          colour RAM banks.  Default is 2K.  Don't use this\n"
"                          option to increase to 8K unless you understand the\n"
"                          problems associated with it.\n"
"\n"
// +++++++++++++++++++++++ Microbee mouse emulation ++++++++++++++++++++++++++++
" Microbee mouse emulation:\n\n"
"  --mouse=x               Microbee mouse emulation. x=on to enable Microbee\n"
"                          mouse emulation on start-up. Default is 'off'.\n"
"\n"
// ++++++++++++++++++++ Parallel printer port emulation ++++++++++++++++++++++++
" Parallel printer port emulation:\n\n"
"                          See 'File path searching' further on for detailed\n"
"                          information.  The default area for printer is:\n"
"\n"
"                          @UBEE512@\\printer\\\n"
"\n"
"  --print=file            Printer output to file,  the output is not modified.\n"
"                          If an open printer file is already in use then that\n"
"                          file will be closed first before creating the new\n"
"                          printer file.\n"
"  --print-close           Closes a currently open printer file. This allows the\n"
"                          file to be accessed externally without exiting the\n"
"                          emulator.\n"
"\n"
"  --printa=file           Printer output to file,  the output is converted to\n"
"                          ASCII decimal.  If an open printer file is already in\n"
"                          use then that file will be closed first before\n"
"                          creating the new printer file.\n"
"  --printa-close          Closes a currently open ASCII printer file. This\n"
"                          allows the file to be accessed externally without\n"
"                          exiting the emulator.\n"
"\n"
// ++++++++++++++++ Real Time Clock (RTC) emulation and time +++++++++++++++++++
" Real Time Clock (RTC) emulation and time:\n\n"
"  --century=n             This value can be used to correct the century date\n"
"                          used in ROMs.  Changes are temporary and are made to\n"
"                          the image(s) in memory only.  An MD5 value will be\n"
"                          used to identify what loaded ROM images need\n"
"                          modifying.  The value of n is expected to be in BCD\n"
"                          format.  The century '20' should use n=0x20 (hex) or\n"
"                          n=32 (dec).  The following ROMs with matching MD5s\n"
"                          will be modified in memory:\n"
"\n"
"                          256TC v1.15: md5=13ddba203bd0b8228f748111421bad5f\n"
"                          256TC v1.20: md5=24d6682ff7603655b0cbf77be6731fb0\n"
"\n"
"  --rtc=n                 Real Time Clock (RTC) emulation. n=1 to enable, n=0\n"
"                          to disable.  The following models use RTC by\n"
"                          default: 256tc, p512k, 512k, p256k, 256k.\n"
"\n"
// +++++++++++++++++++++++++ Serial port emulation +++++++++++++++++++++++++++++
" Serial port emulation:\n\n"
"  --baud=rate             Set serial communications baud rate for both TX and\n"
"                          RX.  A value from 1 to 19200 is allowed.  Default\n"
"                          rate is 300 baud.  If Individual baud rates are\n"
"                          required for TX and RX then use the --baudtx and\n"
"                          --baudrx options instead.  This value must match the\n"
"                          Microbee serial application's value.\n"
"\n"
"  --baudtx=rate           Set serial communications baud rate for TX only.  A\n"
"                          value from 1 to 19200 is allowed.  Default rate is\n"
"                          300 baud.  This value must match the Microbee serial\n"
"                          application's value.\n"
"\n"
"  --baudrx=rate           Set serial communications baud rate for RX only.  A\n"
"                          value from 1 to 19200 is allowed.  Default rate is\n"
"                          300 baud. This value must match the Microbee serial\n"
"                          application's value.\n"
"\n"
"  --coms=port             Serial communications port for emulation of RS232.\n"
"                          On Unices specify a device, on Windows specify the\n"
"                          com port number.  No serial communications will be\n"
"                          emulated if this option is not specified.  If a\n"
"                          serial port is already open then that port will be\n"
"                          closed first before opening a new serial port.\n"
"\n"
"  --coms-close            Closes the RS232 serial port if currently open.\n"
"\n"
"  --datab=bits            Set serial communications number of data bits.  A\n"
"                          value from 5 to 8 is allowed.  Default value is 8\n"
"                          data bits.  This value must match the Microbee\n"
"                          serial application's value.\n"
"\n"
"  --stopb=bits            Set serial communications number of stop bits.  A\n"
"                          value from 1 to 2 is allowed.  Default value is 1\n"
"                          data bits.  This value must match the Microbee\n"
"                          serial application's value for TX.\n"
"\n"
// +++++++++++++++++++++++++++ Sound emulation +++++++++++++++++++++++++++++++++
" Sound emulation:\n\n"
"  --sound=method          Determine the method used for sound,  the default\n"
"                          method is 'prop':\n"
"\n"
"                             off: sound is turned off\n"
"                            prop: sound is proportional to CPU clock frequency\n"
"                          normal: sound rate forced as if 3.375 MHz CPU clock\n"
"\n"
"  --snd-alg1=x            Sound algorithm to prevent sound gaining in time due\n"
"                          to integer remainder losses from division operations.\n"
"                          x=on to enable, x=off to disable.  Default is off.\n"
"  --snd-freq=f            Set the sound sampling rate, f may be a value from\n"
"                          5512 to 176400 Hz.  Default frequency is 44100 Hz.\n"
"  --snd-freqadj=n         Adjusts the sound card sampling rate value. n is a\n"
"                          floating point percentage value from +/- 1 to 50 that\n"
"                          is added to the sound frequency to keep it just under\n"
"                          the speed of the Z80 CPU emulation.  Default is -2.5\%\n"
"  --snd-freqlow=f         Sets the lowest sound frequency that can be produced.\n"
"                          f may be 1 to 20000Hz.  Default is 20Hz.\n"
"  --snd-holdoff=n         Number of milliseconds hold-off delay to apply each\n"
"                          time a new stream of sound bits is started. n may be\n"
"                          0 to 10000. Default is 50 milliseconds.\n"
"  --snd-hq                Sets high quality sound.  How well this works will be\n"
"                          dependent on the host platform.  This option has the\n"
"                          same affect as setting all these values:\n"
"                            --snd-holdoff=50\n"
"                            --snd-samples=2048\n"
"                            --snd-freq=88200\n"
"                            --snd-freqlow=10\n"
"                            --snd-freqadj=-0.25\n"
"                            --snd-alg1=on\n"
"  --snd-mute=x            Sound mute, use to start the emulator with the sound\n"
"                          muted until enabled. x=on to enable, x=off to\n"
"                          disable.  Default is off.\n"
"  --snd-samples=n         Sets the SDL callback data size. n must be a power of\n"
"                          2.  Values from 1 to 32768 are allowed.  Default is\n"
"                          1024 samples.\n"
"  --snd-volume=l --vol=l  Set the sound volume level.  A level of 0 to 100\% is\n"
"                          allowed.  Default is 15\%.\n"
"\n"
// ++++++++++++++++++++++++++++ Speed related ++++++++++++++++++++++++++++++++++
" Speed related:\n\n"
"  --clock=f               Set the Z80 clock frequency for emulation in MHz.\n"
"                          Standard emulation frequencies are 3.375 and 2.0\n"
"                          MHz.  All other frequencies are classed as 'hacking'.\n" 
"                          3.375 MHz is the default clock frequency.  Frequency\n"
"                          is entered as a floating point number.\n"
"\n"
"                          The highest frequency possible is determined by the\n"
"                          host platform.  As the value increases to the point\n"
"                          where uBee512 can no longer regulate the Z80\n"
"                          execution rate the frame rate will decrease (slower\n"
"                          screen update periods).\n"
"\n"
"  --clock-def=f           Set the Z80 clock frequency for emulation in MHz when\n"
"                          the uBee512 API restore function is called.  Default\n"
"                          frequency is 3.375 MHz.\n" 
"\n"
"  --frate=fps             Frame rate, an integer value between 1 and 1,000,000\n"
"                          is allowed.  Default is 50 FPS.\n"
"\n"
"  --maxcpulag=n           This is the maximum time the Z80 CPU emulation is\n"
"                          allowed to get behind before 'catch up' is bypassed\n"
"                          for the currently lagged cycles.  A very high value\n"
"                          for n will cause the 'catch up' mode to always be in\n"
"                          affect,  using a value of 0 for n will effectively\n"
"                          disable this feature and act like 2.5.0 and earlier\n"
"                          versions. Default value is 250ms.\n"
"\n"
"  --speedsel=n            CPU speed selection emulation. n=1 to enable, n=0\n"
"                          to disable.  The following models are enabled by\n"
"                          default: 256tc, p512k, 512k, p256k, 256k.\n"
"\n"
"  -t, --turbo[=x]         Turbo mode, executes Z80 code as fast as possible.\n"
"                          Without this option the emulation attempts to keep\n"
"                          Z80 CPU execution to match the CPU clock value.\n"
"                          If 'x' is specified then turbo mode can be set with\n"
"                          x=on or off with x=off.  This option is intended for\n"
"                          'hacking' and code development use.  There are much\n"
"                          faster methods if more speed is required. (see the\n"
"                          README file)\n"
"\n"
"  --vblank=method         Vertical blanking method to be employed. This is\n"
"                          only intended for 'hacking' when experimenting with\n"
"                          turbo mode and/or high CPU clock speeds.  It is not\n"
"                          required or even recommended to be used if 3.375 MHz\n"
"                          or 2 MHz emulation is desired.\n"
"\n"
"                          0 : 50 Hz VBLANK rate derived from Z80 cycles and is\n"
"                              proportional to the CPU clock frequency.\n"
"                          1 : 50 Hz VBLANK rate derived from the host timer.\n"
"\n"
"                          When running in turbo mode then setting the <method>\n"
"                          equal to 1 will ensure that a VBLANK rate of 50Hz\n"
"                          will be used.  Without this, key repeating may be\n"
"                          too fast.\n"
"\n"
"  -x, --xtal=f            Old non preferred options to set the Z80 clock\n"
"                          frequency.  Use --clock option instead.\n"
"\n"
"  --z80div=n              Determines the number of Z80 blocks emulated per z80\n"
"                          frame.  This value allows the polling rate to be\n"
"                          increased or decreased.  The polling rate per second\n"
"                          is the product of the frame rate (--frate) and this\n"
"                          value. The value of n may range from 1 to 5000. On\n"
"                          versions prior to 2.7.0 this value was 1. Default\n"
"                          value is 25.\n"
"\n"
// +++++++++++++++++++++++++ Tape port emulation +++++++++++++++++++++++++++++++
" Tape port emulation:\n\n"
"                          See 'File path searching' further on for detailed\n"
"                          information.  The default area for tapes is:\n"
"\n"
"                          @UBEE512@\\tapes\\\n"
"\n"
"  --tapei=file            Tape input from a WAV file.  If an open tape input\n"
"                          file is already in use then that file will be closed\n"
"                          first before opening the new tape input file.\n"
"  --tapei-close           Closes a currently open tape input file. This allows\n"
"                          the file to be accessed externally without exiting\n"
"                          the emulator.\n"
"  --tapei-det=value       Optional tape input detection value.  This value is\n"
"                          used in place of the internally set level if non\n"
"                          zero.  Default value is 0.\n"
"\n"
"  --tapeo=file            Tape output to a WAV file.  If an open tape output\n"
"                          file is already in use then that file will be closed\n"
"                          first before creating the new tape output file.\n"
"  --tapeo-close           Closes a currently open tape output file. This allows\n"
"                          the file to be accessed externally without exiting\n"
"                          the emulator.\n"
"\n"
"  --tapesamp=frequency    Tape output sample frequency in Hz.  Default is\n"
"                          22050 Hz.\n"
"\n"
"  --tapevol=level         Tape output wave file volume level.  A level of 0 to\n"
"                          100\% is allowed.  Default is 15\%.\n"
"\n"
// ++++++++++++++++++++++++ Preconfigured variables ++++++++++++++++++++++++++++
" Variables:\n\n"
"                          The --varset option may be used to create variables.\n"
"                          These variables are pre-configured:\n"
"\n"
"  UBEE_USERHOME           User's home path on Unices, or the directory\n"
"                          containing the ubee512 exe on Windows systems.\n"
"  UBEE512 or ubee512      Path to the ubee512 account.\n"
"  UBEE_VERSION            Emulator version.\n"
"  UBEE_HOST               Host system (UNIX or WIN).\n"
"  UBEE_SYS_MAJOR          On Unix systems this will be 'UNIX' On Windows\n"
"                          systems it contains one of the following:\n"
"                          win9x_me, nt4, nt5 or nt6.  Both systems will use\n"
"                          upper case for the variable value.\n"
"  UBEE_SYS_MAJOR_VAL      Windows OSVERSIONINFOEX dwMajorVersion member value.\n"
"  UBEE_SYS_MINOR          On Unix systems this will be the value of the\n"
"                          uname.sysname member.  On Windows systems it contains\n"
"                          one of the following: w95, w98, me, nt4_ws,\n"
"                          nt4_server, w2000, xp, server_2003, or vista. Both\n"
"                          systems will use upper case for the variable value.\n"
"  UBEE_SYS_MINOR_VAL      Windows OSVERSIONINFOEX dwMinorVersion member value.\n"
"\n"
" Time and Date            These variables are normally used to form other\n"
"                          variables, see TD variable usage in 'ubee512rc'.\n"
"\n"
"  SS                      Seconds. (00-59)\n"
"  MM                      Minutes. (00-59)\n"
"  HH                      Hours. (00-23) \n"
"  DD                      Month day. (01-2x/3x) \n"
"  mm                      Month. (01-12) \n"
"  YYYY                    4 digit year. (1900-20xx)\n"
"  YY                      2 digit year. (00-99)\n"
"  ww                      Week day Sun-Sat. (0-6)\n"
"  ac                      Week day capitalised. (Sun-Sat)\n"
"  al                      Week day lower case. (sun-sat)\n"
"  au                      Week day upper case. (SUN-SAT)\n"
"\n"
// ++++++++++++++++++++++++++ General information ++++++++++++++++++++++++++++++
" File path searching:\n\n"
"  Path slash characters\n"
"  ---------------------\n"
"  Forward or back slashes may be used in file paths irrespective of the\n"
"  program being run under Unices or Windows environments when slash\n"
"  conversion is enabled. See --slashes option. Unices will see '\\' as an\n"
"  escape sequence when used on the command line and also when found in\n"
"  configuration files and slash conversion is disabled.\n"
"\n"
"  Files to be opened\n"
"  ------------------\n"
"  Existing files will first be searched for in the current directory, if the\n"
"  path is not found a second search in the default directory will take place.\n"
"  For the second search the file path specified will be appended to the default\n"
"  directory path.  The second search is not carried out if a '\\', '.\\', or\n"
"  '..\\' are the first characters of the path or a ':' character is used under\n"
"  Windows.\n"
"\n"
"  Files to be created\n"
"  -------------------\n"
"  Files to be created will be placed into the default directory unless a path\n"
"  to another location is specified by using a '\\', '.\\', or '..\\' as the\n"
"  first characters of the path or a ':' character is used under Windows.\n"
"\n"
// +++++++++++++++++++++++++++++++++ other +++++++++++++++++++++++++++++++++++++
"If you have any new feature suggestions, bug reports, etc. then please send\n"
"an email to: ubee512@gmail.com\n";

#ifdef MINGW
#define page_help 1
#else
#define page_help 0
#endif
 
 while (1)
    {
     switch (help->state)
        {
         case 0 :
            if ((! page_help) && (! emu.runmode))
               {
                printf("%s", usage); // don't use xprintf() here!
                help->state = -1;    // done
                return;
               }

            if (console.xstdin == 0)
               help->lw = 8;
            else
               help->lw = 24;
            help->index = 0;
            help->lineswanted = help->lw;
            help->state++;
            break;
         case 1 :
            while (1)
               {
                if (! help->lineswanted) // done if no more lines
                   {
                    help->state = -1;
                    return;
                   } 
                ch = (usage[help->index]);
                 
                if (ch != '\n')
                   {
                    xprintf("%c", ch);
                    help->index++;
                   }
                else
                   {
                    help->index++;	// get past the \n character
                    xprintf("\n");

                    if (! usage[help->index])
                       {
                        if (console.xstdin)
                           {
                            xprintf("\r===== End of help information ================================== ESC=done =====");
                            xflush();
                            help->state = 3;
                            return;
                           }
                        else
                           {
                            xprintf("===== End of help information =====\n");
                            help->state = -1;
                            return;                           
                           }   
                       }
                       
                    if (--help->lineswanted == 0)
                       {                 
                        if (console.xstdin)
                           {
                            xprintf("\r===== ENTER=next line ============ SPACE=next screen =========== ESC=done =====");
                            xflush();
                           } 
                        help->state++;
                        return;
                       }
                   }
               }
            break;
         case 2 : // get the ENTER, SPACE or ESC key
            if (console.xstdin)
               ch = getch();
            else
               ch = osd_getkey();
            if ((ch != 13) && (ch != 32) && (ch != 27))
               return;

            if (console.xstdin)
               {
                xprintf("\r                                                                               \r");
                xflush();
               }
                
            switch (ch)
               {
                case 13 : help->lineswanted = 1;
                          break;
                case 32 : help->lineswanted = help->lw;
                          break;
                case 27 : help->lineswanted = 0;
                          if (! console.xstdin)
                             {
                              xprintf("===== User exited help =====\n");
                              help->state = -1;
                              return;                           
                             }
                          break;
               }
            help->state = 1;
            break;
        case 3 : // get the ESC key to exit (used by stdin only)
           while (getch() != 27)
              ;
           help->state = -1;
           xprintf("\r                                                                               \r");
           xflush();
           return;
        }
    }
}

//==============================================================================
// Options usage
//
// The options usage information uses a a state machine.  This function is
// called when the 'options usage' option is processed.
//
// If the OSD console dialogue is active then only the first machine state
// is executed and any further states are called from the OSD module using
// the options_usage_state() function.
//
//   pass: void
// return: void
//==============================================================================
static void options_usage (void)
{
 help.state = 0;

 if (console.xstdin == 0)
    options_usage_state(&help);
 else
    {
     while (help.state != -1)
        options_usage_state(&help);
    }
}

//==============================================================================
// Parameter error message.
//
//   pass: char *parameter
//         char *options_arg
// return: void
//==============================================================================
static void param_error_mesg (char *parameter, char *options_arg)
{
#define PARMERR_MESG "ubee512: option %s argument of '%s' is not permitted\n"
 xprintf(PARMERR_MESG, parameter, options_arg);
}

//==============================================================================
// set a uBee512 environment variable.
//
// The variable uses the format 'variable_name=variable_value'
//
//   pass: s *char                      uBee512 environment variable name/value
// return: int                          0 if no errors, else -1
//==============================================================================
int options_ubee512_envvar_set (char *s)
{
 int l;

 l = strlen(s) + 1;
 
 if (emuenvc < OPTIONS_ENV_SIZE)
    {
     if (! (emuenv[emuenvc] = options_malloc(l)))
        return -1;
     strcpy(emuenv[emuenvc++], s);
     return 0;
    }
 else
    return -1;   
}

//==============================================================================
// unset a uBee512 environment variable.
//
//   pass: s *char                      uBee512 environment variable name
// return: void
//==============================================================================
static void options_ubee512_envvar_unset (char *s)
{
 int i;
 char *x;

 for (i = 0; i < emuenvc; i++)
    {
     x = strstr(emuenv[i], s);
     if (x == emuenv[i])
        {
         free(emuenv[i]);   // free the memory allocated to it
         emuenv[i] = NULL;  // incase we only have one pointer
         emuenvc--;         // one less variable now
         if (emuenvc)       // if 1 or more than move pointers up
            memmove(&emuenv[i], &emuenv[i + 1], sizeof(emuenv) * (emuenvc - i));
         return;   
        }
    }
}

//==============================================================================
// Return a uBee512 environment variable.
//
// Searches a list of variables and returns a pointer to the value if the name
// matches.  A NULL is returned if no matches were found.
//
// The variable uses the format 'variable_name=variable_value'
//
//   pass: s *char                      uBee512 environment variable name
// return: char *                       pointer to variable else NULL
//==============================================================================
static char *options_ubee512_envvar_get (char *s)
{
 int i;
 
 char *x;

 for (i = 0; i < emuenvc; i++)
    {
     x = strstr(emuenv[i], s);
     if ((x == emuenv[i]) && (emuenv[i][strlen(s)] == '='))
        return x + strlen(s) + 1;  // pointer to the string after '='
    }
 return NULL;
}    

//==============================================================================
// Compare a uBee512 'string1,string2' using a strvserscmp() or strcmp()
// function depending on the --if-cmpmode option.
//
// The name and value passed uses the format 'string1,string2'
//
//   pass: s *char                      "string1,string2" (no leading spaces)
// return: int                          0=equals, -1=less, 1=greater, -2=error
//==============================================================================
static int options_ubee512_compare (char *s)
{
 char *x;

 char temp_str1[512];
 char temp_str2[512];

 strcpy(temp_str1, s);
 x = strstr(temp_str1, ",");
 if (x)
    {
     *x = 0;
     strcpy(temp_str2, x+1);
     switch (if_cmp_mode)
        {
         case 0 : return xstrverscmp(temp_str1, temp_str2);
                  break;
         case 1 : return strcmp(temp_str1, temp_str2);
                  break;
        } 
    }
 return -2;
}

//==============================================================================
// Extract application, environment and built-in variables.
//
// Application, environment and built-in variables contained in the
// parameter are referenced by using an @ENVVAR@ format in configuration
// files.  Passing environment variables from the command line can use the
// shell's method instead.
//
// Variables may be defined using the --varset option, set as system
// environment variables or use predifined variables. The search order is as
// described so variables may be redefined as the first matching value will
// be used.
//
// Two variables types are returned, e_option_arg contains the @UBEE512@ as
// is and e_option_q_arg contains the same but with double quotation marks
// placed around the variable if any spaces are found,  this value is
// intended for passing to Z80 applications where needed.
//
// The built-in pre-defined variables are:
//      UBEE512 : path to the ubee512 account.
// UBEE_VERSION : version string of the emulator.
//    UBEE_HOST : host system.
//
// To set uBee512 environment variables using options:
// --if-system=win
//    --varset="UBEE_COM1=1"
// --if-end
// --if-system=linux
//    --varset="UBEE_COM1=/dev/ttyS0"
// --if-end
//
// To set system environment variables using the host system's shell method:
//
// Windows:
// set ENVVAR=value
//
// bash, sh:
// export ENVVAR=value
//
// csh, tsh:
// setenv ENVVAR value
//
//   pass: char *options_arg
//         char *e_options_arg		
//         char *e_options_q_arg	double quoted variable if any spaces
// return: void
//==============================================================================
static void extract_environment_vars (char *options_arg, char *e_options_arg, char *e_options_q_arg)
{
#define ENV_NAME_SIZE 1000
 char env[ENV_NAME_SIZE];

 char *ep;

 int o_save = 0;
 int o_index = 0;
 int e_index = 0;
 int e_index_q = 0;
 int v_index;

 e_options_arg[0] = 0;
 e_options_q_arg[0] = 0;

 if (options_arg == NULL)
    return;

 while ((options_arg[o_index]) && (o_index < OPTIONS_PARM_SIZE) && (e_index < OPTIONS_PARM_SIZE))
    {
     if (options_arg[o_index] != '@') // if not start of ENV string
        {
         e_options_arg[e_index++] = options_arg[o_index];
         e_options_q_arg[e_index_q++] = options_arg[o_index];
         e_options_arg[e_index] = 0;
         e_options_q_arg[e_index_q] = 0;
         o_index++;
        }
     else   
        {
         v_index = 0;
         env[0] = 0; 
         o_index++;
         o_save = o_index;
         while ((options_arg[o_index]) && (options_arg[o_index] != '@') &&
         (o_index < OPTIONS_PARM_SIZE) && (v_index < ENV_NAME_SIZE))
            env[v_index++] = options_arg[o_index++];

         env[v_index] = 0;

         if (options_arg[o_index] == '@')	// if end of ENV string
            {
             o_index++;
             int have_env_var = 0;
#ifdef MINGW
             char s[OPTIONS_PARM_SIZE]; 
             if (GetEnvironmentVariable(env, s, sizeof(s)) != 0)
                {
                 have_env_var = 1;
                 ep = s;
                }
#else
             ep = getenv(env);
             if (ep)
                have_env_var = 1;
#endif
             else
                {
                 ep = options_ubee512_envvar_get(env);
                 if (ep)
                     have_env_var = 1;
                }

             if (have_env_var)
                {
                 strcat(e_options_arg, ep);                
                 if (strchr(ep, ' '))  // if contains any spaces
                    {
                     strcat(e_options_q_arg, "\"");  // double quote
                     strcat(e_options_q_arg, ep);
                     strcat(e_options_q_arg, "\"");  // double quote
                    }
                 else
                    strcat(e_options_q_arg, ep);
                }
             e_index = strlen(e_options_arg);
             e_index_q = strlen(e_options_q_arg);
            }
         else
            {
             o_index = o_save;  // go back to index after first '@' char
             e_options_arg[e_index++] = '@';  // treat it as a normal char.
             e_options_q_arg[e_index_q++] = '@';  // treat it as a normal char.
            } 
        }
    }

 e_options_arg[e_index] = 0;
 e_options_q_arg[e_index_q] = 0;
}

//==============================================================================
// Process options: Short options (0x0000-0x00ff)
//
//   pass: int c			single option number
//         char *argv[]			pointer to pointer to arguments
// return: void
//==============================================================================
static void options_short (int c, char *argv[])
{
 switch (c)
    {
     case   '?' : 
        default : if (! opterr_msg[0])
                     {
                      exitstatus = 1;
                      xprintf("%s: unrecognised getopt error\n", argv[0]);
                      xprintf(TRY_MESG, argv[0], argv[0], argv[0]);
                     }
                  break;   
    }
}

//==============================================================================
// Process options: Control related (0x0100-0x01ff)
//
//   pass: int c			option number
//         char *argv[]
// return: void
//==============================================================================
static void options_control (int c)
{
 char *mouse_wheel_args[] =
 {
  "none",
  "vol",
  "win",  
  ""
 };

 char *cf_args[] =
 {
  "boot",
  "pc85",
  ""
 };

 int x;

 switch (c)
    {
     case OPT_ACCOUNT :
        break;
     case OPT_ALIAS_DISKS :
        x = string_search(offon_args, e_optarg);
        if (x == -1)
           {
            param_error_mesg("'--alias-disks'", e_optarg);
            exitstatus = 1;
           }
        else
           emu.alias_disks = x;
        break;
     case OPT_ALIAS_ROMS :
        x = string_search(offon_args, e_optarg);
        if (x == -1)
           {
            param_error_mesg("'--alias-roms'", e_optarg);
            exitstatus = 1;
           }
        else
           emu.alias_roms = x;
        break; 
     case OPT_ARGS_ERROR :
        snprintf(temp_str, sizeof(temp_str), "%s-", e_optarg);
        if (strstr(temp_str, "-unknown-"))
           args_err_flags = args_err_flags & (0xffffffffL ^ 0x01);
        break;
     case OPT_BOOTKEY :
        if (! e_optarg[0] || e_optarg[1])
           {
            param_error_mesg("'--bootkey'", e_optarg);
            exitstatus = 1;
           }
        else
           {
            x = toupper(e_optarg[0]);
            if (isalpha(x))
               keyb_force((x - 'A') + 1, 1);
            else
               if (isdigit(x))
                  keyb_force((x - '0') + 32, 1);
           }
        break;
     case OPT_CFMODE :
        x = string_search(cf_args, e_optarg);
        if (x == -1)
           {
            param_error_mesg("'--cfmode'", e_optarg);
            exitstatus = 1;
           }
        else
           emu.cfmode = x;
        break;
     case OPT_CONFIG :
        break;
     case OPT_CMD_REPEAT1 :
        if (int_arg < 1)
           {
            param_error_mesg("'--cmd-repeat1'", e_optarg);
            exitstatus = 1;
           }
        else
           emu.cmd_repeat1 = int_arg;
        break;
     case OPT_CMD_REPEAT2 :
        if (int_arg < 1)
           {
            param_error_mesg("'--cmd-repeat2'", e_optarg);
            exitstatus = 1;
           }
        else
           emu.cmd_repeat2 = int_arg;
        break;
     case OPT_CPU_DELAY :
        if ((int_arg < 0) || (int_arg > 2))
           {
            param_error_mesg("'--cpu-delay", e_optarg);
            exitstatus = 1;
           }
        else
           emu.proc_delay_type = int_arg;
        break;        
     case OPT_DCLICK :
        if ((int_arg < 100) || (int_arg > 3000))
           {
            param_error_mesg("'--dclick'", e_optarg);
            exitstatus = 1;
           }
        else
           gui.dclick_time = int_arg;
        break;
     case OPT_EXIT :
        exitstatus = int_arg;
        break;
     case OPT_EXIT_CHECK :
        x = string_search(offon_args, e_optarg);
        if (x == -1)
           {
            param_error_mesg("'--exit-check'", e_optarg);
            exitstatus = 1;
           }
        else
           emu.exit_check = x;
        break;
     case OPT_GUI_PERSIST :
        if (int_arg < 1)
           {
            param_error_mesg("'--gui-persist'", e_optarg);
            exitstatus = 1;
           }
        else
           gui.persist_time = int_arg;
        break;            
     case OPT_KEYSTD_MOD :
        snprintf(temp_str, sizeof(temp_str), "%s+", e_optarg);
        if (strstr(temp_str, "+clear+"))
           keystd.key_mod = 0;
        if (strstr(temp_str, "+ctrl_shift+"))
           keystd.key_mod |= KEYSTD_MOD_CTRL_SHIFT;
        break;
     case OPT_LOCKFIX_WIN32 :
        x = string_search(offon_args, e_optarg);
        if (x == -1)
           {
            param_error_mesg("'--lockfix-win32'", e_optarg);
            exitstatus = 1;
           }
        else
           emu.win32_lock_key_fix = x;
        break;
     case OPT_LOCKFIX_X11 :
        x = string_search(offon_args, e_optarg);
        if (x == -1)
           {
            param_error_mesg("'--lockfix-x11'", e_optarg);
            exitstatus = 1;
           }
        else
           emu.x11_lock_key_fix = x;
        break;            
     case OPT_MMODE :
        keyb_force(0x0d, 100);	// force into monitor mode
        break;
     case OPT_MOUSE_WHEEL :
        x = string_search(mouse_wheel_args, e_optarg);
        if (x == -1)
           {
            param_error_mesg("'--mouse-wheel'", e_optarg);
            exitstatus = 1;
           }
        else
           gui.mouse_wheel = x;
        break;            
     case OPT_NODISK :
        fdc.nodisk = 1;
        break;
     case OPT_OPTIONS_WARN :
        x = string_search(offon_args, e_optarg);
        if (x == -1)
           {
            param_error_mesg("'--options-warn'", e_optarg);
            exitstatus = 1;
           }
        else
           runmode_warn = x;
        break;            
     case OPT_OSD :
        snprintf(temp_str, sizeof(temp_str), "%s+", e_optarg);
        if (strstr(temp_str, "+clear+"))
           osd.flags = 0;
        if (strstr(temp_str, "+animate+"))
           osd.flags |= OSD_FLAG_ANIMATE;
        break;
     case OPT_OUTPUT :
        snprintf(temp_str, sizeof(temp_str), "%s+", e_optarg);
        if (strstr(temp_str, "+none+"))
           console_set_devices(0);
        if (strstr(temp_str, "+osd+"))
           console_add_device(CONSOLE_OSD);
        if (strstr(temp_str, "+stdout+"))
           console_add_device(CONSOLE_STDOUT);
        break;
     case OPT_PREFIX :
        strcpy(emu.prefix_path, e_optarg);
        break;
     case OPT_RUNSECS :
        if ((int_arg != 0) && (int_arg < 5))
           {
            param_error_mesg("'--runsecs'", e_optarg);
            exitstatus = 1;
           }
        else
           emu.secs_exit = int_arg;
        break;
     case OPT_SLASHES :
        x = string_search(offon_args, e_optarg);
        if (x == -1)
           {
            param_error_mesg("'--slashes'", e_optarg);
            exitstatus = 1;
           }
        else
           emu.slashconv = x;
        break;
     case OPT_SPAD :
        if (gui_status_padding(int_arg))
           {
            param_error_mesg("'--spad'", e_optarg);
            exitstatus = 1;
           }
        break;            
     case OPT_STATUS :
        snprintf(temp_str, sizeof(temp_str), "%s+", e_optarg);
        if (strstr(temp_str, "+clear+"))
           memset(&gui_status, 0, sizeof(gui_status));
        if (strstr(temp_str, "+d+"))
           gui_status.shortdrive = 1;
        if (strstr(temp_str, "+drive+"))
           gui_status.longdrive = 1;
        if (strstr(temp_str, "+emu+"))               
           gui_status.emu = 1;
        if (strstr(temp_str, "+emuver+"))               
           gui_status.emuver = 1;               
        if (strstr(temp_str, "+joy+"))               
           gui_status.joy = 1;
        if (strstr(temp_str, "+left+"))               
           gui_status.left = 1;               
        if (strstr(temp_str, "+model+"))               
           gui_status.model = 1;
        if (strstr(temp_str, "+mouse+"))
           gui_status.mouse = 1;
        if (strstr(temp_str, "+mute+"))               
           gui_status.mute = 1;
        if (strstr(temp_str, "+print+"))               
           gui_status.print = 1;
        if (strstr(temp_str, "+ram+"))               
           gui_status.ram = 1;
        if (strstr(temp_str, "+speed+"))               
           gui_status.speed = 1;
        if (strstr(temp_str, "+serial+"))               
           gui_status.serial = 1;
        if (strstr(temp_str, "+sys+"))               
           gui_status.sys = 1;
        if (strstr(temp_str, "+tape+"))
           gui_status.tape = 1;
        if (strstr(temp_str, "+title+"))               
           gui_status.title = 1;               
        if (strstr(temp_str, "+ver+"))
           gui_status.ver = 1;
        if (strstr(temp_str, "+vol+"))
           gui_status.vol = 1;
        if (strstr(temp_str, "+win+"))
           gui_status.win = 1;
        break;
     case OPT_TITLE :
        strncpy(gui.title, e_optarg, sizeof(gui.title));
        gui.title[sizeof(gui.title)-1] = 0;
        break;
     case OPT_VARSET :
        options_ubee512_envvar_set(e_optarg);
        break;
     case OPT_VARUSET :
        options_ubee512_envvar_unset(e_optarg);
        break;
     case OPT_VERBOSE :
        if (! e_optarg[0])
           emu.verbose = 1;           
        else
           emu.verbose = int_arg;
        break;

        // Conditional option parsing
     case OPT_IF_EGT :
        x = options_ubee512_compare(e_optarg);
        if (x == -2)
           {
            param_error_mesg("'--if-egt'", e_optarg);
            exitstatus = 1;
           }
        else            
           if_state[++if_pos] = (x >= 0) & if_state_prev;
        break;
     case OPT_IF_ELT :
        x = options_ubee512_compare(e_optarg);
        if (x == -2)
           {
            param_error_mesg("'--if-elt'", e_optarg);
            exitstatus = 1;
           }
        else            
           if_state[++if_pos] = (x <= 0) & if_state_prev;
        break;
     case OPT_IF_EQ :
        x = options_ubee512_compare(e_optarg);
        if (x == -2)
           {
            param_error_mesg("'--if-eq'", e_optarg);
            exitstatus = 1;
           }
        else            
           if_state[++if_pos] = (x == 0) & if_state_prev;
        break;
     case OPT_IF_GT :
        x = options_ubee512_compare(e_optarg);
        if (x == -2)
           {
            param_error_mesg("'--if-gt'", e_optarg);
            exitstatus = 1;
           }
        else            
           if_state[++if_pos] = (x > 0) & if_state_prev;
        break;
     case OPT_IF_LT :
        x = options_ubee512_compare(e_optarg);
        if (x == -2)
           {
            param_error_mesg("'--if-lt'", e_optarg);
            exitstatus = 1;
           }
        else            
           if_state[++if_pos] = (x < 0) & if_state_prev;
        break;
     case OPT_IF_NEGT :
        x = options_ubee512_compare(e_optarg);
        if (x == -2)
           {
            param_error_mesg("'--if-negt", e_optarg);
            exitstatus = 1;
           }
        else            
           if_state[++if_pos] = (x < 0) & if_state_prev;
        break;
     case OPT_IF_NELT :
        x = options_ubee512_compare(e_optarg);
        if (x == -2)
           {
            param_error_mesg("'--if-nelt", e_optarg);
            exitstatus = 1;
           }
        else            
           if_state[++if_pos] = (x > 0) & if_state_prev;
        break;
     case OPT_IF_NEQ :
        x = options_ubee512_compare(e_optarg);
        if (x == -2)
           {
            param_error_mesg("'--if-neq", e_optarg);
            exitstatus = 1;
           }
        else            
           if_state[++if_pos] = (x != 0) & if_state_prev;
        break;
     case OPT_IF_NGT :
        x = options_ubee512_compare(e_optarg);
        if (x == -2)
           {
            param_error_mesg("'--if-ngt", e_optarg);
            exitstatus = 1;
           }
        else            
           if_state[++if_pos] = (x <= 0) & if_state_prev;
        break;
     case OPT_IF_NLT :
        x = options_ubee512_compare(e_optarg);
        if (x == -2)
           {
            param_error_mesg("'--if-nlt", e_optarg);
            exitstatus = 1;
           }
        else            
           if_state[++if_pos] = (x >= 0) & if_state_prev;
        break;
     case OPT_IF_NSET :
        if_state[++if_pos] = (options_ubee512_envvar_get(e_optarg) == NULL) & if_state_prev;
        break;
     case OPT_IF_SET :
        if_state[++if_pos] = (options_ubee512_envvar_get(e_optarg) != NULL) & if_state_prev;
        break;
     case OPT_IF_SYSTEM :
#ifdef MINGW
        x = ((strcmp(e_optarg_u, "WIN") == 0) || (strcmp(e_optarg_u, win_major_ver) == 0) ||
        (strcmp(e_optarg_u, win_minor_ver) == 0));
#else
        x = ((strcmp(e_optarg_u, "UNIX") == 0) || (strcmp(e_optarg_u, emu.sysname) == 0));
#endif               
        if_state[++if_pos] = x & if_state_prev;
        break;
     case OPT_IF_FALSE :
        if_state[++if_pos] = 0 & if_state_prev;
        break;   
     case OPT_IF_TRUE :
        if_state[++if_pos] = 1 & if_state_prev;
        break;            
     case OPT_IF_ELSE :
        if (if_state[if_pos - 1])
           if_state[if_pos] = ! if_state[if_pos];
        break;   
     case OPT_IF_END :
        if (if_pos)
           if_pos--;
        break;
     case OPT_IF_CMPMODE :
        if ((int_arg < 0) || (int_arg > 1))
           {
            param_error_mesg("'--if-cmpmode'", e_optarg);
            exitstatus = 1;
           }
        else
           if_cmp_mode = int_arg;    
        break;
    }
}

//==============================================================================
// Process options: Debugging (0x0200-0x02ff)
//
//   pass: int c			option number
// return: void
//==============================================================================
static void options_debugging (int c)
{
 switch (c)
    {
     case OPT_BP :
        if ((int_arg < 0) || (int_arg > 0xFFFF))
           {
            param_error_mesg("'--bp'", e_optarg);
            exitstatus = 1;
           }
        else
           debug.break_point[int_arg] = 1;
        break;
     case OPT_BPR :
        if ((int_arg < 0) || (int_arg > 0xFFFF))
           {
            param_error_mesg("'--bpr'", e_optarg);
            exitstatus = 1;
           }
        else
           debug.break_point[int_arg] = 2;
        break;
     case OPT_BPC :
        if (int_arg < 0)
           {
            param_error_mesg("'--bpc'", e_optarg);
            exitstatus = 1;
           }
        else
           debug.break_point_count = int_arg;
        break;         
     case OPT_DEBUG :
        snprintf(temp_str, sizeof(temp_str), "%s+", e_optarg);
        debug.debug = 1;
        if (strstr(temp_str, "+index+"))
           debug.show |= B8(00000001);
        if (strstr(temp_str, "+alt+"))
           debug.show |= B8(00000010);
        if (strstr(temp_str, "+all+"))
           debug.show |= B8(00000011);            
        if (strstr(temp_str, "+count+"))
           debug.showcount = 1;
        if (strstr(temp_str, "+step+"))
           z80debug_command(EMU_CMD_DBGSTEP01);
        if (strstr(temp_str, "+trace+"))
           z80debug_command(EMU_CMD_DBGTRACE);
        break;
     case OPT_DUMP :
        if ((int_arg < 0) || (int_arg > 0xFFFF))
           {
            param_error_mesg("'--dump'", e_optarg);
            exitstatus = 1;
           }
        else
           debug.dump_addr = int_arg;    
        break;
     case OPT_ECHO :
        xprintf("%s\n", e_optarg);
        break;
     case OPT_ECHOQ :
        xprintf("%s\n", e_optarg_q);
        break;
     case OPT_MODIO :
        snprintf(temp_str, sizeof(temp_str), "%s+", e_optarg);
        if (strstr(temp_str, "+log+"))
           modio.level = 1;
        if (strstr(temp_str, "+clock+"))
           modio.clock = 1;
        if (strstr(temp_str, "+crtc+"))
           modio.crtc = 1;
        if (strstr(temp_str, "+fdc+"))
           modio.fdc = 1;
        if (strstr(temp_str, "+fdc_wtd+"))
           modio.fdc_wtd = 1;
        if (strstr(temp_str, "+fdc_wth+"))
           modio.fdc_wth = 1;            
        if (strstr(temp_str, "+func+"))
           modio.func = 1;
        if (strstr(temp_str, "+hdd+"))
           modio.hdd = 1;
        if (strstr(temp_str, "+ide+"))
           modio.ide = 1;
        if (strstr(temp_str, "+joystick+"))
           modio.joystick = 1;
        if (strstr(temp_str, "+keystd+"))
           modio.keystd = 1;
        if (strstr(temp_str, "+keytc+"))
           modio.keytc = 1;
        if (strstr(temp_str, "+mem+"))
           modio.mem = 1;
        if (strstr(temp_str, "+options+"))
           modio.options = 1;
        if (strstr(temp_str, "+roms+"))
           modio.roms = 1;
        if (strstr(temp_str, "+pioa+"))
           modio.pioa = 1;
        if (strstr(temp_str, "+piob+"))
           modio.piob = 1;
        if (strstr(temp_str, "+piocont+"))
           modio.piocont = 1;
        if (strstr(temp_str, "+raminit+"))
           modio.raminit = 1;
        if (strstr(temp_str, "+rtc+"))
           modio.rtc = 1;
        if (strstr(temp_str, "+ubee512+"))
           modio.ubee512 = 1;               
        if (strstr(temp_str, "+vdu+"))
           modio.vdu = 1;
        if (strstr(temp_str, "+vdumem+"))
           modio.vdumem = 1;
        if (strstr(temp_str, "+video+"))
           modio.video = 1;
        if (strstr(temp_str, "+z80+"))
           modio.z80 = 1;
        break;
     case OPT_REGS :
        snprintf(temp_str, sizeof(temp_str), "%s+", e_optarg);
        if (strstr(temp_str, "+clear+"))
           {
            regdump.crtc = 0;
            regdump.rtc = 0;
            regdump.z80 = 0;
           } 
        if (strstr(temp_str, "+crtc+"))
           regdump.crtc = 1;
        if (strstr(temp_str, "+pio+"))
           regdump.pio = 1;
        if (strstr(temp_str, "+rtc+"))
           regdump.rtc = 1;
        if (strstr(temp_str, "+z80+"))
           regdump.z80 = 1;
        break;
    }
}

//==============================================================================
// Process options: Disks (0x0300-0x03ff)
//
//   pass: int c			option number
// return: void
//==============================================================================
static void options_disks (int c)
{
 int i;

 switch (c)
    {
     case OPT_HDD0 : // WD1002-5 Winchester drive
     case OPT_HDD1 : // WD1002-5 Winchester drive
     case OPT_HDD2 : // WD1002-5 Winchester drive
     case OPT_HDD3 : // WD1002-5 floppy drive 1
     case OPT_HDD4 : // WD1002-5 floppy drive 2
     case OPT_HDD5 : // WD1002-5 floppy drive 3
     case OPT_HDD6 : // WD1002-5 floppy drive 4
        i = c - OPT_HDD0;
        strcpy(hdd_d.disk.filename, e_optarg);
#ifdef USE_LIBDSK
        strcpy(hdd_d.disk.libdsk_type, use_driver_type);
        strcpy(hdd_d.disk.libdsk_format, use_format_type);
        if (c >= OPT_HDD3)
           {
            hdd_d.disk.side1as0 = side1as0;
            hdd_d.disk.dstep = dstep;
            hdd_d.disk.dstep_hd = dstep_hd;
           }
        use_driver_type[0] = 0;
        use_format_type[0] = 0;
        side1as0 = 0;
        dstep = 0;
        dstep_hd = 0;
#endif
        hdd_set_drive(i, &hdd_d);
        break;
     case OPT_HDD3_CLOSE :
     case OPT_HDD4_CLOSE :
     case OPT_HDD5_CLOSE :
     case OPT_HDD6_CLOSE :
        i = (c - OPT_HDD3_CLOSE) + 3;
        hdd_unloaddisk(i);
        break;
     case OPT_IDE_A0 :
     case OPT_IDE_A1 :
     case OPT_IDE_B0 :
     case OPT_IDE_B1 :
        i = c - OPT_IDE_A0;
        strcpy(ide_d.disk.filename, e_optarg);
#ifdef USE_LIBDSK
        strcpy(ide_d.disk.libdsk_type, use_driver_type);
        strcpy(ide_d.disk.libdsk_format, use_format_type);
        use_driver_type[0] = 0;
        use_format_type[0] = 0;
#endif
        ide_set_drive(i, &ide_d);
        break;
     case OPT_IMAGE_A : // core board floppy drive 1 (A)
     case OPT_IMAGE_B : // core board floppy drive 2 (B)
     case OPT_IMAGE_C : // core board floppy drive 3 (C)
     case OPT_IMAGE_D : // core board floppy drive 4 (D)
        i = c - OPT_IMAGE_A;
        strcpy(fdc_d.disk.filename, e_optarg);
#ifdef USE_LIBDSK
        strcpy(fdc_d.disk.libdsk_type, use_driver_type);
        strcpy(fdc_d.disk.libdsk_format, use_format_type);
        fdc_d.disk.side1as0 = side1as0;
        fdc_d.disk.dstep = dstep;
        fdc_d.disk.dstep_hd = dstep_hd;
        use_driver_type[0] = 0;
        use_format_type[0] = 0;
        side1as0 = 0;
        dstep = 0;
        dstep_hd = 0;
#endif
        fdc_set_drive(i, &fdc_d);
        break;
     case OPT_A_CLOSE :
     case OPT_B_CLOSE :
     case OPT_C_CLOSE :
     case OPT_D_CLOSE :
        i = c - OPT_A_CLOSE;
        fdc_unloaddisk(i);
        break;
#ifdef USE_LIBDSK
     case OPT_DSTEP :
        dstep = 1;
        break;
     case OPT_DSTEP_HD :
        dstep_hd = 1;
        dstep = 1; 
        break;
     case OPT_FORMAT :
        strncpy(use_format_type, e_optarg, sizeof(use_format_type));
        break;
     case OPT_LFORMAT :
        format = FMT_180K;
	          while (dg_stdformat(NULL, format++, &fname, &fdesc) == DSK_ERR_OK)
	             xprintf("%-10.10s : %s\n", fname, fdesc);
        exitstatus = -1;
        break;
     case OPT_LTYPE :
        i = 0;
        do
           {
            dsk_type_enum(i++, &xstr);
            if (xstr)
               xprintf("%s\n", xstr);
           }    
        while (xstr);
        exitstatus = -1;
        break;
     case OPT_SIDE1AS0 :
        side1as0 = 1;
        break;
     case OPT_TYPE :
        strncpy(use_driver_type, e_optarg, sizeof(use_driver_type));
        break;
#endif
     case OPT_PSEC :
        xprintf("ubee512: Option `--psec' has been removed and is no longer required.\n");
        break; // not an option now, probing is always used now.
    }
}

//==============================================================================
// Process options: Display related (0x0400-0x04ff)
//
//   pass: int c			option number
// return: void
//==============================================================================
static void options_display (int c)
{
 char *video_type_args[] =
 {
  "sw",
  "hw",
  "gl",
  ""
 };

 char *video_depth_args[] =
 {
  "8",
  "8gs",
  "16",
  "32",
  ""
 };

#ifdef USE_OPENGL
 char *gl_filter_args[] =
 {
  "soft",
  "sharp",
  ""
 };
#endif

 int x;

 switch (c)
    {
     case OPT_ASPECT :
        if ((int_arg < 1) || (int_arg > 2))
           {
            param_error_mesg("'--aspect'", e_optarg);
            exitstatus = 1;
           }
        else
           crtc.aspect = int_arg;
        break;
     case OPT_FULLSCREEN :
        if (! e_optarg[0])
           video.fullscreen = ! video.fullscreen;
        else
           {   
            x = string_search(offon_args, e_optarg);
            if (x == -1)
               {
                param_error_mesg("'-f, --fullscreen'", e_optarg);
                exitstatus = 1;
               }
            else
               video.fullscreen = x;
           }
        break;
     case OPT_MONITOR :
        x = string_search(monitor_args, e_optarg);
        if (x == -1)
           {
            param_error_mesg("'-m, --monitor'", e_optarg);
            exitstatus = 1;
           }
        else
           {
            if (x > 5)
               crtc.monitor = x - 6;
            else   
               crtc.monitor = x;
            if (emu.runmode)
               crtc_setmonitor();
           }    
        break;

     case OPT_MON_BG_B : 
     case OPT_MON_BG_G :
     case OPT_MON_BG_R :
     case OPT_MON_BGI_B :
     case OPT_MON_BGI_G :
     case OPT_MON_BGI_R :
     case OPT_MON_FG_B :
     case OPT_MON_FG_G :
     case OPT_MON_FG_R :
     case OPT_MON_FGI_B :
     case OPT_MON_FGI_G :
     case OPT_MON_FGI_R :
        if ((int_arg < 0) || (int_arg > 255))
           {
            param_error_mesg("'--mon-[bgh/bg/fg]-x'", e_optarg);
            exitstatus = 1;
           }
        else
           crtc_set_mon_table(c - OPT_MON_BG_B, int_arg);
        break;

     case OPT_MON_FGL_B :
     case OPT_MON_FGL_G :
     case OPT_MON_FGL_R :
        xprintf("ubee512: Options `--mon-fgl-x' are no longer supported and are ignored.\n");
        xprintf("ubee512: These should be removed and the replacemnt options used instead.\n");
        xprintf("ubee512: See the ubee512rc.sample and README files.\n");
        break;

     case OPT_RGB_00_R : 
     case OPT_RGB_00_G :
     case OPT_RGB_00_B :
     case OPT_RGB_01_R :
     case OPT_RGB_01_G :
     case OPT_RGB_01_B :
     case OPT_RGB_02_R :
     case OPT_RGB_02_G :
     case OPT_RGB_02_B :
     case OPT_RGB_03_R :
     case OPT_RGB_03_G :
     case OPT_RGB_03_B :
     case OPT_RGB_04_R :
     case OPT_RGB_04_G :
     case OPT_RGB_04_B :
     case OPT_RGB_05_R :
     case OPT_RGB_05_G :
     case OPT_RGB_05_B :
     case OPT_RGB_06_R :
     case OPT_RGB_06_G :
     case OPT_RGB_06_B :
     case OPT_RGB_07_R :
     case OPT_RGB_07_G :
     case OPT_RGB_07_B :
     case OPT_RGB_08_R :
     case OPT_RGB_08_G :
     case OPT_RGB_08_B :
     case OPT_RGB_09_R :
     case OPT_RGB_09_G :
     case OPT_RGB_09_B :
     case OPT_RGB_10_R :
     case OPT_RGB_10_G :
     case OPT_RGB_10_B :
     case OPT_RGB_11_R :
     case OPT_RGB_11_G :
     case OPT_RGB_11_B :
     case OPT_RGB_12_R :
     case OPT_RGB_12_G :
     case OPT_RGB_12_B :
     case OPT_RGB_13_R :
     case OPT_RGB_13_G :
     case OPT_RGB_13_B :
     case OPT_RGB_14_R :
     case OPT_RGB_14_G :
     case OPT_RGB_14_B :
     case OPT_RGB_15_R :
     case OPT_RGB_15_G :
     case OPT_RGB_15_B :
        if ((int_arg < 0) || (int_arg > 255))
           {
            param_error_mesg("'--rgb-nn-x'", e_optarg);
            exitstatus = 1;
           }
        else
           col_table_p[c - OPT_RGB_00_R] = int_arg;
        break;

     case OPT_VIDEO :
        x = string_search(offon_args, e_optarg);
        if (x == -1)
           {
            param_error_mesg("'--video'", e_optarg);
            exitstatus = 1;
           }
        else
           crtc.video = x;
        break;
     case OPT_VIDEO_DEPTH :
        x = string_search(video_depth_args, e_optarg);
        if (x == -1)
           {
            param_error_mesg("'--video-depth'", e_optarg);
            exitstatus = 1;
           }
        else
           video.depth = x;
        break;
     case OPT_VIDEO_TYPE : 
#ifndef USE_OPENGL
        if (strcmp(e_optarg, "gl") == 0)
           break;
#endif
        x = string_search(video_type_args, e_optarg);
        if (x == -1)
           {
            param_error_mesg("'--video-type'", e_optarg);
            exitstatus = 1;
           }
        else
           video.type = x;
        break;
#ifdef USE_OPENGL
     // OpenGL rendering
     case OPT_GL_ASPECT_BEE :
        if (video_gl_aspect_bee(float_arg) == -1)
           {
            param_error_mesg("'--gl-aspect-bee'", e_optarg);
            exitstatus = 1;
           }
        break;
     case OPT_GL_ASPECT_MON :
        if (video_gl_aspect_mon(float_arg) == -1)
           {
            param_error_mesg("'--gl-aspect-mon'", e_optarg);
            exitstatus = 1;
           }         
        break;
     case OPT_GL_FILTER_FS :
        x = string_search(gl_filter_args, e_optarg);
        if (x == -1)
           {
            param_error_mesg("'--gl-filter-fs'", e_optarg);
            exitstatus = 1;
           }
        else
           {
            video.filter_fs = x;
            video_gl_filter_update();
           } 
        break;
     case OPT_GL_FILTER_MAX :
        x = string_search(gl_filter_args, e_optarg);
        if (x == -1)
           {
            param_error_mesg("'--gl-filter-max'", e_optarg);
            exitstatus = 1;
           }
        else
           {
            video.filter_max = x;
            video_gl_filter_update();
           } 
        break;            
     case OPT_GL_FILTER_WIN :
        x = string_search(gl_filter_args, e_optarg);
        if (x == -1)
           {
            param_error_mesg("'--gl-filter-win'", e_optarg);
            exitstatus = 1;
           }
        else
           {
            video.filter_win = x;
            video_gl_filter_update();
           } 
        break;
     case OPT_GL_MAX :
#ifdef MINGW
        xprintf("ubee512: '--gl-max' option is not currently supported under Windows.\n");
        exitstatus = 1; 
#else
        x = string_search(offon_args, e_optarg);
        if (x == -1)
           {
            param_error_mesg("'--gl-max'", e_optarg);
            exitstatus = 1;
           }
        else
           video.max = x;
#endif
        break;
     case OPT_GL_VSYNC :
        x = string_search(offon_args, e_optarg);
        if (x == -1)
           {
            param_error_mesg("'--gl-vsync'", e_optarg);
            exitstatus = 1;
           }
        else
           video.vsync = x;
        break;
     case OPT_GL_WINPCT :
        if (video_gl_size_percent(int_arg) == -1)
           {
            param_error_mesg("'--gl-winpct'", e_optarg);
            exitstatus = 1;
           }
        break;
     case OPT_GL_WINPIX :
        if (video_gl_size_pixels(int_arg) == -1)
           {
            param_error_mesg("'--gl-winpix'", e_optarg);
            exitstatus = 1;
           }
        break;
#endif
    }
}

//==============================================================================
// Process options: Model (0x0500-0x05ff)
//
//   pass: int c			option number
// return: void
//==============================================================================
static void options_model (int c)
{
 char *piob7_args[] =
 {
  "pup", 
  "vsync",
  "rtc",  
  "net",
  ""
 };

 char *flash_args[] =
 {
  "off", 
  "on",
  "v4",  
  "v3",
  ""
 };

 char *flashrate_v3_args[] =
 {
  "20",
  "40",
  "80",
  "160",
  "320",
  "640",  
  "1280",
  "2560",
  "w61",
  "w62",
  "w64",  // w63 and w64 are swapped!
  "w63",
  ""
 };

 char *flashrate_v4_args[] =
 {
  "20",
  "40",
  "80",
  "160",
  "320",
  "640",  
  "1280",
  "2560",
  "w61",
  "w62",
  "w63",
  "w64",
  ""
 };

 int x;

 switch (c)
    {
     case OPT_BASIC :
     case OPT_BASICA :
        strncpy(modelc.basica, e_optarg, sizeof(modelc.basica));
        modelc.basica[sizeof(modelc.basica)-1] = 0;
        break;
     case OPT_BASICB :
        strncpy(modelc.basicb, e_optarg, sizeof(modelc.basicb));
        modelc.basicb[sizeof(modelc.basicb)-1] = 0;
        break;
     case OPT_BASICC :
        strncpy(modelc.basicc, e_optarg, sizeof(modelc.basicc));
        modelc.basicc[sizeof(modelc.basicc)-1] = 0;
        break;
     case OPT_BASICD :
        strncpy(modelc.basicd, e_optarg, sizeof(modelc.basicd));
        modelc.basicd[sizeof(modelc.basicd)-1] = 0;
        break;

     case OPT_BASRAM :
        modelc.basram = 1;
        break;

     case OPT_CHARROM :
        strncpy(modelc.charrom, e_optarg, sizeof(modelc.charrom));
        modelc.charrom[sizeof(modelc.charrom)-1] = 0;
        break;

     case OPT_COL :
        if (! modelx.alphap)
           modelx.colour = 1;
        break;
     case OPT_COL_TYPE :
        if ((int_arg < 0) || (int_arg > 1))
           {
            param_error_mesg("'--col-type'", e_optarg);
            exitstatus = 1;
           }
        else
           {
            crtc.std_col_type = int_arg;
            if (! modelx.alphap)
               modelx.colour = 1;
            if (emu.runmode)
               crtc_init();   
           } 
        break;            
     case OPT_COLPROM :
        strncpy(modelc.colprom, e_optarg, sizeof(modelc.colprom));
        modelc.basica[sizeof(modelc.colprom)-1] = 0;
        break;

     case OPT_DINT :
     case OPT_HINT :
        x = string_search(offon_args, e_optarg);
        if (x == -1)
           {
            param_error_mesg("'--dint, --hint'", e_optarg);
            exitstatus = 1;
           }
        else
           modelx.halfint = x;
        break;
     case OPT_HARDWARE :
        snprintf(temp_str, sizeof(temp_str), "%s-", e_optarg);
        if (strstr(temp_str, "-wd2793-"))
           emu.hardware &= ~HW_WD2793;
        break;
     case OPT_HWFLASH :
        x = string_search(flash_args, e_optarg);
        if (x == -1)
           {
            param_error_mesg("'--hwflash'", e_optarg);
            exitstatus = 1;
           }
        else
           {
            if (x > 1)
               x--;
            modelx.hwflash = x;
           } 
        break;
     case OPT_HWFLASHR :
        if (modelx.hwflash == 2)  // if version 3
           x = string_search(flashrate_v3_args, e_optarg);
        else   
           x = string_search(flashrate_v4_args, e_optarg);
        if (crtc_set_flash_rate(x) == -1)
           {
            param_error_mesg("'--hwflashr'", e_optarg);
            exitstatus = 1;
           }
        break;
     case OPT_LMODEL :
        x = 0;
        while (model_args[x][0])
           xprintf("%s\n", model_args[x++]);
        exitstatus = -1;
        break;
     case OPT_LPEN :
        modelx.lpen = 1;
        break;
     case OPT_MODEL :
        x = string_search(model_args, e_optarg);
        if (x == -1)
           {
            param_error_mesg("'--model'", e_optarg);
            exitstatus = 1;
           }
        else
           {
            emu.model = x;
            memcpy(&modelx, &model_data[emu.model], sizeof(model_t));
           }
        break;
     case OPT_MONO :
        if (modelx.alphap)
           {
            crtc.monitor = string_search(monitor_args, "g");  // can't be an error
            if (emu.runmode)
               crtc_setmonitor();
           } 
        else   
           modelx.colour = 0;
        break;
     case OPT_NETRAM :
        modelc.netram = 1;
        break;
     case OPT_NETROM :
        strncpy(modelc.netrom, e_optarg, sizeof(modelc.netrom));
        modelc.netrom[sizeof(modelc.netrom)-1] = 0; break;
        break;

     case OPT_PAK :
     case OPT_PAK0 :
     case OPT_PAK1 :
     case OPT_PAK2 :
     case OPT_PAK3 :
     case OPT_PAK4 :
     case OPT_PAK5 :
     case OPT_PAK6 :
     case OPT_PAK7 :
        if (c == OPT_PAK)
           x = 0;
        else   
           x = c - OPT_PAK0;
        strncpy(modelc.pak[x], e_optarg, sizeof(modelc.pak[x]));
        modelc.pak[x][sizeof(modelc.pak[x])-1] = 0;
        break;

     case OPT_PAKRAM :
        if ((int_arg < 0) || (int_arg > 7))
           {
            param_error_mesg("'--pakram'", e_optarg);
            exitstatus = 1;
           }
        else
           modelc.pakram[int_arg] = 1;
        break;

     case OPT_PCG :
        if ((int_arg < 2) || (int_arg > 32) || (int_arg % 2))
           {
            param_error_mesg("'--pcg'", e_optarg);
            exitstatus = 1;
           }
        else
           if (modelx.alphap)
              modelx.pcg = int_arg / 2;
        break;
     case OPT_PIOB7 :
        x = string_search(piob7_args, e_optarg);
        if (x == -1) 
           {
            param_error_mesg("'--piob7'", e_optarg);
            exitstatus = 1;
           }
        else
           modelx.piob7 = x;
        break;

     case OPT_ROM1 :
        strncpy(modelc.rom1, e_optarg, sizeof(modelc.rom1));
        modelc.rom1[sizeof(modelc.rom1)-1] = 0;
        break;
     case OPT_ROM2 :
        strncpy(modelc.rom2, e_optarg, sizeof(modelc.rom2));
        modelc.rom2[sizeof(modelc.rom2)-1] = 0;
        break;
     case OPT_ROM3 :
        strncpy(modelc.rom3, e_optarg, sizeof(modelc.rom3));
        modelc.rom3[sizeof(modelc.rom3)-1] = 0;
        break;

     case OPT_ROM256K :
        strncpy(modelc.rom256k, e_optarg, sizeof(modelc.rom256k));
        modelc.rom256k[sizeof(modelc.rom256k)-1] = 0;
        break;

     case OPT_SYS :
        strncpy(modelc.systname, e_optarg, sizeof(modelc.systname));
        modelc.systname[sizeof(modelc.systname)-1] = 0;
        break;
     case OPT_VDU :
        if ((int_arg != 2) && (int_arg != 8))
           {
            param_error_mesg("'--vdu'", e_optarg);
            exitstatus = 1;
           }
        else
           if (modelx.alphap)
              modelx.vdu = (int_arg / 2) - 1;
        break;
    }
}

//==============================================================================
// Process options: Information output (0x0600-0x06ff)
//
//   pass: int c			option number
// return: void
//==============================================================================
static void options_information (int c)
{
 int i;
 int x;
 char vers[20];

 switch (c)
    {
     case OPT_CONIO :
        console.force_stdout = 1;
        break;
     case OPT_HELP :
     case OPT_USAGE :
        options_usage();
        exitstatus = -2;
        break;
     case OPT_LCON :
        for (i = list_config_start; i < ndefsc; i++)
           xprintf("%s\n", ndefsv[i]);
        exitstatus = -1;
        break;
     case OPT_LCONW :
        i = list_config_start;
        x = 0;
        while (i < ndefsc)
           {
            xprintf("%-16s", ndefsv[i++]);
            x++;
            if ((x % 5) == 0)
               xprintf("\n");
           }    
        if ((x % 5) != 0)
           xprintf("\n");
        exitstatus = -1;
        break;
     case OPT_LCONS :
        if ((int_arg < 1) || (int_arg > ndefsc)) 
           {
            param_error_mesg("'--lcons'", e_optarg);
            exitstatus = 1;
           }
        else
           list_config_start = int_arg - 1;
        break;
     case OPT_VERSION :
        xprintf("%s\n", APPVER);
        sdlv = SDL_Linked_Version();
        xprintf("SDL %u.%u.%u\n", sdlv->major, sdlv->minor, sdlv->patch);
        z80api_get_version(vers, sizeof(vers));
        xprintf("%s\n", vers);
        exitstatus = 1;
        break;
    }
}    

//==============================================================================
// Process options: Parallel port (0x0700-0x07ff)
//
//   pass: int c			option number
// return: void
//==============================================================================
static void options_parallel_port (int c)
{
 switch (c)
    {
     case OPT_PRINT :
        printer_b_open(e_optarg, emu.runmode);
        break;
     case OPT_PRINT_CLOSE :
        printer_b_close();
        break;
     case OPT_PRINTA :
        printer_a_open(e_optarg, emu.runmode);
        break;
     case OPT_PRINTA_CLOSE :
        printer_a_close();
        break;
    }
}

//==============================================================================
// Process options: Serial port (0x0800-0x08ff)
//
//   pass: int c			option number
// return: void
//==============================================================================
static void options_serial_port (int c)
{
 switch (c)
    {
     case OPT_BAUD :
        if ((int_arg < 1) || (int_arg > 19200))
           {
            param_error_mesg("'--baud'", e_optarg);
            exitstatus = 1;
           }
        else
           {
            serial.tx_baud = int_arg;
            serial.rx_baud = int_arg;
            if (emu.runmode)
               serial_config(emu.cpuclock);
           } 
        break;
     case OPT_BAUDRX :
        if ((int_arg < 1) || (int_arg > 19200))
           {
            param_error_mesg("'--baudrx'", e_optarg);
            exitstatus = 1;
           }
        else
           {
            serial.rx_baud = int_arg;
            if (emu.runmode)
               serial_config(emu.cpuclock);
           }
        break;
     case OPT_BAUDTX :
        if ((int_arg < 1) || (int_arg > 19200))
           {
            param_error_mesg("'--baudtx'", e_optarg);
            exitstatus = 1;
           }
        else
           {
            serial.tx_baud = int_arg;
            if (emu.runmode)
               serial_config(emu.cpuclock);
           }
        break;
     case OPT_COMS :
        if (serial_open(e_optarg, 0, emu.runmode) == -1)
           {
            param_error_mesg("'--coms1'", e_optarg);
            exitstatus = 1;
           }
        break;
     case OPT_COMS_CLOSE :
        serial_close(0);
        break;
     case OPT_DATAB :
        if ((int_arg < 5) || (int_arg > 8))
           {
            param_error_mesg("'--datab'", e_optarg);
            exitstatus = 1;
           }
        else
           {
            serial.databits = int_arg;
            if (emu.runmode)
               serial_config(emu.cpuclock);
           }
        break;
     case OPT_STOPB :
        if ((int_arg < 1) || (int_arg > 2))
           {
            param_error_mesg("'--stopb'", e_optarg);
            exitstatus = 1;
           }
        else
           {
            serial.stopbits = int_arg;
            if (emu.runmode)
               serial_config(emu.cpuclock);
           }
        break;
    }
}

//==============================================================================
// Process options: Sound emulation (0x0900-0x09ff)
//
//   pass: int c			option number
// return: void
//==============================================================================
static void options_sound (int c)
{
 char *sound_args[] =
 {
  "off",
  "prop",
  "normal",
  ""
 };

 int x;

 switch (c)
    {
     case OPT_SOUND :
        x = string_search(sound_args, e_optarg);
        if (x == -1) 
           {
            param_error_mesg("'--sound'", e_optarg);
            exitstatus = 1;
           } 
        else
           sound.enabled = x;
        break;
     case OPT_SND_ALG1 :
        x = string_search(offon_args, e_optarg);
        if (x == -1)
           {
            param_error_mesg("'--snd-alg1'", e_optarg);
            exitstatus = 1;
           }
        else
           sound.use_alg1 = x;
        break;
     case OPT_SND_FREQ :
        if ((int_arg < 5512) || (int_arg > 176400))
           {
            param_error_mesg("'--snd-freq'", e_optarg);
            exitstatus = 1;
           }
        else
           sound.frequency = int_arg;    
        break;
     case OPT_SND_FREQADJ :
        if ((float_arg < -50.0) || (float_arg > 50.0))
           {
            param_error_mesg("'--snd-freqadj'", e_optarg);
            exitstatus = 1;
           }
        else
           sound.freq_adjust = float_arg;    
        break;
     case OPT_SND_FREQLOW :
        if ((int_arg < 1) || (int_arg > 20000))
           {
            param_error_mesg("'--snd-freqlow'", e_optarg);
            exitstatus = 1;
           }
        else
           sound.freq_lowest = int_arg;    
        break;
     case OPT_SND_HOLDOFF :
        if ((int_arg < 0) || (int_arg > 10000))
           {
            param_error_mesg("'--snd-holdoff'", e_optarg);
            exitstatus = 1;
           }
        else
           sound.holdoff = int_arg;    
        break;
     case OPT_SND_HQ :
        sound.holdoff = 50;
        sound.samples = 2048;
        sound.frequency = 88200;
        sound.freq_lowest = 10;
        sound.freq_adjust = -0.25;
        sound.use_alg1 = 1;
        break;
     case OPT_SND_MUTE :
        x = string_search(offon_args, e_optarg);
        if (x == -1)
           {
            param_error_mesg("'--snd-mute'", e_optarg);
            exitstatus = 1;
           }
        else
           sound.mute = x;
        break;
     case OPT_SND_SAMPLES : if (((int_arg & (int_arg - 1)) != 0) || (int_arg > 16384)) // check is power of 2
           {
            param_error_mesg("'--snd-samples'", e_optarg);
            exitstatus = 1;
           }
        else
           sound.samples = int_arg;    
        break;
     case OPT_SND_VOLUME :
     case OPT_VOL :
        if ((int_arg < 0) || (int_arg > 100))
           {
            param_error_mesg("'--snd-vol, --vol'", e_optarg);
            exitstatus = 1;
           }
        else
           {
            sound.volume = int_arg;
            sound_volume(sound.volume);
           } 
        break;
    }
}

//==============================================================================
// Process options: Speed related (0x0a00-0x0aff)
//
//   pass: int c			option number
// return: void
//==============================================================================
static void options_speed (int c)
{
 int x;
 
 switch (c)
    {
     case OPT_CLOCK :
     case OPT_XTAL :
        if (float_arg <= 0.0)
           {
            param_error_mesg("'--clock, -x, --xtal'", e_optarg);
            exitstatus = 1;
           }
        else
           {
            modelx.cpuclock = float_arg;
            if (emu.runmode)
               set_clock_speed(modelx.cpuclock, 0, 0);
           } 
        break;
     case OPT_CLOCK_DEF :
        if (float_arg <= 0.0)
           {
            param_error_mesg("'--clock-def'", e_optarg);
            exitstatus = 1;
           }
        else
           emu.cpuclock_def = float_arg;    
        break;
     case OPT_FRATE :
        if ((int_arg < 1) || (int_arg > 1000000L))
           {
            param_error_mesg("'--frate'", e_optarg);
            exitstatus = 1;
           }
        else
           {
            emu.framerate = int_arg;
            if (emu.runmode)
               set_clock_speed(modelx.cpuclock, emu.z80_divider, emu.framerate);
           }
        break;
     case OPT_MAXCPULAG :
        if (int_arg < 0)
           {
            param_error_mesg("'--maxcpulag'", e_optarg);
            exitstatus = 1;
           }
        else
           emu.maxcpulag = int_arg;    
        break;
     case OPT_VBLANK :
        if ((int_arg < 0) || (int_arg > 1))
           {
            param_error_mesg("'--vblank'", e_optarg);
            exitstatus = 1;
           }
        else
           crtc.vblank_method = int_arg;    
        break;
     case OPT_SPEEDSEL :
        if ((int_arg < 0) || (int_arg > 1))
           {
            param_error_mesg("'--speedsel'", e_optarg);
            exitstatus = 1;
           }
        else
           modelx.speed = int_arg;
        break;
     case OPT_TURBO :
        if (! e_optarg[0])
           emu.turbo = 1;
        else
           {
            x = string_search(offon_args, e_optarg);
            if (x == -1)
               {
                param_error_mesg("'-t, --turbo'", e_optarg);
                exitstatus = 1;
               }
            else
               emu.turbo = x;
           }
        break;
     case OPT_Z80DIV :
        if ((int_arg < 1) || (int_arg > 5000))
           {
            param_error_mesg("'--z80div'", e_optarg);
            exitstatus = 1;
           }
        else
           {
            emu.z80_divider = int_arg;
            if (emu.runmode)
               set_clock_speed(modelx.cpuclock, emu.z80_divider, 0);
           }
        break;            
    }
}

//==============================================================================
// Process options: Tape port (0x0b00-0x0bff)
//
//   pass: int c			option number
// return: void
//==============================================================================
static void options_tape (int c)
{
 switch (c)
    {
     case OPT_TAPEI :
        if (tape_check(tape.tapeo, e_optarg) == 0)
           tape_i_open(e_optarg, emu.runmode);
        else
           {
            param_error_mesg("'--tapei'", e_optarg);
            exitstatus = 1;
           }
        break;
     case OPT_TAPEI_CLOSE :
        tape_i_close();
        break;
     case OPT_TAPE_DET :
        tape.detect = (uint32_t)int_arg;
        break;
     case OPT_TAPEO :
        if (tape_check(tape.tapei, e_optarg) == 0)
           tape_o_open(e_optarg, emu.runmode);
        else
           {
            param_error_mesg("'--tapeo'", e_optarg);
            exitstatus = 1;
           }
        break;   
     case OPT_TAPEO_CLOSE :
        tape_o_close();
        break;
     case OPT_TAPESAMP :
        if ((int_arg < 1) || (int_arg > 1000000L))
           {
            param_error_mesg("'--tapesamp'", e_optarg);
            exitstatus = 1;
           }
        else
           {
            tape.orate = int_arg;
            if (emu.runmode)
               tape_config_out(modelx.cpuclock);
           }
        break;
     case OPT_TAPEVOL :
        if ((int_arg < 0) || (int_arg > 100))
           {
            param_error_mesg("'--tapevol'", e_optarg);
            exitstatus = 1;
           }
        else
           tape.olevel = (int)(127 * (int_arg / 100.0));
        break;
    }
}

//==============================================================================
// Process options: Real Time Clock (RTC) emulation and time (0x0c00-0x0cff)
//
//   pass: int c			option number
// return: void
//==============================================================================
static void options_rtc (int c)
{
 switch (c)
    {
     case OPT_CENTURY :
        if ((int_arg < 0) || (int_arg > 255))
           {
            param_error_mesg("'--century'", e_optarg);
            exitstatus = 1;
           }
        else
           emu.century = int_arg;
        break;
     case OPT_RTC :
        if ((int_arg != 0) && (int_arg != 1))
           {
            param_error_mesg("'--rtc'", e_optarg);
            exitstatus = 1;
           }
        else
           modelx.rtc = int_arg;
        break;
    }
}

//==============================================================================
// Process options: Joystick emulation (0x0d00-0x0dff)
//
//   pass: int c			option number
// return: void
//==============================================================================
static void options_joystick (int c)
{
 int x;

 switch (c)
    {
     case OPT_JS :
        if ((int_arg < -1) || (int_arg > 127))	// yes -1 !
           {
            param_error_mesg("'--js'", e_optarg);
            exitstatus = 1;
           }
        else
           joystick.used = int_arg;
        break;

     case OPT_JS_AXIS :
        x = string_search(offon_args, e_optarg);
        if (x == -1)
           {
            param_error_mesg("'--js-axis'", e_optarg);
            exitstatus = 1;
           }
        else
           joystick.axis_used = x;
        break;
     case OPT_JS_AXISB :
        if ((int_arg < 0) || (int_arg > 124))
           {
            param_error_mesg("'--js-axisb'", e_optarg);
            exitstatus = 1;
           }
        else
           joystick.axis_buttons = int_arg;
        break;
     case OPT_JS_AXISL :
        if ((int_arg < 1) || (int_arg > 32767))
           {
            param_error_mesg("'--js-axisl'", e_optarg);
            exitstatus = 1;
           }
        else
           joystick.axis_level = int_arg;
        break;                        

     case OPT_JS_CLEAR :
        joystick_mbjoy_clear();
        break;

     case OPT_JS_MBEE :
        x = string_search(offon_args, e_optarg);
        if (x == -1)
           {
            param_error_mesg("'--js-mbee'", e_optarg);
            exitstatus = 1;
           }
        else
           joystick.mbee = x;
        break;

     case OPT_JS_UP :
        if (joystick_mbjoy_set_action(JOY_MB_UP, int_arg))
           {
            param_error_mesg("'--js-up'", e_optarg);
            exitstatus = 1;
           }
        break;
     case OPT_JS_RIGHT :
        if (joystick_mbjoy_set_action(JOY_MB_RIGHT, int_arg))
           {
            param_error_mesg("'--js-right'", e_optarg);
            exitstatus = 1;
           }
        break;
     case OPT_JS_DOWN :
        if (joystick_mbjoy_set_action(JOY_MB_DOWN, int_arg))
           {
            param_error_mesg("'--js-down'", e_optarg);
            exitstatus = 1;
           }
        break;
     case OPT_JS_LEFT :
        if (joystick_mbjoy_set_action(JOY_MB_LEFT, int_arg))
           {
            param_error_mesg("'--js-left'", e_optarg);
            exitstatus = 1;
           }
        break;
     case OPT_JS_FIRE :
        if (joystick_mbjoy_set_action(JOY_MB_FIRE, int_arg))
           {
            param_error_mesg("'--js-fire'", e_optarg);
            exitstatus = 1;
           }
        break;
     case OPT_JS_PLAY1 :
        if (joystick_mbjoy_set_action(JOY_MB_PLAY1, int_arg))
           {
            param_error_mesg("'--js-play1'", e_optarg);
            exitstatus = 1;
           }
        break;
     case OPT_JS_PLAY2 :
        if (joystick_mbjoy_set_action(JOY_MB_PLAY2, int_arg))
           {
            param_error_mesg("'--js-play2'", e_optarg);
            exitstatus = 1;
           }
        break;
     case OPT_JS_SPARE :
        if (joystick_mbjoy_set_action(JOY_MB_SPARE, int_arg))
           {
            param_error_mesg("'--js-spare'", e_optarg);
            exitstatus = 1;
           }
        break;

     case OPT_JS_CLIST :
        joystick_kbjoy_listcommands();
        exitstatus = -1;
        break;
     case OPT_JS_KLIST :
        joystick_kbjoy_listkeys();
        exitstatus = -1;
        break;
     case OPT_JS_KBD :
        x = string_search(offon_args, e_optarg);
        if (x == -1)
           {
            param_error_mesg("'--js-kbd'", e_optarg);
            exitstatus = 1;
           }
        else
           joystick.kbd = x;
        break;
     case OPT_JS_KK :
        if (joystick_kbjoy_key(e_optarg))
           {
            param_error_mesg("'--js-kk'", e_optarg);
            exitstatus = 1;
           }
        break;
     case OPT_JS_KB :
        if (joystick_kbjoy_button(int_arg))
           {
            param_error_mesg("'--js-kb'", e_optarg);
            exitstatus = 1;
           }
        break;
     case OPT_JS_KSET :
        if (joystick_kbjoy_set(int_arg, e_optarg))
           {
            param_error_mesg("'--js-kset'", e_optarg);
            exitstatus = 1;
           }
        else
           {
            joystick.kbd = 1;
            joystick_kbjoy_select(int_arg, e_optarg);
           } 
        break;
     case OPT_JS_KSEL :
        if (joystick_kbjoy_select(int_arg, e_optarg))
           {
            param_error_mesg("'--js-ksel'", e_optarg);
            exitstatus = 1;
           }
        break;
    }
}

//==============================================================================
// Process options: Mouse emulation (0x0e00-0x0eff)
//
//   pass: int c			option number
// return: void
//==============================================================================
static void options_mouse (int c)
{
 char *mouse_args[] =
 {
  "off", 
  "on",
  ""
 };

 int x;

 switch (c)
    {
     case OPT_MOUSE :
        x = string_search(mouse_args, e_optarg);
        if (x == -1) 
           {
            param_error_mesg("'--mouse'", e_optarg);
            exitstatus = 1;
           }
        else
           mouse.active = x;
        break;
    }
}

//==============================================================================
// Process options: Application dependent (0x0f00-0x0fff)
//
//   pass: int c			option number
// return: void
//==============================================================================
static void options_application (int c)
{
 int x;

 switch (c)
    {
     case OPT_FILE_APP :
        strncpy(func.file_app, e_optarg, sizeof(func.file_app));
        func.file_app[sizeof(func.file_app)-1] = 0;
        break;
     case OPT_FILE_EXEC :
        if ((int_arg < 0) || (int_arg > 65535))
           {
            param_error_mesg("'--file-exec'", e_optarg);
            exitstatus = 1;
           }
        else
           func.file_exec = int_arg;
        break;
     case OPT_FILE_EXIT :
        x = string_search(offon_args, e_optarg);
        if (x == -1)
           {
            param_error_mesg("'--file-exit'", e_optarg);
            exitstatus = 1;
           }
        else   
           func.file_exit = x;
        break;
     case OPT_FILE_LIST :
        if (func.file_list_count == FILE_LIST_ENTRIES)
           {
            param_error_mesg("'--file-list'", e_optarg);
            exitstatus = 1;
           }
        else   
           {
            strncpy(func.file_list[func.file_list_count], e_optarg_q, FILE_STR_SIZE);
            func.file_list[func.file_list_count][FILE_STR_SIZE-1] = 0;
            convert_slash(func.file_list[func.file_list_count++]);
           }
        break;
     case OPT_FILE_LIST_Q :
        if (func.file_list_count == FILE_LIST_ENTRIES)
           {
            param_error_mesg("'--file-list-q'", e_optarg);
            exitstatus = 1;
           }
        else
           {
            snprintf(func.file_list[func.file_list_count], FILE_STR_SIZE, "\"%s\"", e_optarg);
            func.file_list[func.file_list_count][FILE_STR_SIZE-1] = 0;
            convert_slash(func.file_list[func.file_list_count++]);
           } 
        break;
     case OPT_FILE_LOAD :
        if ((int_arg < 0) || (int_arg > 65535))
           {
            param_error_mesg("'--file-load'", e_optarg);
            exitstatus = 1;
           }
        else
           func.file_load = int_arg;
        break;   
     case OPT_FILE_RUN :
        strncpy(func.file_run, e_optarg, sizeof(func.file_run));
        func.file_run[sizeof(func.file_run)-1] = 0;
        break;
    }
}

//==============================================================================
// Process options.
//
// If any non-options are found these are processed last (see man getopt). 
// If any options processed causes exiting the options module before those
// non-options are reached then no error will be seen.
//
// An OPT_RUN constant is added to the option number if allowed to be
// processed when the emulator is already running.  An OPT_Z constant is
// added if the option can not be used when already running, some of these
// may be able to be added at a later date.
//
// In the "short options" string passed in xgetopt() a single ':' following
// an option letter indicates an argument for the optionis required.  Two
// '::' indicates an optional argument and no colon indicates no option. 
// The optarg string will be empty if an option is set to 'optional
// argument' type and no argument was sepecified.
//
// 256 option groups are possible (0x00nn-0xffnn) with each each group
// allowed upto 256 options.
//
//   pass: int argc			number of arguments
//         char *argv[]			pointer to pointer to arguments
// return: void
//==============================================================================
static void options_getopt (int argc, char *argv[])
{
 static struct option long_options[] =
    {
     // Control related
     {"account",        required_argument, 0, OPT_ACCOUNT       + OPT_Z  },
     {"alias-disks",    required_argument, 0, OPT_ALIAS_DISKS   + OPT_RUN},
     {"alias-roms",     required_argument, 0, OPT_ALIAS_ROMS    + OPT_RUN},
     {"args-error",     required_argument, 0, OPT_ARGS_ERROR    + OPT_RUN},
     {"bootkey",        required_argument, 0, OPT_BOOTKEY       + OPT_RUN},
     {"cfmode",         required_argument, 0, OPT_CFMODE        + OPT_Z  },
     {"config",         required_argument, 0, OPT_CONFIG        + OPT_RUN},
     {"cmd-repeat1",    required_argument, 0, OPT_CMD_REPEAT1   + OPT_RUN},
     {"cmd-repeat2",    required_argument, 0, OPT_CMD_REPEAT2   + OPT_RUN},
     {"cpu-delay",      required_argument, 0, OPT_CPU_DELAY     + OPT_RUN},
     {"dclick",         required_argument, 0, OPT_DCLICK        + OPT_RUN},
     {"exit",           required_argument, 0, OPT_EXIT          + OPT_RUN},
     {"exit-check",     required_argument, 0, OPT_EXIT_CHECK    + OPT_RUN},
     {"gui-persist",    required_argument, 0, OPT_GUI_PERSIST   + OPT_RUN},
     {"keystd-mod",     required_argument, 0, OPT_KEYSTD_MOD    + OPT_RUN},
     {"lockfix-win32",  required_argument, 0, OPT_LOCKFIX_WIN32 + OPT_RUN},
     {"lockfix-x11",    required_argument, 0, OPT_LOCKFIX_X11   + OPT_RUN},
     {"mmode",          no_argument,       0, OPT_MMODE         + OPT_RUN},
     {"mouse-wheel",    required_argument, 0, OPT_MOUSE_WHEEL   + OPT_RUN},
     {"nodisk",         no_argument,       0, OPT_NODISK        + OPT_RUN},
     {"options-warn",   required_argument, 0, OPT_OPTIONS_WARN  + OPT_RUN},
     {"osd",            required_argument, 0, OPT_OSD           + OPT_RUN},
     {"output",         required_argument, 0, OPT_OUTPUT        + OPT_RUN},
     {"prefix",         required_argument, 0, OPT_PREFIX        + OPT_Z  },
     {"runsecs",        required_argument, 0, OPT_RUNSECS       + OPT_RUN},
     {"slashes",        required_argument, 0, OPT_SLASHES       + OPT_RUN},
     {"spad",           required_argument, 0, OPT_SPAD          + OPT_RUN},
     {"status",         required_argument, 0, OPT_STATUS        + OPT_RUN},
     {"title",          required_argument, 0, OPT_TITLE         + OPT_RUN},
     {"varset",         required_argument, 0, OPT_VARSET        + OPT_RUN},
     {"varuset",        required_argument, 0, OPT_VARUSET       + OPT_RUN},
     {"verbose",        optional_argument, 0, OPT_VERBOSE       + OPT_RUN},

     // Conditional option parsing
     {"if-egt",         required_argument, 0, OPT_IF_EGT        + OPT_RUN},
     {"if-elt",         required_argument, 0, OPT_IF_ELT        + OPT_RUN},
     {"if-eq",          required_argument, 0, OPT_IF_EQ         + OPT_RUN},
     {"if-gt",          required_argument, 0, OPT_IF_GT         + OPT_RUN},
     {"if-lt",          required_argument, 0, OPT_IF_LT         + OPT_RUN},
     {"if-negt",        required_argument, 0, OPT_IF_NEGT       + OPT_RUN},
     {"if-nelt",        required_argument, 0, OPT_IF_NELT       + OPT_RUN},
     {"if-neq",         required_argument, 0, OPT_IF_NEQ        + OPT_RUN},
     {"if-ngt",         required_argument, 0, OPT_IF_NGT        + OPT_RUN},
     {"if-nlt",         required_argument, 0, OPT_IF_NLT        + OPT_RUN},
     {"if-nset",        required_argument, 0, OPT_IF_NSET       + OPT_RUN},
     {"if-set",         required_argument, 0, OPT_IF_SET        + OPT_RUN},
     {"if-system",      required_argument, 0, OPT_IF_SYSTEM     + OPT_RUN},
     {"if-false",       no_argument,       0, OPT_IF_FALSE      + OPT_RUN},
     {"if-true",        no_argument,       0, OPT_IF_TRUE       + OPT_RUN},
     {"if-else",        no_argument,       0, OPT_IF_ELSE       + OPT_RUN},
     {"if-end",         no_argument,       0, OPT_IF_END        + OPT_RUN},
     {"if-cmpmode",     required_argument, 0, OPT_IF_CMPMODE    + OPT_RUN},

     // Debugging tools
     {"bp",             required_argument, 0, OPT_BP            + OPT_RUN},
     {"bpr",            required_argument, 0, OPT_BPR           + OPT_RUN},
     {"bpc",            required_argument, 0, OPT_BPC           + OPT_RUN},
     {"debug",          required_argument, 0, OPT_DEBUG         + OPT_RUN}, // option (-z)
     {"dump",           required_argument, 0, OPT_DUMP          + OPT_RUN},
     {"echo",           required_argument, 0, OPT_ECHO          + OPT_RUN},
     {"echoq",          required_argument, 0, OPT_ECHOQ         + OPT_RUN},
     {"modio",          required_argument, 0, OPT_MODIO         + OPT_RUN},
     {"regs",           required_argument, 0, OPT_REGS          + OPT_RUN},

     // Disk drive images
     {"hdd0",           required_argument, 0, OPT_HDD0          + OPT_Z  }, // 0-2 are HDDs
     {"hdd1",           required_argument, 0, OPT_HDD1          + OPT_Z  },
     {"hdd2",           required_argument, 0, OPT_HDD2          + OPT_Z  },
     {"hdd3",           required_argument, 0, OPT_HDD3          + OPT_RUN}, // 3-6 are floppies
     {"hdd4",           required_argument, 0, OPT_HDD4          + OPT_RUN},
     {"hdd5",           required_argument, 0, OPT_HDD5          + OPT_RUN},
     {"hdd6",           required_argument, 0, OPT_HDD6          + OPT_RUN},

     {"hdd3-close",     no_argument,       0, OPT_HDD3_CLOSE    + OPT_RUN},
     {"hdd4-close",     no_argument,       0, OPT_HDD4_CLOSE    + OPT_RUN},
     {"hdd5-close",     no_argument,       0, OPT_HDD5_CLOSE    + OPT_RUN},
     {"hdd6-close",     no_argument,       0, OPT_HDD6_CLOSE    + OPT_RUN},

     {"ide-a0",         required_argument, 0, OPT_IDE_A0        + OPT_Z  },
     {"ide-a1",         required_argument, 0, OPT_IDE_A1        + OPT_Z  },
     {"ide-b0",         required_argument, 0, OPT_IDE_B0        + OPT_Z  },
     {"ide-b1",         required_argument, 0, OPT_IDE_B1        + OPT_Z  },

     {"image_a",        required_argument, 0, OPT_IMAGE_A       + OPT_RUN}, // option (-a)
     {"image_b",        required_argument, 0, OPT_IMAGE_B       + OPT_RUN}, // option (-b)
     {"image_c",        required_argument, 0, OPT_IMAGE_C       + OPT_RUN}, // option (-c)
     {"image_d",        required_argument, 0, OPT_IMAGE_D       + OPT_RUN}, // option (-d)

     {"a-close",        no_argument,       0, OPT_A_CLOSE       + OPT_RUN},
     {"b-close",        no_argument,       0, OPT_B_CLOSE       + OPT_RUN},
     {"c-close",        no_argument,       0, OPT_C_CLOSE       + OPT_RUN},
     {"d-close",        no_argument,       0, OPT_D_CLOSE       + OPT_RUN},

#ifdef USE_LIBDSK
     {"dstep",          no_argument,       0, OPT_DSTEP         + OPT_RUN},
     {"dstep-hd",       no_argument,       0, OPT_DSTEP_HD      + OPT_RUN},
     {"format",         required_argument, 0, OPT_FORMAT        + OPT_RUN},
     {"lformat",        no_argument,       0, OPT_LFORMAT       + OPT_RUN},
     {"ltype",          no_argument,       0, OPT_LTYPE         + OPT_RUN},
     {"side1as0",       no_argument,       0, OPT_SIDE1AS0      + OPT_RUN},
     {"type",           required_argument, 0, OPT_TYPE          + OPT_RUN},
#endif
     {"psec",           no_argument,       0, OPT_PSEC          + OPT_RUN},

     // Display related
     {"aspect",         required_argument, 0, OPT_ASPECT        + OPT_Z  },
     {"fullscreen",     optional_argument, 0, OPT_FULLSCREEN    + OPT_Z  }, // option (-f)
     {"monitor",        required_argument, 0, OPT_MONITOR       + OPT_RUN}, // option (-m)

     {"mon-bg-b",       required_argument, 0, OPT_MON_BG_B      + OPT_RUN},
     {"mon-bg-g",       required_argument, 0, OPT_MON_BG_G      + OPT_RUN},
     {"mon-bg-r",       required_argument, 0, OPT_MON_BG_R      + OPT_RUN},
     {"mon-bgi-b",      required_argument, 0, OPT_MON_BGI_B     + OPT_RUN},
     {"mon-bgi-g",      required_argument, 0, OPT_MON_BGI_G     + OPT_RUN},
     {"mon-bgi-r",      required_argument, 0, OPT_MON_BGI_R     + OPT_RUN},
     {"mon-fg-b",       required_argument, 0, OPT_MON_FG_B      + OPT_RUN},
     {"mon-fg-g",       required_argument, 0, OPT_MON_FG_G      + OPT_RUN},
     {"mon-fg-r",       required_argument, 0, OPT_MON_FG_R      + OPT_RUN},
     {"mon-fgi-b",      required_argument, 0, OPT_MON_FGI_B     + OPT_RUN},
     {"mon-fgi-g",      required_argument, 0, OPT_MON_FGI_G     + OPT_RUN},
     {"mon-fgi-r",      required_argument, 0, OPT_MON_FGI_R     + OPT_RUN},

     {"mon-fgl-b",      required_argument, 0, OPT_MON_FGL_B     + OPT_RUN}, // these 3 report warnings
     {"mon-fgl-g",      required_argument, 0, OPT_MON_FGL_G     + OPT_RUN},
     {"mon-fgl-r",      required_argument, 0, OPT_MON_FGL_R     + OPT_RUN},

     {"rgb-00-r",       required_argument, 0, OPT_RGB_00_R      + OPT_RUN},
     {"rgb-00-g",       required_argument, 0, OPT_RGB_00_G      + OPT_RUN},
     {"rgb-00-b",       required_argument, 0, OPT_RGB_00_B      + OPT_RUN},
     {"rgb-01-r",       required_argument, 0, OPT_RGB_01_R      + OPT_RUN},
     {"rgb-01-g",       required_argument, 0, OPT_RGB_01_G      + OPT_RUN},
     {"rgb-01-b",       required_argument, 0, OPT_RGB_01_B      + OPT_RUN},
     {"rgb-02-r",       required_argument, 0, OPT_RGB_02_R      + OPT_RUN},
     {"rgb-02-g",       required_argument, 0, OPT_RGB_02_G      + OPT_RUN},
     {"rgb-02-b",       required_argument, 0, OPT_RGB_02_B      + OPT_RUN},
     {"rgb-03-r",       required_argument, 0, OPT_RGB_03_R      + OPT_RUN},
     {"rgb-03-g",       required_argument, 0, OPT_RGB_03_G      + OPT_RUN},
     {"rgb-03-b",       required_argument, 0, OPT_RGB_03_B      + OPT_RUN},
     {"rgb-04-r",       required_argument, 0, OPT_RGB_04_R      + OPT_RUN},
     {"rgb-04-g",       required_argument, 0, OPT_RGB_04_G      + OPT_RUN},
     {"rgb-04-b",       required_argument, 0, OPT_RGB_04_B      + OPT_RUN},
     {"rgb-05-r",       required_argument, 0, OPT_RGB_05_R      + OPT_RUN},
     {"rgb-05-g",       required_argument, 0, OPT_RGB_05_G      + OPT_RUN},
     {"rgb-05-b",       required_argument, 0, OPT_RGB_05_B      + OPT_RUN},
     {"rgb-06-r",       required_argument, 0, OPT_RGB_06_R      + OPT_RUN},
     {"rgb-06-g",       required_argument, 0, OPT_RGB_06_G      + OPT_RUN},
     {"rgb-06-b",       required_argument, 0, OPT_RGB_06_B      + OPT_RUN},
     {"rgb-07-r",       required_argument, 0, OPT_RGB_07_R      + OPT_RUN},
     {"rgb-07-g",       required_argument, 0, OPT_RGB_07_G      + OPT_RUN},
     {"rgb-07-b",       required_argument, 0, OPT_RGB_07_B      + OPT_RUN},
     {"rgb-08-r",       required_argument, 0, OPT_RGB_08_R      + OPT_RUN},
     {"rgb-08-g",       required_argument, 0, OPT_RGB_08_G      + OPT_RUN},
     {"rgb-08-b",       required_argument, 0, OPT_RGB_08_B      + OPT_RUN},
     {"rgb-09-r",       required_argument, 0, OPT_RGB_09_R      + OPT_RUN},
     {"rgb-09-g",       required_argument, 0, OPT_RGB_09_G      + OPT_RUN},
     {"rgb-09-b",       required_argument, 0, OPT_RGB_09_B      + OPT_RUN},
     {"rgb-10-r",       required_argument, 0, OPT_RGB_10_R      + OPT_RUN},
     {"rgb-10-g",       required_argument, 0, OPT_RGB_10_G      + OPT_RUN},
     {"rgb-10-b",       required_argument, 0, OPT_RGB_10_B      + OPT_RUN},
     {"rgb-11-r",       required_argument, 0, OPT_RGB_11_R      + OPT_RUN},
     {"rgb-11-g",       required_argument, 0, OPT_RGB_11_G      + OPT_RUN},
     {"rgb-11-b",       required_argument, 0, OPT_RGB_11_B      + OPT_RUN},
     {"rgb-12-r",       required_argument, 0, OPT_RGB_12_R      + OPT_RUN},
     {"rgb-12-g",       required_argument, 0, OPT_RGB_12_G      + OPT_RUN},
     {"rgb-12-b",       required_argument, 0, OPT_RGB_12_B      + OPT_RUN},
     {"rgb-13-r",       required_argument, 0, OPT_RGB_13_R      + OPT_RUN},
     {"rgb-13-g",       required_argument, 0, OPT_RGB_13_G      + OPT_RUN},
     {"rgb-13-b",       required_argument, 0, OPT_RGB_13_B      + OPT_RUN},
     {"rgb-14-r",       required_argument, 0, OPT_RGB_14_R      + OPT_RUN},
     {"rgb-14-g",       required_argument, 0, OPT_RGB_14_G      + OPT_RUN},
     {"rgb-14-b",       required_argument, 0, OPT_RGB_14_B      + OPT_RUN},
     {"rgb-15-r",       required_argument, 0, OPT_RGB_15_R      + OPT_RUN},
     {"rgb-15-g",       required_argument, 0, OPT_RGB_15_G      + OPT_RUN},
     {"rgb-15-b",       required_argument, 0, OPT_RGB_15_B      + OPT_RUN},
     
     {"video",          required_argument, 0, OPT_VIDEO         + OPT_RUN},
     {"video-depth",    required_argument, 0, OPT_VIDEO_DEPTH   + OPT_Z  },
     {"video-type",     required_argument, 0, OPT_VIDEO_TYPE    + OPT_Z  },

#ifdef USE_OPENGL
     // OpenGL rendering
     {"gl-aspect-bee",  required_argument, 0, OPT_GL_ASPECT_BEE + OPT_RUN},
     {"gl-aspect-mon",  required_argument, 0, OPT_GL_ASPECT_MON + OPT_RUN},
     {"gl-filter-fs",   required_argument, 0, OPT_GL_FILTER_FS  + OPT_RUN},
     {"gl-filter-max",  required_argument, 0, OPT_GL_FILTER_MAX + OPT_RUN},
     {"gl-filter-win",  required_argument, 0, OPT_GL_FILTER_WIN + OPT_RUN},
     {"gl-max",         required_argument, 0, OPT_GL_MAX        + OPT_Z  },
     {"gl-vsync",       required_argument, 0, OPT_GL_VSYNC      + OPT_Z  },
     {"gl-winpct",      required_argument, 0, OPT_GL_WINPCT     + OPT_Z  },
     {"gl-winpix",      required_argument, 0, OPT_GL_WINPIX     + OPT_Z  },
#endif

     // Model emulation
     {"basic",          required_argument, 0, OPT_BASIC         + OPT_Z  },
     {"basica",         required_argument, 0, OPT_BASICA        + OPT_Z  },
     {"basicb",         required_argument, 0, OPT_BASICB        + OPT_Z  },
     {"basicc",         required_argument, 0, OPT_BASICC        + OPT_Z  },
     {"basicd",         required_argument, 0, OPT_BASICD        + OPT_Z  },

     {"basram",         no_argument,       0, OPT_BASRAM        + OPT_Z  },
     {"charrom",        required_argument, 0, OPT_CHARROM       + OPT_Z  },
     {"col",            no_argument,       0, OPT_COL           + OPT_RUN},
     {"col-type",       required_argument, 0, OPT_COL_TYPE      + OPT_RUN},
     {"colprom",        required_argument, 0, OPT_COLPROM       + OPT_Z  },
     {"dint",           required_argument, 0, OPT_DINT          + OPT_RUN},
     {"hint",           required_argument, 0, OPT_HINT          + OPT_RUN},
     {"hardware",       required_argument, 0, OPT_HARDWARE      + OPT_Z  },
     {"hwflash",        required_argument, 0, OPT_HWFLASH       + OPT_RUN},
     {"hwflashr",       required_argument, 0, OPT_HWFLASHR      + OPT_RUN},
     {"lmodel",         no_argument,       0, OPT_LMODEL        + OPT_RUN},
     {"lpen",           no_argument,       0, OPT_LPEN          + OPT_RUN},
     {"model",          required_argument, 0, OPT_MODEL         + OPT_Z  },
     {"mono",           no_argument,       0, OPT_MONO          + OPT_RUN},
     {"netram",         no_argument,       0, OPT_NETRAM        + OPT_Z  },
     {"netrom",         required_argument, 0, OPT_NETROM        + OPT_Z  },

     {"pak",            required_argument, 0, OPT_PAK           + OPT_Z  },
     {"pak0",           required_argument, 0, OPT_PAK0          + OPT_Z  },
     {"pak1",           required_argument, 0, OPT_PAK1          + OPT_Z  },
     {"pak2",           required_argument, 0, OPT_PAK2          + OPT_Z  },
     {"pak3",           required_argument, 0, OPT_PAK3          + OPT_Z  },
     {"pak4",           required_argument, 0, OPT_PAK4          + OPT_Z  },
     {"pak5",           required_argument, 0, OPT_PAK5          + OPT_Z  },
     {"pak6",           required_argument, 0, OPT_PAK6          + OPT_Z  },
     {"pak7",           required_argument, 0, OPT_PAK7          + OPT_Z  },

     {"pakram",         required_argument, 0, OPT_PAKRAM        + OPT_Z  },

     {"pcg",            required_argument, 0, OPT_PCG           + OPT_RUN},
     {"piob7",          required_argument, 0, OPT_PIOB7         + OPT_RUN},

     {"rom1",           required_argument, 0, OPT_ROM1          + OPT_Z  },
     {"rom2",           required_argument, 0, OPT_ROM2          + OPT_Z  },
     {"rom3",           required_argument, 0, OPT_ROM3          + OPT_Z  },

     {"rom256k",        required_argument, 0, OPT_ROM256K       + OPT_Z  },

     {"sys",            required_argument, 0, OPT_SYS           + OPT_Z  },
     {"vdu",            required_argument, 0, OPT_VDU           + OPT_RUN},

     // Information output
     {"conio",          no_argument,       0, OPT_CONIO         + OPT_Z  },
     {"help",           no_argument,       0, OPT_HELP          + OPT_RUN}, // option (-h)
     {"lcon",           no_argument,       0, OPT_LCON          + OPT_RUN},
     {"lconw",          no_argument,       0, OPT_LCONW         + OPT_RUN},
     {"lcons",          required_argument, 0, OPT_LCONS         + OPT_RUN},
     {"usage",          no_argument,       0, OPT_USAGE         + OPT_RUN}, // option (-h)
     {"version",        no_argument,       0, OPT_VERSION       + OPT_RUN},

     // Parallel printer port emulation
     {"print",          required_argument, 0, OPT_PRINT         + OPT_RUN},
     {"print-close",    no_argument,       0, OPT_PRINT_CLOSE   + OPT_RUN},
     {"printa",         required_argument, 0, OPT_PRINTA        + OPT_RUN},
     {"printa-close",   no_argument,       0, OPT_PRINTA_CLOSE  + OPT_RUN},

     // Serial port emulation
     {"baud",           required_argument, 0, OPT_BAUD          + OPT_RUN},
     {"baudrx",         required_argument, 0, OPT_BAUDRX        + OPT_RUN},
     {"baudtx",         required_argument, 0, OPT_BAUDTX        + OPT_RUN},
     {"coms",           required_argument, 0, OPT_COMS          + OPT_RUN},
     {"coms-close",     no_argument,       0, OPT_COMS_CLOSE    + OPT_RUN},
     {"datab",          required_argument, 0, OPT_DATAB         + OPT_RUN},
     {"stopb",          required_argument, 0, OPT_STOPB         + OPT_RUN},

     // Sound emulation
     {"sound",          required_argument, 0, OPT_SOUND         + OPT_Z  },
     {"snd-alg1",       required_argument, 0, OPT_SND_ALG1      + OPT_RUN},
     {"snd-freq",       required_argument, 0, OPT_SND_FREQ      + OPT_Z  },
     {"snd-freqadj",    required_argument, 0, OPT_SND_FREQADJ   + OPT_Z  },
     {"snd-freqlow",    required_argument, 0, OPT_SND_FREQLOW   + OPT_Z  },
     {"snd-holdoff",    required_argument, 0, OPT_SND_HOLDOFF   + OPT_RUN},
     {"snd-hq",         no_argument,       0, OPT_SND_HQ        + OPT_Z  },
     {"snd-mute",       required_argument, 0, OPT_SND_MUTE      + OPT_RUN},
     {"snd-samples",    required_argument, 0, OPT_SND_SAMPLES   + OPT_Z  },
     {"snd-volume",     required_argument, 0, OPT_SND_VOLUME    + OPT_RUN}, // option (-v)
     {"vol",            required_argument, 0, OPT_VOL           + OPT_RUN}, // option (-v)

     // Speed related
     {"clock",          required_argument, 0, OPT_CLOCK         + OPT_RUN}, // same as 'xtal'
     {"clock-def",      required_argument, 0, OPT_CLOCK_DEF     + OPT_Z  },
     {"frate",          required_argument, 0, OPT_FRATE         + OPT_RUN},
     {"maxcpulag",      required_argument, 0, OPT_MAXCPULAG     + OPT_RUN},
     {"vblank",         required_argument, 0, OPT_VBLANK        + OPT_RUN},
     {"xtal",           required_argument, 0, OPT_XTAL          + OPT_RUN}, // option (-x)
     {"speedsel",       required_argument, 0, OPT_SPEEDSEL      + OPT_RUN},
     {"turbo",          optional_argument, 0, OPT_TURBO         + OPT_RUN}, // option (-t)
     {"z80div",         required_argument, 0, OPT_Z80DIV        + OPT_RUN},

     // Tape port emulation
     {"tapei",          required_argument, 0, OPT_TAPEI         + OPT_RUN},
     {"tapei-close",    no_argument,       0, OPT_TAPEI_CLOSE   + OPT_RUN},
     {"tapei-det",      required_argument, 0, OPT_TAPE_DET      + OPT_RUN},
     {"tapeo",          required_argument, 0, OPT_TAPEO         + OPT_RUN},
     {"tapeo-close",    no_argument,       0, OPT_TAPEO_CLOSE   + OPT_RUN},
     {"tapesamp",       required_argument, 0, OPT_TAPESAMP      + OPT_RUN},
     {"tapevol",        required_argument, 0, OPT_TAPEVOL       + OPT_RUN},

     // Real Time Clock (RTC) emulation and time
     {"century",        required_argument, 0, OPT_CENTURY       + OPT_Z  },
     {"rtc",            required_argument, 0, OPT_RTC           + OPT_Z  },

     // Joystick emulation
     {"js",             required_argument, 0, OPT_JS            + OPT_Z  },

     {"js-axis",        required_argument, 0, OPT_JS_AXIS       + OPT_RUN},
     {"js-axisb",       required_argument, 0, OPT_JS_AXISB      + OPT_RUN},
     {"js-axisl",       required_argument, 0, OPT_JS_AXISL      + OPT_RUN},

     {"js-clear",       no_argument,       0, OPT_JS_CLEAR      + OPT_RUN},
     {"js-mbee",        required_argument, 0, OPT_JS_MBEE       + OPT_RUN},
     {"js-up",          required_argument, 0, OPT_JS_UP         + OPT_RUN},
     {"js-right",       required_argument, 0, OPT_JS_RIGHT      + OPT_RUN},
     {"js-down",        required_argument, 0, OPT_JS_DOWN       + OPT_RUN},
     {"js-left",        required_argument, 0, OPT_JS_LEFT       + OPT_RUN},
     {"js-fire",        required_argument, 0, OPT_JS_FIRE       + OPT_RUN},
     {"js-play1",       required_argument, 0, OPT_JS_PLAY1      + OPT_RUN},
     {"js-play2",       required_argument, 0, OPT_JS_PLAY2      + OPT_RUN},
     {"js-spare",       required_argument, 0, OPT_JS_SPARE      + OPT_RUN},

     {"js-clist",       no_argument,       0, OPT_JS_CLIST      + OPT_RUN},
     {"js-klist",       no_argument,       0, OPT_JS_KLIST      + OPT_RUN},
     {"js-kbd",         required_argument, 0, OPT_JS_KBD        + OPT_RUN},
     {"js-kk",          required_argument, 0, OPT_JS_KK         + OPT_RUN},
     {"js-kb",          required_argument, 0, OPT_JS_KB         + OPT_RUN},
     {"js-kset",        required_argument, 0, OPT_JS_KSET       + OPT_RUN},
     {"js-ksel",        required_argument, 0, OPT_JS_KSEL       + OPT_RUN},

     // Mouse emulation     
     {"mouse",          required_argument, 0, OPT_MOUSE         + OPT_RUN},     

     // Application dependent
     {"file-app",       required_argument, 0, OPT_FILE_APP      + OPT_RUN},
     {"file-exec",      required_argument, 0, OPT_FILE_EXEC     + OPT_RUN},
     {"file-exit",      required_argument, 0, OPT_FILE_EXIT     + OPT_RUN},
     {"file-list",      required_argument, 0, OPT_FILE_LIST     + OPT_RUN},
     {"file-list-q",    required_argument, 0, OPT_FILE_LIST_Q   + OPT_RUN},
     {"file-load",      required_argument, 0, OPT_FILE_LOAD     + OPT_RUN},
     {"file-run",       required_argument, 0, OPT_FILE_RUN      + OPT_RUN},

     {0,                0,                 0,               0}
    };

 // table to convert short options to a long option number >= 0x100
 static short_options_trans_t short_options[] =
    {
     {'a', OPT_IMAGE_A    + OPT_RUN},
     {'b', OPT_IMAGE_B    + OPT_RUN},
     {'c', OPT_IMAGE_C    + OPT_RUN},
     {'d', OPT_IMAGE_D    + OPT_RUN},
     {'f', OPT_FULLSCREEN + OPT_Z  },
     {'h', OPT_HELP       + OPT_RUN},
     {'m', OPT_MONITOR    + OPT_RUN},
     {'t', OPT_TURBO      + OPT_RUN},
     {'v', OPT_SND_VOLUME + OPT_RUN},
     {'x', OPT_CLOCK      + OPT_RUN},
     {'z', OPT_DEBUG      + OPT_RUN},
     {0,                     0x0000}
    };

 int i;
 int c;
 int short_option;
 int allowed;

 // xgetopt_long stores the long option index here.
 int long_index = 0;

 // clear these each time the options are processed
#ifdef USE_LIBDSK
 use_driver_type[0] = 0;
 use_format_type[0] = 0;
 side1as0 = 0;  
 dstep = 0;
 dstep_hd = 0; 
#endif 

 // the xgetopt variables must be reset each time this function is re-entered.
 xgetopt_init();

 while (! exitstatus)
    {
     c = xgetopt_long(argc, argv, "f::ht::a:b:c:d:m:v:x:z:", long_options, &long_index);

     // print any error messages generated by xgetopt_long()
     if (opterr_msg[0])
        {
         exitstatus = 1;
         xprintf("%s", opterr_msg);
         xprintf(TRY_MESG, argv[0], argv[0], argv[0]);
        }

     // detect the end of the options
     if (c == -1)
        break;

     // translate short options to a long option number
     short_option = (c < 0x100) * c;
     if (short_option)
        {
         i = 0;
         while ((short_options[i].option) && (short_options[i].option != c))
            i++;
         if (short_options[i].option == c)
            c = short_options[i].longno;
        }

     // check if processing an option is allowed when in run mode
     allowed = (! emu.runmode) || (c < 0x100) ||  ((c & OPT_RUN) != 0);
     
     if (! allowed)   
        {
         if (runmode_warn)
            {
             if (short_option)
                xprintf("ubee512: WARNING ! option `-%c' not supported in run mode (ignored).\n", short_option);
             else
                xprintf("ubee512: WARNING ! option `--%s' not supported in run mode (ignored).\n", long_options[long_index].name);
             c = 0;  // keep going but don't process the option
            }
         else
            {
             if (short_option)
                xprintf("ubee512: option `-%c' not supported in run mode.\n", short_option);
             else
                xprintf("ubee512: option `--%s' not supported in run mode.\n", long_options[long_index].name);
             exitstatus = -1;
             break;  // abort the remainder of the command line
            } 
        }

     c = c & 0x0000ffff;  // strip of the flag bits
     
     // don't process options if currently off (except for --if-x options)
     if (((c < 0x0180) || (c > 0x01ff)) && ((c != '?') && (c != ':')))
        {
         if (if_pos >= OPTIONS_MAXCOND)
            {
             xprintf("ubee512: all %d levels of conditionals used up!\n", OPTIONS_MAXCOND);
             exitstatus = -1;
             c = 0;
            } 
         if ((if_state[if_pos]) == 0)
            c = 0;
        }
     
     if_state_prev = if_state[if_pos];

     if (c)
        {
         // extract environment variables from the argument
         extract_environment_vars(optarg, e_optarg, e_optarg_q);
         toupper_string(e_optarg_u, e_optarg);

//       xprintf("optarg=%s, e_optarg=%s e_optarg_q=%s\n", optarg, e_optarg, e_optarg_q);

         // do integer and floating point conversions
         if (e_optarg != NULL)
            {
             int_arg = -1;
             if ((strchr(e_optarg, 'x') == NULL) && (strchr(e_optarg, 'X') == NULL))
                sscanf(e_optarg, "%d", &int_arg);
             else   
                sscanf(e_optarg, "%x", &int_arg);         
             float_arg = -1.0;
             sscanf(e_optarg, "%f", &float_arg);
            }

         switch (c & 0xff00)
            {
             case 0x0000 : // Short options
                options_short(c, argv);
                break;
             case 0x0100 : // Control related 
                options_control(c);
                break;
             case 0x0200 : // Debugging tools
                options_debugging(c);
                break;
             case 0x0300 : // Disk drive images
                options_disks(c);
                break;
             case 0x0400 : // Display related
                options_display(c);
                break;
             case 0x0500 : // Model emulation
                options_model(c);
                break;
             case 0x0600 : // Information output
                options_information(c);
                break;
             case 0x0700 : // Parallel printer port emulation
                options_parallel_port(c);
                break;
             case 0x0800 : // Serial port
                options_serial_port(c);
                break;
             case 0x0900 : // Sound emulation
                options_sound(c);
                break;
             case 0x0a00 : // Speed related
                options_speed(c);
                break;
             case 0x0b00 : // Tape port
                options_tape(c);
                break;
             case 0x0c00 : // Real Time Clock (RTC) emulation
                options_rtc(c);
                break;
             case 0x0d00 : // Joystick emulation
                options_joystick(c);
                break;
             case 0x0e00 : // Mouse emulation
                options_mouse(c);
                break;
             case 0x0f00 : // Application dependent
                options_application(c);
                break;
             case 0x1000 : // reserved
                break;
                default : exitstatus = 1;
                break;
            }
        }
    }

 // Check if any other arguments are incorrectly specified
 if ((optind < argc) && (exitstatus == 0) && (args_err_flags & 0x01))
    {
     exitstatus = 1;
     xprintf("%s: %d additional arguments were specified that are not recognised:\n", argv[0], argc - optind);
     xprintf("%s: ", argv[0]);
     while (optind < argc)
        xprintf("%s ", argv[optind++]);
     xprintf("\n");
     xprintf(TRY_MESG, argv[0], argv[0], argv[0]);
    }
}

//==============================================================================
// Allocate memory.
//
//   pass: int size
// return: char *			result of memory allocate
//==============================================================================
static char *options_malloc (int size)
{
 char *ptr;
 
 if (! (ptr = malloc(size)))
    {
     exitstatus = 1;
     xprintf("options_malloc: unable to malloc %d bytes of memory\n", size);
    } 

 return ptr; 
}

//==============================================================================
// Build a list of definitions found in the configuration file.
//
//   pass: void
// return: void
//==============================================================================
static void options_buildlist (void)
{
 int l;
 char s[OPTIONS_SIZE];
 
 if (! fp)
    return;

 fseek(fp, 0, SEEK_SET);

 while ((l = file_readline(fp, s, OPTIONS_SIZE)))
    {
     if ((s[0] == '[') && (s[l-1] == ']'))
        {
         if (! (ndefsv[ndefsc] = options_malloc(l-1)))  // space for name + 0
            return;
         strncpy(ndefsv[ndefsc], &s[1], l-2);
         ndefsv[ndefsc++][l-2] = 0;
        }
    }
}

//==============================================================================
// Find a definitions entry in the names list.
//
//   pass: const char *name		name of the definition to look for
// return: int				0 if no match found, else index +1
//==============================================================================
static int options_findentry (const char *name)
{
 int i = 0;
 
 if (! fp)
    return 0;

 while ((i < ndefsc) && (strcmp(ndefsv[i], name) != 0))
    i++;
    
 if (i < ndefsc)
    return i+1;
    
 return 0;
}

//==============================================================================
// Get options for the named entry from the configuration file.
//
//   pass: const char *name		definition name
//         char *options		return options found for definition
//                                      string if entry not found or empty
// return: void
//==============================================================================
static void options_getoptstr (const char *name, char *options)
{
 int l;
 int finish = 0;

 char s[OPTIONS_SIZE];

 if (! fp)
    return;

 fseek(fp, 0, SEEK_SET);

 options[0] = 0;

 // find the definition entry 
 while ((! finish) && (l = file_readline(fp, s, OPTIONS_SIZE)))
    {
     if ((s[0] == '[') && (s[l-1] == ']'))
        {
         strncpy(s, &s[1], l-2);
         s[l-2] = 0;
         finish = (strcmp(s, name) == 0);
        }
    }

 if (! finish)
    return;

 finish = 0;

 // concatenate all the definition entries into one string
 while (! finish)
    {
     if ((l = file_readline(fp, s, OPTIONS_SIZE)))
        {
         if (! (finish = ((s[0] == '[') && (s[l-1] == ']'))))
            {
             l = OPTIONS_SIZE - (strlen(options) + 1);
             if (l > 0)
                {
                 strcat(options, " ");
                 strncat(options, s, l-1);
                } 
            }
        }
     else
        finish = 1;
    }

 return;   
}

//==============================================================================
// Parse the options from the config file entry string and place into a
// xargv variable.
//
//   pass: const char *options		options to be parsed
// return: void
//==============================================================================
static void options_parse (const char *options)
{
 int i = 0;
 int l;

 char s[OPTIONS_SIZE];
 
 while (options[i])
    {
     while ((options[i]) && (options[i] <= ' '))  // move past white space
        i++;

     l = 0;   
     while ((options[i]) && (options[i] != ' ') && (options[i] != '"'))     
        {
#ifdef MINGW
#else
         if ((options[i] == '\\') && (! emu.slashconv))
            i++;
#endif   
         if (options[i])
            {
             s[l++] = options[i];
             i++;
            } 
        }
     if (options[i] == '"')
        {
         i++;
         while ((options[i]) && (options[i] != '"'))
            {
             s[l++] = options[i];
             i++;
            }
         i++;   
        }   
     s[l] = 0;
     
     if (! (xargv[xargc] = options_malloc(l+1)))  // space for argument + 0
        return;
     strcpy(xargv[xargc++], s);
    }
}

//==============================================================================
// Process early options found on the command line, these are special options
// and if used must be declared before all others.  These options will be
// ignored by the options_getopt() function.
//
// --account option:
// By default the account is @UBEE_USERHOME@/.ubee512 on Unices and the
// location of the executed ubee512.exe on Windows. This option may be used
// when an alternative account location is required. If this option is used
// it must be the first option on the command line.
//
// --config option:
// By default the ubee512rc file found in the home account is used unless
// another file is specified. If this option is used it must be the first or
// second option on the command line.
//
// Note: The number of arguments (argc) consists of the number of command
// line space delimited arguments (except where quoted), the program name is
// also included in the count.  The first argument is therefore the second
// argv pointer.
//
//   pass: char *s			pointer to string
//         int argc			number of arguments
//         char *argv[]			pointer to pointer to arguments
// return: int				0 if no error, 1 if error, neg if exit
//==============================================================================
static int options_early (char *s, int argc, char *argv[])
{
 int argv_pos = 1;

 // test if --account is the first command line option, this is used to set an
 // alternative '.ubee512' home account directory.
 if (argc > 1)
    {
     // option & arg may be 1 entry (i.e. --option=arg)
     if (strncmp("--account=", argv[argv_pos], 10) == 0)
        {
         extract_environment_vars(&argv[argv_pos][10], userhome, e_optarg_q);
         emu.home_account_set = 1;
         argv_pos++;
         argc--;
        } 
     else
        // option & arg may be broken up over 2 entries (i.e. --option arg)
        if (strcmp("--account", argv[argv_pos]) == 0)
           {
            argv_pos++;
            extract_environment_vars(argv[argv_pos], userhome, e_optarg_q);
            emu.home_account_set = 1;
            argv_pos++;
            argc -= 2;
           }   
    }

 // set all the account paths now that we know the home account to be used.
 if (set_account_paths())
    return 1;

 // test if --config is the next command line option,  this is used
 // to set an alternative configuration file.
 if (argc > 1)
    {
     // option & arg may be 1 entry (i.e. --option=arg)
     if (strncmp("--config=", argv[argv_pos], 9) == 0)
        extract_environment_vars(&argv[argv_pos][9], s, e_optarg_q);
     else
        // option & arg may be broken up over 2 entries (i.e. --option arg)
        if (strcmp("--config", argv[argv_pos]) == 0)
           extract_environment_vars(argv[++argv_pos], s, e_optarg_q);
    }

// xprintf("config=%s  userhome=%s\n", s, userhome);
    
 return 0;   
}

//==============================================================================
// Process all options from the pointer list and the configuration file.
//
// If processing in run mode then the [global-start] and [global-end]
// sections will be ignored.  Other sections in the configuration file will
// be processed if specified.
//
//   pass: int argc			number of arguments
//         char *argv[]			pointer to pointer to arguments
// return: int				0 if no error, 1 if error, neg if exit
//==============================================================================
int options_process (int argc, char *argv[])
{
 int i;

 char filepath[SSIZE1];
 char file_options[OPTIONS_SIZE];
 char file_section[40];

 exitstatus = 0;

 // close any open configuration file
 if (fp)
    fclose(fp);

 // free existing pointers in use for section entry names list
 while (ndefsc)
    free(ndefsv[--ndefsc]);

 // free existing pointers in use for arguments
 while (xargc)
    free(xargv[--xargc]);

 // set conditionals state
 if_pos = 0;
 if_state[if_pos] = 1;  // true at the base level

 // set default configuration file to use if none was set previously
 if (! config_file[0])
    strcpy(config_file, "ubee512rc");

 // process any special early options from the command line, the
 // configuration file may be changed with an option here.
 i = options_early(config_file, argc, argv);
 if (i)
    return i;

 // if use no configuration file has been requested then NULL the file pointer
 if (strcmp("none", config_file) == 0)
    fp = NULL;
 else
    {
     fp = open_file(config_file, userhome_confpath, filepath, "r");
     if (strlen(config_file) && (! fp) && (emu.verbose))
        xprintf("options_process: Configuration file not found: %s\n", config_file);
    }

 // build a list of the definitions contained in the configuration file
 options_buildlist();
 if (exitstatus)
    return exitstatus;

 // first argument will be the program name and must be inserted first
 if (! (xargv[xargc] = options_malloc(strlen(argv[0])+1))) // space for argument + 0
    return exitstatus;

 strcpy(xargv[xargc++], argv[0]);

 // process the configuration file's 'global-start' or 'global-start-runmode' options
 if (emu.runmode)
    strcpy(file_section, "global-start-runmode");
 else
    strcpy(file_section, "global-start");
 if (options_findentry(file_section))
    {
     options_getoptstr(file_section, file_options);
     if (strlen(file_options))
        options_parse(file_options);
    }
 if (exitstatus)
    return exitstatus;
    
 // process all the command line and configuration file arguments after the program name
 for (i = 1; i < argc; i++)
    {
     // if it's a command line option (-) or the entry is not a section entry
     if ((*argv[i] == '-') || (! options_findentry(argv[i])))
        {
         if (! (xargv[xargc] = options_malloc(strlen(argv[i])+1)))  // space for argument + 0
            return exitstatus;

         strcpy(xargv[xargc++], argv[i]);
        }
     else
        {
         // otherwise if a section entry grab all the options from the file
         if (options_findentry(argv[i]))
            {
             options_getoptstr(argv[i], file_options);
             if (strlen(file_options))
                options_parse(file_options);
            }
        }
    }

 // process the configuration file's 'global-end' or 'global-end-runmode' options
 if (emu.runmode)
    strcpy(file_section, "global-end-runmode");
 else
    strcpy(file_section, "global-end");
 if (options_findentry(file_section))
    {
     options_getoptstr(file_section, file_options);
     if (strlen(file_options))
        options_parse(file_options);
    }

 // process all the options
 if (! exitstatus)
    options_getopt(xargc, xargv);

 // report any options information if requested
 options_modio_info();
 
 return exitstatus;
}
